/*
Module : PyDebugDoc.cpp
Purpose: Defines the implementation for the classes which encapsulate the MFC Document functionality of PyDebug++
Created: PJN / 20-08-2015
History: PJN / 28-08-2016 1. Fixed a bug where breakpoints would not get disabled in the Scintilla UI when the 
                          document is modified.
         PJN / 01-08-2022 1. Updated all the code in to use C++ uniform initialization for all variable
                          declarations.

Copyright (c) 2015 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "PyDebugDoc.h"
#include "PyDebugView.h"
#include "PyDebugMainFrm.h"


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26409 26433 26440 26477)
IMPLEMENT_DYNCREATE(CPyDebugDoc, CScintillaDoc)

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CPyDebugDoc, CScintillaDoc) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
END_MESSAGE_MAP()

void CPyDebugDoc::UpdateBreakpointsToUI(_In_z_ LPCTSTR lpszPathName)
{
  //Validate our parameters
  CPyDebugFrameWnd* pMainFrame{GetMainFrameWnd()};
  ASSERT(pMainFrame != nullptr);

  DeleteAllBreakpoints();

  {
    CSingleLock lock{&pMainFrame->m_csBreakpoints, TRUE};
#pragma warning(suppress: 26489)
    for (const auto& breakpoint : pMainFrame->m_Breakpoints)
    {
#pragma warning(suppress: 26489)
      if (breakpoint.m_sModule == lpszPathName)
#pragma warning(suppress: 26486 26489)
        UpdateBreakpoint(breakpoint);
    }
  }
}

void CPyDebugDoc::UpdateBreakpointsFromUI()
{
  //Validate our parameters
#pragma warning(suppress: 26429)
  CPyDebugFrameWnd* pMainFrame{GetMainFrameWnd()};
  ASSERT(pMainFrame != nullptr);

#pragma warning(suppress: 26429)
  auto* pView{GetView()};
  ASSERT(pView != nullptr);
  auto& rCtrl{pView->GetCtrl()};
  CString sThisModule{GetPathName()};

  //Update all the breakpoints in-memory from the markers in the UI
  {
    CSingleLock lock{&pMainFrame->m_csBreakpoints, TRUE};

    //Delete all breakpoints for this document
#pragma warning(suppress: 26489)
    pMainFrame->m_Breakpoints.erase(std::remove_if(pMainFrame->m_Breakpoints.begin(), pMainFrame->m_Breakpoints.end(), [&sThisModule](const CPyDebugBreakpoint& element) noexcept
      {
        return element.m_sModule == sThisModule;
      }), pMainFrame->m_Breakpoints.end());

    //Now re-add the breakpoints from the markers in the UI
    auto foundLine{rCtrl.MarkerNext(0, 0x3)};
    while (foundLine >= 0)
    {
      //Add to the in-memory
      CPyDebugBreakpoint breakpoint;
      breakpoint.m_LineNumber = foundLine + 1;
      breakpoint.m_sModule = sThisModule;
      const int nMarkers{rCtrl.MarkerGet(foundLine)};
      if (nMarkers & 0x1)
      {
        const CPyDebugFrameWnd::DEBUG_STATE eDebugState{pMainFrame->m_DebugState};
        if ((eDebugState != CPyDebugFrameWnd::DEBUG_STATE::NOT_DEBUGGING) && (eDebugState != CPyDebugFrameWnd::DEBUG_STATE::DEBUG_THREAD_EXITED))
        {
          breakpoint.m_bEnabled = false;
          //Update the marker in the UI
          rCtrl.MarkerDelete(foundLine, SC_PYTHON_MARK_ENABLED_BREAKPOINT);
          rCtrl.MarkerAdd(foundLine, SC_PYTHON_MARK_DISABLED_BREAKPOINT);
        }
        else
        {
          breakpoint.m_bEnabled = IsModified() ? false : true;
          //Update the marker in the UI
          rCtrl.MarkerDelete(foundLine, SC_PYTHON_MARK_ENABLED_BREAKPOINT);
          rCtrl.MarkerAdd(foundLine, breakpoint.m_bEnabled ? SC_PYTHON_MARK_ENABLED_BREAKPOINT : SC_PYTHON_MARK_DISABLED_BREAKPOINT);
        }
      }
      else
      {
        ASSERT(nMarkers & 0x2);
        breakpoint.m_bEnabled = false;
      }
#pragma warning(suppress: 26489)
      pMainFrame->m_Breakpoints.push_back(breakpoint);

      //Prepare for the next time around
      foundLine = rCtrl.MarkerNext(foundLine + 1, 0x3);
    }
  }

  //Update the breakpoints window
  pMainFrame->UpdateBreakpointsWindow();
}

void CPyDebugDoc::DisableAllBreakpoints()
{
  //Validate our parameters
#pragma warning(suppress: 26429)
  auto* pView{GetView()};
  ASSERT(pView != nullptr);
  auto& rCtrl{pView->GetCtrl()};

  //Show all the markers as disabled in the UI
  auto foundLine{rCtrl.MarkerNext(0, 0x3)};
  while (foundLine >= 0)
  {
    rCtrl.MarkerDelete(foundLine, SC_PYTHON_MARK_ENABLED_BREAKPOINT);
    rCtrl.MarkerAdd(foundLine, SC_PYTHON_MARK_DISABLED_BREAKPOINT);
    foundLine = rCtrl.MarkerNext(foundLine + 1, 0x3);
  }
}

void CPyDebugDoc::DeleteAllBreakpoints()
{
  //Validate our parameters
#pragma warning(suppress: 26429)
  auto* pView{GetView()};
  ASSERT(pView != nullptr);
  auto& rCtrl{pView->GetCtrl()};

  rCtrl.MarkerDeleteAll(SC_PYTHON_MARK_ENABLED_BREAKPOINT);
  rCtrl.MarkerDeleteAll(SC_PYTHON_MARK_DISABLED_BREAKPOINT);
}

void CPyDebugDoc::NavigateToLine(_In_ Scintilla::Line lineNumber)
{
  //Validate our parameters
#pragma warning(suppress: 26429)
  auto* pView{GetView()};
  ASSERT(pView != nullptr);
  auto& rCtrl{pView->GetCtrl()};

  //Navigate to the breakpoint in the UI
  rCtrl.SetFocus();
  rCtrl.GotoLine(lineNumber);
}

void CPyDebugDoc::UpdateBreakpoint(_In_ const CPyDebugBreakpoint& breakpoint)
{
  //Validate our parameters
#pragma warning(suppress: 26429)
  auto* pView{GetView()};
  ASSERT(pView != nullptr);
  auto& rCtrl{pView->GetCtrl()};

  //Remove the current marker(s)
  DeleteBreakpoint(breakpoint);

  //Readd the new marker
  if (breakpoint.m_bEnabled)
    rCtrl.MarkerAdd(breakpoint.m_LineNumber - 1, SC_PYTHON_MARK_ENABLED_BREAKPOINT);
  else
    rCtrl.MarkerAdd(breakpoint.m_LineNumber - 1, SC_PYTHON_MARK_DISABLED_BREAKPOINT);
}

void CPyDebugDoc::DeleteBreakpoint(_In_ const CPyDebugBreakpoint& breakpoint)
{
  //Validate our parameters
#pragma warning(suppress: 26429)
  auto* pView{GetView()};
  ASSERT(pView != nullptr);
  auto& rCtrl{pView->GetCtrl()};

  //Remove the current marker(s)
  rCtrl.MarkerDelete(breakpoint.m_LineNumber - 1, SC_PYTHON_MARK_ENABLED_BREAKPOINT);
  rCtrl.MarkerDelete(breakpoint.m_LineNumber - 1, SC_PYTHON_MARK_DISABLED_BREAKPOINT);
}

void CPyDebugDoc::DeleteContents()
{
  //Validate our parameters
  auto* pView{GetView()};
  if (pView == nullptr)
    return;
  auto& rCtrl{pView->GetCtrl()};

  //Disable modification notifications while DeleteContents is being called
  rCtrl.SetModEventMask(Scintilla::ModificationFlags::None);
  __super::DeleteContents();
  rCtrl.SetModEventMask(Scintilla::ModificationFlags::InsertText | Scintilla::ModificationFlags::DeleteText);
}

BOOL CPyDebugDoc::OnOpenDocument(_In_z_ LPCTSTR lpszPathName)
{
  //Validate our parameters
#pragma warning(suppress: 26429)
  auto* pView{GetView()};
  ASSERT(pView != nullptr);
  auto& rCtrl{pView->GetCtrl()};

  //Disable modification notifications while OnOpenDocument is being called
  rCtrl.SetModEventMask(Scintilla::ModificationFlags::None);

  //Let the base class do its thing
  const BOOL bLoaded{__super::OnOpenDocument(lpszPathName)};

  //Reset the modification notifications
  rCtrl.SetModEventMask(Scintilla::ModificationFlags::InsertText | Scintilla::ModificationFlags::DeleteText);

  if (!bLoaded)
    return FALSE;

  //Add any breakpoint markers back to the UI
  UpdateBreakpointsToUI(lpszPathName);

  return TRUE;
}

CPyDebugFrameWnd* CPyDebugDoc::GetMainFrameWnd()
{
  //Validate our parameters
#pragma warning(suppress: 26429)
  const auto* pView{GetView()};
  ASSERT(pView != nullptr);
#pragma warning(suppress: 26429)
  const CFrameWnd* pParentFrame{pView->GetParentFrame()};
  ASSERT(pParentFrame != nullptr);

#pragma warning(suppress: 26466)
  auto pMainFrame{static_cast<CPyDebugFrameWnd*>(pParentFrame->GetParentFrame())};
  ASSERT(pMainFrame != nullptr);
  ASSERT(pMainFrame->IsKindOf(RUNTIME_CLASS(CPyDebugFrameWnd)));

  return pMainFrame;
}
