/*
Module : PyDebugMainFrm.cpp
Purpose: Defines the implementation for the classes which encapsulate the main frame window functionality of PyDebug++
Created: PJN / 20-08-2015
History: PJN / 20-08-2015 1. Initial public release.
         PJN / 22-08-2015 1. Move additional string literals for the various window column names and the attributes shown 
                          in the properties window to string resources to aid in translation
                          2. Updated the code to handle the menu text and description text for the ID_PYDEBUG_WINDOW_MANAGER
                          menu item.
                          3. Verified the code compiles cleanly using VC 2015.
                          4. Removed the global function "GetPyDebugFrameWnd" and modified all code which called this
                          method to use different logic to get the main CPyDebugFrameWnd instance at runtime.
                          5. Replaced all occurrences of NULL with nullptr throughout codebase
         PJN / 19-12-2015 1. Refactored the setup of the various docking windows into separate virtual methods to allow for
                          easier customization in client applications.
                          2. File->New and File->Open menu functionality is now implemented with PyDebug++ specific
                          menu ids. This allows for easier integration in client applications.
                          3. Refactored the functionality of the "DebugThread" method into smaller new virtual methods of
                          "SetupArgvEx", "SetupRedirectIO" & "SetupCodeObject".
                          4. Introduced a new "m_bSupportPySys_SetArgvEx" boolean member variable which decides if
                          PySys_SetArgvEx is called just prior to running/debugging the Python code.
                          5. Sample app now ships with Python v3.5.1 binaries
         PJN / 24-07-2016 1. Updated copyright details.
                          2. Sample app now ships with Python v3.5.2 binaries
                          3. CPyDebugFrameWnd::PyStrToCString method has been updated to avoid unnecessary code which 
                          null terminates the return value string.
         PJN / 27-08-2016 1. Added support for showing expandable Python variables in the Watch, Locals, Built-Ins & 
                          Globals windows.
                          2. Numerous improvements to the handling of keyboard and mouse interaction in the Watch, 
                          Locals, Built-Ins & Globals windows.
         PJN / 03-04-2017 1. Updated copyright details.
                          2. Updated the code to work with the latest version of the Scintilla wrapper classes.
                          3. Updated the code to compile cleanly on VC 2017.
         PJN / 05-11-2017 1. Sample app now ships with Python v3.6.3 binaries
                          2. Reworked the Python initialization to occur in the main UI thread of PyDebugPP instead of
                          in CPyDebugFrameWnd::_DebugThread. Also reworked the code to correctly lock the Python GIL
                          throughout the codebase using a new CPyGILAutoLock class available in PythonGILAutoLock.h
                          3. Fixed some compiler warnings in CPyDebugOutputWnd::OnContextMenu, 
                          CPyDebugFrameWnd::OnUpdateDebugNewBreakpoint & CPyDebugFrameWnd::OnDebugToggleBreakpoint 
                          when compiled for x64 mode.
                          4. Replaced CString::operator LPC*STR() calls with CString::GetString calls
         PJN / 11-11-2017 1. Reworked CPyDebugFrameWnd::PyStrToCString to convert an arbitrary Python object to a 
                          CString.
                          2. Renamed the "PyStrToCString" method to "PyToCString"
                          3. Reworked CPyDebugFrameWnd::GetPythonErrorString to print detailed traceback details if it
                          can.
         PJN / 18-11-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that the code
                          will now only compile on VC 2017 or later.
                          3. Replaced enum throughout the codebase with enum class
         PJN / 19-04-2019 1. Updated copyright details
                          2. Updated the code to clean compile on VC 2019
                          3. Sample app now ships with Python v3.7.3 binaries
         PJN / 11-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 09-11-2019 1. Updated initialization of various structs to use C++ 11 list initialization
         PJN / 28-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 12-02-2020 1. Updated copyright details.
                          2. Fixed an issue where you would get an exception of "I/O operation on closed file" if you 
                          debug with the Allocate Console setting to true and then debug again with the Allocate 
                          Console setting set to false. Thanks to Changyun Kim for reporting this issue.
         PJN / 18-06-2020 1. Added code to the CPyDebugFrameWnd::DebugCode method to disable the close button
                          in the allocated console window. This prevents the main app from closing when you click the 
                          close button in the console window. Thanks to Changyun Kim for reporting this issue.
         PJN / 08-06-2021 1. Updated copyright details
                          2. Updated the code to work with the latest version of the Scintilla wrapper classes.
                          3. Update the code in PyDebugPP.cpp to use the Python Initialization Configuration and 
                          Finalization API's introduced in Python v3.8.
                          4. Sample app now ships with Python v3.9.5 binaries
         PJN / 01-08-2022 1. Updated all the code in to use C++ uniform initialization for all variable
                          declarations.
         PJN / 23-01-2023 1. Update the code to work with Python v3.11.1.
                          2. Updated copyright details.
                          3. Removed CPyDebugFrameWnd::SetupArgvEx method.

Copyright (c) 2015 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#ifndef _ALGORITHM_
#pragma message("To avoid this message, please put algorithm in your pre compiled header (normally stdafx.h)")
#include <algorithm>
#endif //#ifndef _ALGORITHM_
#include "PyDebugMainFrm.h"
#include "resource.h"
#include "PyDebugDoc.h"
#include "PyDebugView.h"
#include "PythonGILAutoLock.h"


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG

constexpr const char* g_CAPSULE_IDENTIFIER = "PYDEBUG_THIS";
constexpr int WM_HANDLE_PENDING_OUTPUT_TEXT = WM_USER + 1;
constexpr int WM_PYTHON_DEBUG_NOTIFICATION = WM_USER + 2;
constexpr int DEBUG_NOTIFCATION_UPDATE_INSTRUCTION_POINTER = 0;
constexpr int DEBUG_NOTIFCATION_UPDATE_WATCH_WINDOW = 1;
constexpr int DEBUG_NOTIFCATION_UPDATE_WATCH_WINDOW2 = 2;

std::array<UINT, 1> g_indicators{ ID_SEPARATOR }; //status line indicator


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26409 26433 26440 26477)
IMPLEMENT_DYNCREATE(CPyDebugFrameWnd, CMDIFrameWndEx)

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CPyDebugFrameWnd, CMDIFrameWndEx) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_CREATE()
  ON_COMMAND(ID_PYDEBUG_WINDOW_MANAGER, &CPyDebugFrameWnd::OnWindowManager)
  ON_COMMAND(ID_PYDEBUG_VIEW_CUSTOMIZE, &CPyDebugFrameWnd::OnViewCustomize)
  ON_REGISTERED_MESSAGE(AFX_WM_CREATETOOLBAR, &CPyDebugFrameWnd::OnToolbarCreateNew)
  ON_MESSAGE(WM_HANDLE_PENDING_OUTPUT_TEXT, &CPyDebugFrameWnd::HandlePendingOutputText)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_GLOBALS, &CMDIFrameWndEx::OnUpdatePaneMenu)
  ON_COMMAND_EX(ID_PYDEBUG_GLOBALS, &CPyDebugFrameWnd::OnPaneCheck)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_LOCALS, &CMDIFrameWndEx::OnUpdatePaneMenu)
  ON_COMMAND_EX(ID_PYDEBUG_LOCALS, &CPyDebugFrameWnd::OnPaneCheck)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_BUILTINS, &CMDIFrameWndEx::OnUpdatePaneMenu)
  ON_COMMAND_EX(ID_PYDEBUG_BUILTINS, &CPyDebugFrameWnd::OnPaneCheck)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_OUTPUT, &CMDIFrameWndEx::OnUpdatePaneMenu)
  ON_COMMAND_EX(ID_PYDEBUG_OUTPUT, &CPyDebugFrameWnd::OnPaneCheck)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_CALLSTACK, &CMDIFrameWndEx::OnUpdatePaneMenu)
  ON_COMMAND_EX(ID_PYDEBUG_CALLSTACK, &CPyDebugFrameWnd::OnPaneCheck)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_MODULES, &CMDIFrameWndEx::OnUpdatePaneMenu)
  ON_COMMAND_EX(ID_PYDEBUG_MODULES, &CPyDebugFrameWnd::OnPaneCheck)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_BREAKPOINTS, &CMDIFrameWndEx::OnUpdatePaneMenu)
  ON_COMMAND_EX(ID_PYDEBUG_BREAKPOINTS, &CPyDebugFrameWnd::OnPaneCheck)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_THREADS, &CMDIFrameWndEx::OnUpdatePaneMenu)
  ON_COMMAND_EX(ID_PYDEBUG_THREADS, &CPyDebugFrameWnd::OnPaneCheck)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_WATCH, &CMDIFrameWndEx::OnUpdatePaneMenu)
  ON_COMMAND_EX(ID_PYDEBUG_WATCH, &CPyDebugFrameWnd::OnPaneCheck)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_PROPERTIES, &CMDIFrameWndEx::OnUpdatePaneMenu)
  ON_COMMAND_EX(ID_PYDEBUG_PROPERTIES, &CPyDebugFrameWnd::OnPaneCheck)
  ON_COMMAND(ID_PYDEBUG_START, &CPyDebugFrameWnd::OnDebugStart)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_START, &CPyDebugFrameWnd::OnUpdateDebugStart)
  ON_COMMAND(ID_PYDEBUG_STOP, &CPyDebugFrameWnd::OnDebugStop)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_STOP, &CPyDebugFrameWnd::OnUpdateDebugStop)
  ON_COMMAND(ID_PYDEBUG_TOGGLEBREAKPOINT, &CPyDebugFrameWnd::OnDebugToggleBreakpoint)
  ON_COMMAND(ID_PYDEBUG_DELETEALLBREAKPOINTS, &CPyDebugFrameWnd::OnDebugDeleteAllBreakpoints)
  ON_COMMAND(ID_PYDEBUG_NEWBREAKPOINT, &CPyDebugFrameWnd::OnDebugNewBreakpoint)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_NEWBREAKPOINT, &CPyDebugFrameWnd::OnUpdateDebugNewBreakpoint)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_TOGGLEBREAKPOINT, &CPyDebugFrameWnd::OnUpdateDebugToggleBreakpoint)
  ON_COMMAND(ID_PYDEBUG_DISABLEALLBREAKPOINTS, &CPyDebugFrameWnd::OnDebugDisableAllBreakpoints)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_BREAKALL, &CPyDebugFrameWnd::OnUpdateDebugBreakAll)
  ON_COMMAND(ID_PYDEBUG_BREAKALL, &CPyDebugFrameWnd::OnDebugBreakAll)
  ON_COMMAND(ID_PYDEBUG_RESTART, &CPyDebugFrameWnd::OnDebugRestart)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_RESTART, &CPyDebugFrameWnd::OnUpdateDebugRestart)
  ON_COMMAND(ID_PYDEBUG_STEPINTO, &CPyDebugFrameWnd::OnDebugStepInto)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_STEPINTO, &CPyDebugFrameWnd::OnUpdateDebugStepInto)
  ON_COMMAND(ID_PYDEBUG_STEPOVER, &CPyDebugFrameWnd::OnDebugStepOver)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_STEPOVER, &CPyDebugFrameWnd::OnUpdateDebugStepOver)
  ON_COMMAND(ID_PYDEBUG_STEPOUT, &CPyDebugFrameWnd::OnDebugStepOut)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_STEPOUT, &CPyDebugFrameWnd::OnUpdateDebugStepOut)
  ON_MESSAGE(WM_PYTHON_DEBUG_NOTIFICATION, &CPyDebugFrameWnd::OnDebugNotification)
  ON_WM_SYSKEYDOWN()
  ON_COMMAND(ID_PYDEBUG_FILE_OPEN, &CPyDebugFrameWnd::OnFileOpen)
  ON_COMMAND(ID_PYDEBUG_FILE_NEW, &CPyDebugFrameWnd::OnFileNew)
END_MESSAGE_MAP()

CPyDebugFrameWnd::~CPyDebugFrameWnd()
{
  //Stop down debugging if required
  if (m_pDebugThread != nullptr)
#pragma warning(suppress: 26447)
    OnDebugStop();
}

#pragma warning(suppress: 26434)
int CPyDebugFrameWnd::OnCreate(_In_ LPCREATESTRUCT lpCreateStruct)
{
  //Let the base class do its thing
  if (__super::OnCreate(lpCreateStruct) == -1)
    return -1;

  //Create the menu bar
  if (!m_wndMenuBar.Create(this))
  {
    TRACE(_T("Failed to create menubar\n"));
    return -1;
  }
  m_wndMenuBar.SetPaneStyle(m_wndMenuBar.GetPaneStyle() | CBRS_SIZE_DYNAMIC | CBRS_TOOLTIPS | CBRS_FLYBY);

  //prevent the menu bar from taking the focus on activation
  CMFCPopupMenu::SetForceMenuFocus(FALSE);

  //Create the standard toolbar
  if (!m_wndStandardToolBar.CreateEx(this, TBSTYLE_FLAT, WS_CHILD | WS_VISIBLE | CBRS_TOP | CBRS_GRIPPER | CBRS_TOOLTIPS | CBRS_FLYBY | CBRS_SIZE_DYNAMIC, CRect{1, 1, 1, 1}, AFX_IDW_TOOLBAR) ||
      !m_wndStandardToolBar.LoadToolBar(m_bHiColorIcons ? IDR_PYDEBUG_MAINFRAME_HC : IDR_PYDEBUG_MAINFRAME))
  {
    TRACE(_T("Failed to create standard toolbar\n"));
    return -1;
  }
  CString strToolBarName;
  if (!strToolBarName.LoadString(IDS_PYDEBUG_TOOLBAR_STANDARD))
  {
    TRACE(_T("Failed to load standard toolbar caption text\n"));
    return -1;
  }
  m_wndStandardToolBar.SetWindowText(strToolBarName);
  CString strCustomize;
  if (!strCustomize.LoadString(IDS_PYDEBUG_TOOLBAR_CUSTOMIZE))
  {
    TRACE(_T("Failed to load customize toolbar text\n"));
    return -1;
  }
  m_wndStandardToolBar.EnableCustomizeButton(TRUE, ID_PYDEBUG_VIEW_CUSTOMIZE, strCustomize);

  //Create the debug toolbar
  if (!m_wndDebugToolBar.CreateEx(this, TBSTYLE_FLAT, WS_CHILD | WS_VISIBLE | CBRS_TOP | CBRS_GRIPPER | CBRS_TOOLTIPS | CBRS_FLYBY | CBRS_SIZE_DYNAMIC, CRect{1, 1, 1, 1}, IDW_PYDEBUG_DEBUG_TOOLBAR) ||
      !m_wndDebugToolBar.LoadToolBar(m_bHiColorIcons ? IDR_PYDEBUG_DEBUG_HC : IDR_PYDEBUG_DEBUG))
  {
    TRACE(_T("Failed to create debug toolbar\n"));
    return -1;
  }
  if (!strToolBarName.LoadString(IDS_PYDEBUG_TOOLBAR_DEBUG))
  {
    TRACE(_T("Failed to load debug toolbar caption text\n"));
    return -1;
  }
  m_wndDebugToolBar.SetWindowText(strToolBarName);
  m_wndDebugToolBar.EnableCustomizeButton(TRUE, ID_PYDEBUG_VIEW_CUSTOMIZE, strCustomize);

  //Create the status bar
  if (!m_wndStatusBar.Create(this))
  {
    TRACE(_T("Failed to create status bar\n"));
    return -1;
  }
#pragma warning(suppress: 26472)
  m_wndStatusBar.SetIndicators(g_indicators.data(), static_cast<int>(g_indicators.size()));

  m_wndMenuBar.EnableDocking(CBRS_ALIGN_ANY);
  m_wndStandardToolBar.EnableDocking(CBRS_ALIGN_ANY);
  m_wndDebugToolBar.EnableDocking(CBRS_ALIGN_ANY);
  EnableDocking(CBRS_ALIGN_ANY);
  DockPane(&m_wndMenuBar);
  DockPane(&m_wndDebugToolBar);
  DockPaneLeftOf(&m_wndStandardToolBar, &m_wndDebugToolBar);

  CDockingManager::SetDockingMode(DT_SMART);
  EnableAutoHidePanes(CBRS_ALIGN_ANY);
  EnableDocking(CBRS_ALIGN_TOP | CBRS_ALIGN_BOTTOM | CBRS_ALIGN_RIGHT | CBRS_ALIGN_LEFT);
  EnableAutoHidePanes(CBRS_ALIGN_RIGHT);

  //Load menu item image (not placed on any standard toolbars)
  CMFCToolBar::AddToolBarForImageCollection(IDR_PYDEBUG_MENU_IMAGES, m_bHiColorIcons ? IDB_PYDEBUG_MENU_IMAGES_HC : 0);

  //create docking windows
  if (!CreateDockingWindows())
  {
    TRACE(_T("Failed to create docking windows\n"));
    return -1;
  }

  //set the visual manager used to draw all user interface elements
  CMFCVisualManager::SetDefaultManager(RUNTIME_CLASS(CMFCVisualManagerVS2008));

  //Enable enhanced windows management dialog
  EnableWindowsDialog(ID_PYDEBUG_WINDOW_MANAGER, ID_PYDEBUG_WINDOW_MANAGER, TRUE);

  //Enable toolbar and docking window menu replacement
  EnablePaneMenu(TRUE, ID_PYDEBUG_VIEW_CUSTOMIZE, strCustomize, ID_VIEW_TOOLBAR);

  //enable quick (Alt+drag) toolbar customization
  CMFCToolBar::EnableQuickCustomization();

  return 0;
}

BOOL CPyDebugFrameWnd::CreateOutputDockingWindow()
{
  //Create output window
  CString sCaption;
  if (!sCaption.LoadString(IDS_PYDEBUG_OUTPUT_WND))
    return FALSE;
  if (!m_wndOutput.Create(sCaption, this, CRect{0, 0, 100, 300}, TRUE, ID_PYDEBUG_OUTPUT, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CBRS_BOTTOM | CBRS_FLOAT_MULTI))
    return FALSE;

  //Set the icon 
  const int nCXIcon = ::GetSystemMetrics(SM_CXSMICON);
  const int nCYIcon = ::GetSystemMetrics(SM_CYSMICON);
  auto hOutputIcon = static_cast<HICON>(::LoadImage(::AfxGetResourceHandle(), MAKEINTRESOURCE(IDI_PYDEBUG_OUTPUT_WND), IMAGE_ICON, nCXIcon, nCYIcon, 0));
  m_wndOutput.SetIcon(hOutputIcon, FALSE);

  //Setup the docking options
  m_wndOutput.EnableDocking(CBRS_ALIGN_ANY);
  DockPane(&m_wndOutput);

  return TRUE;
}

BOOL CPyDebugFrameWnd::CreatePropertiesDockingWindow()
{
  //Create properties window
  m_wndProperties.SetPyDebugMainFrame(this);
  CString sCaption;
  if (!sCaption.LoadString(IDS_PYDEBUG_PROPERTIES_WND))
    return FALSE;
  if (!m_wndProperties.Create(sCaption, this, CRect{0, 0, 400, 200}, TRUE, ID_PYDEBUG_PROPERTIES, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CBRS_RIGHT | CBRS_FLOAT_MULTI))
    return FALSE;

  //Set the icon
  const int nCXIcon = ::GetSystemMetrics(SM_CXSMICON);
  const int nCYIcon = ::GetSystemMetrics(SM_CYSMICON);
  auto hPropertiesBarIcon = static_cast<HICON>(::LoadImage(::AfxGetResourceHandle(), MAKEINTRESOURCE(IDI_PYDEBUG_PROPERTIES_WND), IMAGE_ICON, nCXIcon, nCYIcon, 0));
  m_wndProperties.SetIcon(hPropertiesBarIcon, FALSE);

  //Setup the docking options
  m_wndProperties.EnableDocking(CBRS_ALIGN_ANY);
  DockPane(&m_wndProperties);

  return TRUE;
}

BOOL CPyDebugFrameWnd::CreateGlobalsDockingWindow()
{
  //Create globals watch window
  CString sCaption;
  if (!sCaption.LoadString(IDS_PYDEBUG_GLOBALS_WND))
    return FALSE;
  m_wndGlobals.SetPyDebugMainFrame(this);
  m_wndGlobals.SetEditable(false);
  m_wndGlobals.m_wndCtrl.SetType(CPyDebugVariablesListCtrl::TYPE::GLOBALS);
  if (!m_wndGlobals.Create(sCaption, this, CRect{0, 0, 100, 300}, TRUE, ID_PYDEBUG_GLOBALS, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CBRS_BOTTOM | CBRS_FLOAT_MULTI))
    return FALSE;

  //Set the icon
  const int nCXIcon = ::GetSystemMetrics(SM_CXSMICON);
  const int nCYIcon = ::GetSystemMetrics(SM_CYSMICON);
  auto hGlobalsIcon = static_cast<HICON>(::LoadImage(::AfxGetResourceHandle(), MAKEINTRESOURCE(IDI_PYDEBUG_GLOBALS_WND), IMAGE_ICON, nCXIcon, nCYIcon, 0));
  m_wndGlobals.SetIcon(hGlobalsIcon, FALSE);

  //Setup the docking options
  m_wndGlobals.EnableDocking(CBRS_ALIGN_ANY);
  m_wndGlobals.AttachToTabWnd(&m_wndOutput, DM_STANDARD);

  return TRUE;
}

BOOL CPyDebugFrameWnd::CreateLocalsDockingWindow()
{
  //Create locals watch window
  CString sCaption;
  if (!sCaption.LoadString(IDS_PYDEBUG_LOCALS_WND))
    return FALSE;
  m_wndLocals.SetPyDebugMainFrame(this);
  m_wndLocals.SetEditable(false);
  m_wndLocals.m_wndCtrl.SetType(CPyDebugVariablesListCtrl::TYPE::LOCALS);
  if (!m_wndLocals.Create(sCaption, this, CRect{0, 0, 100, 300}, TRUE, ID_PYDEBUG_LOCALS, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CBRS_BOTTOM | CBRS_FLOAT_MULTI))
    return FALSE;

  //Set the icon
  const int nCXIcon = ::GetSystemMetrics(SM_CXSMICON);
  const int nCYIcon = ::GetSystemMetrics(SM_CYSMICON);
  auto hLocalsIcon = static_cast<HICON>(::LoadImage(::AfxGetResourceHandle(), MAKEINTRESOURCE(IDI_PYDEBUG_LOCALS_WND), IMAGE_ICON, nCXIcon, nCYIcon, 0));
  m_wndLocals.SetIcon(hLocalsIcon, FALSE);

  //Setup the docking options
  m_wndLocals.EnableDocking(CBRS_ALIGN_ANY);
  m_wndLocals.AttachToTabWnd(&m_wndOutput, DM_STANDARD);

  return TRUE;
}

BOOL CPyDebugFrameWnd::CreateBuiltInsDockingWindow()
{
  //Create BuiltIns watch window
  CString sCaption;
  if (!sCaption.LoadString(IDS_PYDEBUG_BUILTINS_WND))
    return FALSE;
  m_wndBuiltIns.SetPyDebugMainFrame(this);
  m_wndBuiltIns.SetEditable(false);
  m_wndBuiltIns.m_wndCtrl.SetType(CPyDebugVariablesListCtrl::TYPE::BUILTINS);
  if (!m_wndBuiltIns.Create(sCaption, this, CRect{0, 0, 100, 300}, TRUE, ID_PYDEBUG_BUILTINS, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CBRS_BOTTOM | CBRS_FLOAT_MULTI))
    return FALSE;

  //Set the icon
  const int nCXIcon = ::GetSystemMetrics(SM_CXSMICON);
  const int nCYIcon = ::GetSystemMetrics(SM_CYSMICON);
  auto hBuiltInsIcon = static_cast<HICON>(::LoadImage(::AfxGetResourceHandle(), MAKEINTRESOURCE(IDI_PYDEBUG_BUILTINS_WND), IMAGE_ICON, nCXIcon, nCYIcon, 0));
  m_wndBuiltIns.SetIcon(hBuiltInsIcon, FALSE);

  //Setup the docking options
  m_wndBuiltIns.EnableDocking(CBRS_ALIGN_ANY);
  m_wndBuiltIns.AttachToTabWnd(&m_wndOutput, DM_STANDARD);

  return TRUE;
}

BOOL CPyDebugFrameWnd::CreateCallStackDockingWindow()
{
  //Create Call Stack window
  m_wndCallStack.SePyDebugMainFrame(this);
  CString sCaption;
  if (!sCaption.LoadString(IDS_PYDEBUG_CALLSTACK_WND))
    return FALSE;
  if (!m_wndCallStack.Create(sCaption, this, CRect{0, 0, 100, 300}, TRUE, ID_PYDEBUG_CALLSTACK, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CBRS_BOTTOM | CBRS_FLOAT_MULTI))
    return FALSE;

  //Set the icon
  const int nCXIcon = ::GetSystemMetrics(SM_CXSMICON);
  const int nCYIcon = ::GetSystemMetrics(SM_CYSMICON);
  auto hCallStackIcon = static_cast<HICON>(::LoadImage(::AfxGetResourceHandle(), MAKEINTRESOURCE(IDI_PYDEBUG_CALLSTACK_WND), IMAGE_ICON, nCXIcon, nCYIcon, 0));
  m_wndCallStack.SetIcon(hCallStackIcon, FALSE);

  //Setup the docking options
  m_wndCallStack.EnableDocking(CBRS_ALIGN_ANY);
  m_wndCallStack.AttachToTabWnd(&m_wndOutput, DM_STANDARD);

  return TRUE;
}

BOOL CPyDebugFrameWnd::CreateModulesDockingWindow()
{
  //Create Modules window
  m_wndModules.SetPyDebugMainFrame(this);
  CString sCaption;
  if (!sCaption.LoadString(IDS_PYDEBUG_MODULES_WND))
    return FALSE;
  if (!m_wndModules.Create(sCaption, this, CRect{0, 0, 100, 300}, TRUE, ID_PYDEBUG_MODULES, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CBRS_BOTTOM | CBRS_FLOAT_MULTI))
    return FALSE;

  //Set the icon
  const int nCXIcon = ::GetSystemMetrics(SM_CXSMICON);
  const int nCYIcon = ::GetSystemMetrics(SM_CYSMICON);
  auto hModulesIcon = static_cast<HICON>(::LoadImage(::AfxGetResourceHandle(), MAKEINTRESOURCE(IDI_PYDEBUG_MODULES_WND), IMAGE_ICON, nCXIcon, nCYIcon, 0));
  m_wndModules.SetIcon(hModulesIcon, FALSE);

  //Setup the docking options
  m_wndModules.EnableDocking(CBRS_ALIGN_ANY);
  m_wndModules.AttachToTabWnd(&m_wndOutput, DM_STANDARD);

  return TRUE;
}

BOOL CPyDebugFrameWnd::CreateBreakpointsDockingWindow()
{
  //Create Breakpoints window
  m_wndBreakpoints.SePyDebugMainFrame(this);
  CString sCaption;
  if (!sCaption.LoadString(IDS_PYDEBUG_BREAKPOINTS_WND))
    return FALSE;
  if (!m_wndBreakpoints.Create(sCaption, this, CRect{0, 0, 100, 300}, TRUE, ID_PYDEBUG_BREAKPOINTS, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CBRS_BOTTOM | CBRS_FLOAT_MULTI))
    return FALSE;

  //Set the icon
  const int nCXIcon = ::GetSystemMetrics(SM_CXSMICON);
  const int nCYIcon = ::GetSystemMetrics(SM_CYSMICON);
  auto hBreakpointsIcon = static_cast<HICON>(::LoadImage(::AfxGetResourceHandle(), MAKEINTRESOURCE(IDI_PYDEBUG_BREAKPOINTS_WND), IMAGE_ICON, nCXIcon, nCYIcon, 0));
  m_wndBreakpoints.SetIcon(hBreakpointsIcon, FALSE);

  //Setup the docking options
  m_wndBreakpoints.EnableDocking(CBRS_ALIGN_ANY);
  m_wndBreakpoints.AttachToTabWnd(&m_wndOutput, DM_STANDARD);

  return TRUE;
}

BOOL CPyDebugFrameWnd::CreateThreadsDockingWindow()
{
  //Create Threads window
  CString sCaption;
  if (!sCaption.LoadString(IDS_PYDEBUG_THREADS))
    return FALSE;
  if (!m_wndThreads.Create(sCaption, this, CRect{0, 0, 100, 300}, TRUE, ID_PYDEBUG_THREADS, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CBRS_BOTTOM | CBRS_FLOAT_MULTI))
    return FALSE;

  //Set the icon
  const int nCXIcon = ::GetSystemMetrics(SM_CXSMICON);
  const int nCYIcon = ::GetSystemMetrics(SM_CYSMICON);
  auto hThreadsIcon = static_cast<HICON>(::LoadImage(::AfxGetResourceHandle(), MAKEINTRESOURCE(IDI_PYDEBUG_THREADS_WND), IMAGE_ICON, nCXIcon, nCYIcon, 0));
  m_wndThreads.SetIcon(hThreadsIcon, FALSE);

  //Setup the docking options
  m_wndThreads.EnableDocking(CBRS_ALIGN_ANY);
  m_wndThreads.AttachToTabWnd(&m_wndOutput, DM_STANDARD);

  return TRUE;
}

BOOL CPyDebugFrameWnd::CreateWatchDockingWindow()
{
  //Create watch window
  m_wndWatch.SetPyDebugMainFrame(this);
  m_wndWatch.SetEditable(true);
  m_wndWatch.m_wndCtrl.SetType(CPyDebugVariablesListCtrl::TYPE::WATCH);
  CString sCaption;
  if (!sCaption.LoadString(IDS_PYDEBUG_WATCH))
    return FALSE;
  if (!m_wndWatch.Create(sCaption, this, CRect{0, 0, 100, 300}, TRUE, ID_PYDEBUG_WATCH, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CBRS_BOTTOM | CBRS_FLOAT_MULTI))
    return FALSE;

  //Set the icon
  const int nCXIcon{::GetSystemMetrics(SM_CXSMICON)};
  const int nCYIcon{::GetSystemMetrics(SM_CYSMICON)};
  auto hWatchIcon{static_cast<HICON>(::LoadImage(::AfxGetResourceHandle(), MAKEINTRESOURCE(IDI_PYDEBUG_WATCH_WND), IMAGE_ICON, nCXIcon, nCYIcon, 0))};
  m_wndWatch.SetIcon(hWatchIcon, FALSE);

  m_wndWatch.m_wndCtrl.UpdateVariableWindow();

  //Setup the docking options
  m_wndWatch.EnableDocking(CBRS_ALIGN_ANY);
  m_wndWatch.AttachToTabWnd(&m_wndOutput, DM_STANDARD);

  return TRUE;
}

BOOL CPyDebugFrameWnd::CreateDockingWindows()
{
  //Delegate the work to the various virtual methods for each docking window
  if (!CreateOutputDockingWindow())
    return FALSE;
  if (!CreatePropertiesDockingWindow())
    return FALSE;
  if (!CreateGlobalsDockingWindow())
    return FALSE;
  if (!CreateLocalsDockingWindow())
    return FALSE;
  if (!CreateBuiltInsDockingWindow())
    return FALSE;
  if (!CreateCallStackDockingWindow())
    return FALSE;
  if (!CreateModulesDockingWindow())
    return FALSE;
  if (!CreateBreakpointsDockingWindow())
    return FALSE;
  if (!CreateThreadsDockingWindow())
    return FALSE;
  if (!CreateWatchDockingWindow())
    return FALSE;
  return TRUE;
}

void CPyDebugFrameWnd::OnViewCustomize()
{
#pragma warning(suppress: 26400 26403 26409)
  auto pDlgCust{new CMFCToolBarsCustomizeDialog(this, TRUE)};
  pDlgCust->Create();
}

#pragma warning(suppress: 26434)
LRESULT CPyDebugFrameWnd::OnToolbarCreateNew(_In_ WPARAM wParam, _In_ LPARAM lParam)
{
  //Let the base class do its thing
  LRESULT lres{__super::OnToolbarCreateNew(wParam, lParam)};
  if (lres == 0)
    return 0;

#pragma warning(suppress: 26429 26490)
  auto pUserToolbar{reinterpret_cast<CMFCToolBar*>(lres)};
#pragma warning(suppress: 26486)
  ASSERT_VALID(pUserToolbar);
  CString strCustomize;
  if (!strCustomize.LoadString(IDS_PYDEBUG_TOOLBAR_CUSTOMIZE))
    return 0;
#pragma warning(suppress: 26486)
  pUserToolbar->EnableCustomizeButton(TRUE, ID_PYDEBUG_VIEW_CUSTOMIZE, strCustomize);
  return lres;
}

void CPyDebugFrameWnd::AddOutputText(_In_ UINT nResourceID)
{
  CString sText;
  if (!sText.LoadString(nResourceID))
    return;

  //Delegate to the other version of the method
  AddOutputText(sText);
}

void CPyDebugFrameWnd::AddOutputTextU(_In_ UINT nResourceID, _In_ UINT nValue)
{
  CString sValue;
  sValue.Format(_T("%u"), nValue);

  //Delegate to the other version of the method
  AddOutputText(nResourceID, sValue);
}

void CPyDebugFrameWnd::AddOutputTextI(_In_ UINT nResourceID, _In_ int nValue)
{
  CString sValue;
  sValue.Format(_T("%d"), nValue);

  //Delegate to the other version of the method
  AddOutputText(nResourceID, sValue);
}


void CPyDebugFrameWnd::AddOutputText(_In_ UINT nResourceID, _In_z_ LPCTSTR pszValue)
{
  CString sText;
  AfxFormatString1(sText, nResourceID, pszValue);

  //Delegate to the other version of the method
  AddOutputText(sText);
}

void CPyDebugFrameWnd::AddOutputText(_In_ UINT nResourceID, _In_z_ LPCTSTR pszValue1, _In_z_ LPCTSTR pszValue2)
{
  std::array<LPCTSTR, 2> rglpsz{pszValue1, pszValue2};
  CString sText;
#pragma warning(suppress: 26472)
  AfxFormatStrings(sText, nResourceID, rglpsz.data(), static_cast<int>(rglpsz.size()));

  //Delegate to the other version of the method
  AddOutputText(sText);
}

void CPyDebugFrameWnd::AddOutputText(_In_ UINT nResourceID, _In_ int nValue1, _In_z_ LPCTSTR pszValue2, _In_z_ LPCTSTR pszValue3)
{
  CString sValue1;
  sValue1.Format(_T("%d"), nValue1);

  std::array<LPCTSTR, 3> rglpsz{sValue1, pszValue2, pszValue3};
  CString sText;
#pragma warning(suppress: 26472)
  AfxFormatStrings(sText, nResourceID, rglpsz.data(), static_cast<int>(rglpsz.size()));

  //Delegate to the other version of the method
  AddOutputText(sText);
}

void CPyDebugFrameWnd::AddOutputText(_In_ UINT nResourceID, _In_z_ LPCTSTR pszValue1, _In_ int nValue2, _In_z_ LPCTSTR pszValue3, _In_z_ LPCTSTR pszValue4)
{
  CString sValue2;
  sValue2.Format(_T("%d"), nValue2);

  std::array<LPCTSTR, 4> rglpsz{pszValue1, sValue2, pszValue3, pszValue4};
  CString sText;
#pragma warning(suppress: 26472)
  AfxFormatStrings(sText, nResourceID, rglpsz.data(), static_cast<int>(rglpsz.size()));

  //Delegate to the other version of the method
  AddOutputText(sText);
}

void CPyDebugFrameWnd::AddOutputText(_In_z_ LPCTSTR pszText)
{
  if (!IsWindow(GetSafeHwnd()))
    return;

  //Add to the pending queue
  {
    CSingleLock sl{&m_csOutputText, TRUE};
    m_pendingOutputText.emplace_back(pszText);
  }

  //Delegate to the UI thread
  PostMessage(WM_HANDLE_PENDING_OUTPUT_TEXT);
}

LRESULT CPyDebugFrameWnd::HandlePendingOutputText(_In_ WPARAM /*wParam*/, _In_ LPARAM /*lParam*/)
{
  CSingleLock lock{&m_csOutputText, TRUE};
  for (const auto& sText : m_pendingOutputText)
    m_wndOutput.AddOutputText(sText);
  m_pendingOutputText.clear();

  return 0L;
}

BOOL CPyDebugFrameWnd::PreTranslateMessage(MSG* pMsg)
{
  //Handle the reserved F10 key
  if ((pMsg->message == WM_SYSKEYDOWN) && (pMsg->wParam == VK_F10))
  {
    SendMessage(WM_COMMAND, ID_PYDEBUG_STEPOVER);
    return TRUE;
  }
  else if (((pMsg->message == WM_SYSKEYUP) || (pMsg->message == WM_KEYDOWN)) && (pMsg->wParam == VK_F10))
    return TRUE;
  else
    return __super::PreTranslateMessage(pMsg);
}

#pragma warning(suppress: 26434)
void CPyDebugFrameWnd::OnSysKeyDown(_In_ UINT nChar, _In_ UINT nRepCnt, _In_ UINT nFlags)
{
  if (nChar != VK_F10)
    __super::OnSysKeyDown(nChar, nRepCnt, nFlags);
}

CPyDebugDoc* CPyDebugFrameWnd::FindActiveDebuggerDocument()
{
  const CFrameWnd* pFrameWnd{GetActiveFrame()};
  if (pFrameWnd != nullptr)
  {
    const CView* pView{pFrameWnd->GetActiveView()};
    if (pView != nullptr)
    {
#pragma warning(suppress: 26429 26466)
      auto pDoc{static_cast<CPyDebugDoc*>(pView->GetDocument())};
      ASSERT(pDoc != nullptr);
      if (pDoc->IsKindOf(RUNTIME_CLASS(CPyDebugDoc)))
        return pDoc;
    }
  }

  return nullptr;
}

void CPyDebugFrameWnd::NavigateToDocumentLine(_In_opt_z_ LPCTSTR pszFilename, _In_ Scintilla::Line lineNumber, _In_ bool bAllowOpen)
{
  //Find the correct document, navigate to the specified location and activate the parent frame
  CPyDebugDoc* pDoc{FindDebuggerDocument(pszFilename, bAllowOpen)};
  if (pDoc != nullptr)
  {
    pDoc->NavigateToLine(lineNumber);
#pragma warning(suppress: 26429)
    const auto* pView{pDoc->GetView()};
    ASSERT(pView != nullptr);
#pragma warning(suppress: 26429)
    CFrameWnd* pActiveFrame{pView->GetParentFrame()};
    ASSERT(pActiveFrame != nullptr);
    pActiveFrame->ActivateFrame();
  }
}

#pragma warning(suppress: 26434)
BOOL CPyDebugFrameWnd::OnPaneCheck(_In_ UINT nID)
{
  ASSERT_VALID(this);

  CBasePane* pBar{GetPane(nID)};
  if (pBar != nullptr)
  {
    ShowPane(pBar, (pBar->GetStyle() & WS_VISIBLE) == 0, FALSE, TRUE);
    return TRUE;
  }

  return FALSE;
}

//Simple helper function to convert a python object to an MFC CString
CString CPyDebugFrameWnd::PyToCString(_In_opt_ PyObject* pObject)
{
  //What will be the return value from this method
  CString sString;

  if (pObject != nullptr)
  {
    CPyObject pStr{PyObject_Str(pObject)};
    if (pStr != nullptr)
    {
      CPyObject pUTF8Str{PyUnicode_AsUTF8String(pStr)};
      if (pUTF8Str != nullptr)
#pragma warning(suppress: 26485)
        sString = ATL::CA2W(PyBytes_AS_STRING(pUTF8Str.operator PyObject*()), CP_UTF8).operator LPWSTR();
    }
  }

  return sString;
}

void CPyDebugFrameWnd::OnDebugStart()
{
  //Implement continue functionality
  DEBUG_STATE desired{DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL};
  if (m_DebugState.compare_exchange_strong(desired, DEBUG_STATE::REQUEST_CONTINUE))
  {
    UpdateUIForCodeContinue();
    return;
  }

  //Delegate the main work to a helper method
  StartDebug(START_DEBUGGING_TYPE::NORMAL_START);
}

void CPyDebugFrameWnd::StartDebug(_In_ START_DEBUGGING_TYPE eType)
{
  //Validate our parameters
#pragma warning(suppress: 26429)
  CPyDebugDoc* pDoc{FindActiveDebuggerDocument()};
  ASSERT(pDoc != nullptr);

  //Implement Start debugging functionality

  //get name/title of document
  CString sModuleName;
  if (pDoc->GetPathName().IsEmpty()) //NOLINT(clang-analyzer-core.CallAndMessage)
  {
    //get name based on caption
    sModuleName = pDoc->GetTitle();
    if (sModuleName.IsEmpty())
#pragma warning(suppress: 26496)
      ENSURE(sModuleName.LoadString(AFX_IDS_UNTITLED));
  }
  else
  {
    //get name based on file title of path name
    sModuleName = pDoc->GetPathName();
  }

  //We can't debug if the document is marked modified
  if (pDoc->IsModified())
  {
    //get name/title of document
    CString sName;
    if (pDoc->GetPathName().IsEmpty())
      sName = sModuleName;
    else
    {
      GetFileTitle(sModuleName, sName.GetBuffer(_MAX_PATH), _MAX_PATH);
      sName.ReleaseBuffer();
    }

    //Prompt the user to save, and save if required before we start debugging
    CString prompt;
    AfxFormatString1(prompt, AFX_IDP_ASK_TO_SAVE, sName);
    switch (AfxMessageBox(prompt, MB_YESNO, AFX_IDP_ASK_TO_SAVE))
    {
      case IDYES:
      {
        if (!pDoc->DoFileSave())
          return;
        break;
      }
      case IDNO:
      {
        return;
        break;
      }
      default:
      {
        ASSERT(FALSE);
        break;
      }
    }
  }

  //Extract the code to execute
#pragma warning(suppress: 26429)
  auto* pView{pDoc->GetView()};
  ASSERT(pView != nullptr);
  auto& rCtrl{pView->GetCtrl()};
#pragma warning(suppress: 26472)
  CStringW sUnicodeCode{rCtrl.GetText(static_cast<int>(rCtrl.GetTextLength() + 1))};

  //Run the code
  DebugCode(sUnicodeCode, sModuleName, eType);
}

BOOL CPyDebugFrameWnd::GetToolbarButtonToolTipText(CMFCToolBarButton* pButton, CString& strTTText)
{
  //Validate our parameters
  ASSERT(pButton != nullptr);

  if ((pButton->m_nID == ID_PYDEBUG_START) && (m_DebugState == DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL))
  {
    const BOOL bReturn{strTTText.LoadString(ID_PYDEBUG_CONTINUE)};
    if (bReturn)
      AfxExtractSubString(strTTText, strTTText, 1);
    return bReturn;
  }
  else if (pButton->m_nID == ID_PYDEBUG_WINDOW_MANAGER)
  {
    const BOOL bReturn{strTTText.LoadString(ID_PYDEBUG_WINDOW_MANAGER_DESCRIPTION)};
    if (bReturn)
      AfxExtractSubString(strTTText, strTTText, 1);
    return bReturn;
  }
  else
    return __super::GetToolbarButtonToolTipText(pButton, strTTText);
}

void CPyDebugFrameWnd::GetMessageString(UINT nID, CString& rMessage) const
{
  if ((nID == ID_PYDEBUG_START) && (m_DebugState == DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL))
  {
#pragma warning(suppress: 26496)
    ENSURE(rMessage.LoadString(ID_PYDEBUG_CONTINUE));
    AfxExtractSubString(rMessage, rMessage, 0);
  }
  else if (nID == ID_PYDEBUG_WINDOW_MANAGER)
  {
#pragma warning(suppress: 26496)
    ENSURE(rMessage.LoadString(ID_PYDEBUG_WINDOW_MANAGER_DESCRIPTION));
    AfxExtractSubString(rMessage, rMessage, 0);
  }
  else
    return __super::GetMessageString(nID, rMessage);
}

#pragma warning(suppress: 26429)
void CPyDebugFrameWnd::OnUpdateDebugStart(_In_ CCmdUI* pCmdUI)
{
  //Get the active document
  const DEBUG_STATE eDebugState{m_DebugState};
  const CPyDebugDoc* pDoc{FindActiveDebuggerDocument()};

  //Change the text for continue
  if (eDebugState == DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL)
  {
    //Change the command text
    CString sText;
#pragma warning(suppress: 26496)
    ENSURE(sText.LoadString(ID_PYDEBUG_CONTINUE));
    AfxExtractSubString(sText, sText, 1);
#pragma warning(suppress: 26486)
    pCmdUI->SetText(sText);
  }
  else
  {
    //Change the command text
    CString sText;
#pragma warning(suppress: 26496)
    ENSURE(sText.LoadString(ID_PYDEBUG_START));
    AfxExtractSubString(sText, sText, 1);
#pragma warning(suppress: 26486)
    pCmdUI->SetText(sText);
  }

  //Handle the case of no document currently open
  if (pDoc == nullptr)
  {
    //Allow continue when no document is currently open
    if (eDebugState == DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL)
#pragma warning(suppress: 26486)
      pCmdUI->Enable(TRUE);
    else
#pragma warning(suppress: 26486)
      pCmdUI->Enable(FALSE);
    return;
  }

  //Allow debugging so long as their is some code and we are not debugging already, the debugging thread has exit or we are currently break all in the debugger
#pragma warning(suppress: 26429)
  auto* pView{pDoc->GetView()};
  ASSERT(pView != nullptr);
#pragma warning(suppress: 26486)
  pCmdUI->Enable((pView->GetCtrl().GetWindowTextLength() != 0) && ((eDebugState == DEBUG_STATE::NOT_DEBUGGING) || (eDebugState == DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL) || (eDebugState == DEBUG_STATE::DEBUG_THREAD_EXITED)));
}

void CPyDebugFrameWnd::RemoveAllInstructionPointers()
{
#pragma warning(suppress: 26496)
  AFXASSUME(m_pDocTemplate != nullptr); //Did you forget to call CPyDebugFrameWnd::SetDocTemplate
  POSITION posDoc{m_pDocTemplate->GetFirstDocPosition()};
  while (posDoc != nullptr)
  {
#pragma warning(suppress: 26429 26466)
    const auto* pDoc{static_cast<CPyDebugDoc*>(m_pDocTemplate->GetNextDoc(posDoc))};
    ASSERT(pDoc != nullptr);
#pragma warning(suppress: 26429)
    auto* pView{pDoc->GetView()};
    ASSERT(pView != nullptr);
    auto& rCtrl{pView->GetCtrl()};

    rCtrl.MarkerDeleteAll(SC_PYTHON_MARK_INSTRUCTION_POINTER);
  }
}

void CPyDebugFrameWnd::OnDebugStop()
{
  if (m_pDebugThread != nullptr)
  {
    //Request the debug thread to exit
    m_DebugState = DEBUG_STATE::REQUEST_STOP;

    //Shutdown the debugger thread
    WaitForSingleObject(m_pDebugThread->m_hThread, INFINITE);
    delete m_pDebugThread;
    m_pDebugThread = nullptr;

    m_DebugState = DEBUG_STATE::NOT_DEBUGGING;
  }

  //Get the active document
  RemoveAllInstructionPointers();

  //Reset the locals window
  {
    if (m_wndLocals.m_wndCtrl.GetSafeHwnd())
    {
      m_wndLocals.m_wndCtrl.DeleteAllItems();
      m_wndLocals.m_wndCtrl.EnableWindow(FALSE);
    }
    CSingleLock sl(&m_wndLocals.m_wndCtrl.m_csVariables, TRUE);
    m_wndLocals.m_wndCtrl.m_Variables.clear();
  }

  //Reset the globals window
  {
    if (m_wndGlobals.m_wndCtrl.GetSafeHwnd())
    {
      m_wndGlobals.m_wndCtrl.DeleteAllItems();
      m_wndGlobals.m_wndCtrl.EnableWindow(FALSE);
    }
    CSingleLock sl(&m_wndGlobals.m_wndCtrl.m_csVariables, TRUE);
    m_wndGlobals.m_wndCtrl.m_Variables.clear();
  }

  //Reset the Built-Ins window
  {
    if (m_wndBuiltIns.m_wndCtrl.GetSafeHwnd())
    {
      m_wndBuiltIns.m_wndCtrl.DeleteAllItems();
      m_wndBuiltIns.m_wndCtrl.EnableWindow(FALSE);
    }
    CSingleLock sl(&m_wndBuiltIns.m_wndCtrl.m_csVariables, TRUE);
    m_wndBuiltIns.m_wndCtrl.m_Variables.clear();
  }

  //Reset the Call Stack window
  {
    if (m_wndCallStack.m_wndCtrl.GetSafeHwnd())
    {
      m_wndCallStack.m_wndCtrl.DeleteAllItems();
      m_wndCallStack.m_wndCtrl.EnableWindow(FALSE);
    }
    CSingleLock sl(&m_csStackFrames, TRUE);
    m_StackFrames.clear();
  }

  //Reset the Modules window
  {
    if (m_wndModules.m_wndCtrl.GetSafeHwnd())
    {
      m_wndModules.m_wndCtrl.DeleteAllItems();
      m_wndModules.m_wndCtrl.EnableWindow(FALSE);
    }
    CSingleLock sl(&m_csModules, TRUE);
    m_Modules.clear();
  }

  //Reset the breakpoints window
  if (m_wndBreakpoints.m_wndCtrl.GetSafeHwnd())
  {
    const int nBreakpoints{m_wndBreakpoints.m_wndCtrl.GetItemCount()};
    for (int i{0}; i<nBreakpoints; i++)
      m_wndBreakpoints.m_wndCtrl.SetItemData(i, 0);
    m_wndBreakpoints.m_wndCtrl.UpdateWindow();
  }

  //Reset the threads window
  {
    if (m_wndThreads.m_wndCtrl.GetSafeHwnd())
    {
      m_wndThreads.m_wndCtrl.DeleteAllItems();
      m_wndThreads.m_wndCtrl.EnableWindow(FALSE);
    }
    CSingleLock sl(&m_csThreads, TRUE);
    m_Threads.clear();
  }

  //Reset the Watch window
  m_wndWatch.m_wndCtrl.ResetValuesForAllVariables();
  m_wndWatch.m_wndCtrl.UpdateVariableWindow();

  //Release the console
  if (m_bConsoleAllocated)
  {
    if (!FreeConsole())
      AddOutputTextU(IDS_PYDEBUG_FAIL_FREE_CONSOLE, GetLastError());
    m_bConsoleAllocated = false;
  }
}

#pragma warning(suppress: 26429)
void CPyDebugFrameWnd::OnUpdateDebugStop(_In_ CCmdUI* pCmdUI)
{
  //Allow stop if we are debugging
  const DEBUG_STATE eDebugState{m_DebugState};
#pragma warning(suppress: 26486)
  pCmdUI->Enable((eDebugState != DEBUG_STATE::NOT_DEBUGGING) && (eDebugState != DEBUG_STATE::DEBUG_THREAD_EXITED));

  //Do the cleanup if the thread has exited
  if (eDebugState == DEBUG_STATE::DEBUG_THREAD_EXITED)
    OnDebugStop();
}

UINT CPyDebugFrameWnd::_DebugThread(_In_opt_ LPVOID pParam)
{
  //Get back the "this" pointer from the pParam parameter
#pragma warning(suppress: 26429)
  auto pThis{static_cast<CPyDebugFrameWnd*>(pParam)};
  ASSERT(pThis != nullptr);

  return pThis->DebugThread();
}

bool CPyDebugFrameWnd::RedirectIO(_In_opt_ PyObject* pIOModule, _In_opt_ PyObject* pSysModule, _In_opt_z_ LPCSTR pszPythonName, _In_opt_z_ LPCSTR pszName, _In_opt_z_ LPCSTR pszMode, _In_ CPyObject& fileObject, _Inout_ CPyObject& originalIO)
{
  //Validate our parameters
  ASSERT(pIOModule != nullptr);
  ASSERT(pSysModule != nullptr);

  fileObject = PyObject_CallMethod(pIOModule, "open", "ss", pszName, pszMode);
  if (fileObject == nullptr)
  {
    AddOutputText(IDS_PYDEBUG_FAIL_OPEN_FILE_HANDLE, GetPythonErrorString());
    return false;
  }

  originalIO = PyObject_GetAttrString(pSysModule, pszPythonName);
  const int nReturn{PyObject_SetAttrString(pSysModule, pszPythonName, fileObject)};
  if (nReturn == -1)
  {
    AddOutputText(IDS_PYDEBUG_FAIL_REDIRECT_IO, GetPythonErrorString());
    return false;
  }
  return true;
}

#pragma warning(suppress: 26440)
void CPyDebugFrameWnd::ClosePythonFile(_In_opt_ PyObject* pFile)
{
  //Validate our parameters
  ASSERT(pFile != nullptr);

  CPyObject pResult{PyObject_CallMethod(pFile, "close", "")};
}

CString CPyDebugFrameWnd::GetPythonErrorString()
{
  PyObject* pType{nullptr};
  PyObject* pValue{nullptr};
  PyObject* pTraceback{nullptr};
#pragma warning(suppress: 26486)
  PyErr_Fetch(&pType, &pValue, &pTraceback);
  if (pType == nullptr)
  {
#pragma warning(suppress: 26429 26477 26486)
    Py_XDECREF(pValue);
#pragma warning(suppress: 26429 26477 26486)
    Py_XDECREF(pTraceback);
    return {};
  }
  PyErr_NormalizeException(&pType, &pValue, &pTraceback);
  if (pValue == nullptr)
  {
    pValue = Py_None;
    Py_INCREF(pValue);
  }
  CPyObject type{pType};
  CPyObject value{pValue};
  CPyObject traceback{pTraceback};
  CString sTracebackDetails;
  if (pTraceback != nullptr)
  {
    CPyObject moduleName{PyUnicode_FromString("traceback")};
    CPyObject PythonModule{PyImport_Import(moduleName)};
    if (PythonModule != nullptr)
    {
      CPyObject function{PyObject_GetAttrString(PythonModule, "format_exception")};
      if (function != nullptr)
      {
        if (PyCallable_Check(function))
        {
          CPyObject pythonValue{PyObject_CallFunctionObjArgs(function, type.operator PyObject*(), value.operator PyObject*(), pTraceback, nullptr)};
          sTracebackDetails = PyToCString(pythonValue);
        }
      }
    }
  }
  CString sType{PyToCString(type)};
  CString sValue{PyToCString(value)};
  CString sTempTraceback{PyToCString(traceback)};
  std::array<LPCTSTR, 3> rglpsz{sType, sValue, nullptr};
  if (sTracebackDetails.GetLength())
#pragma warning(suppress: 26446)
    rglpsz[2] = sTracebackDetails;
  else
#pragma warning(suppress: 26446)
    rglpsz[2] = sTempTraceback;
  CString sMessage;
#pragma warning(suppress: 26472)
  AfxFormatStrings(sMessage, IDS_PYDEBUG_EXCEPTION_MESSAGE, rglpsz.data(), static_cast<int>(rglpsz.size()));
  return sMessage;
}

void CPyDebugFrameWnd::CloseIOFiles(_In_ PyObject* pStdErr, _In_ PyObject* pStdOut, _In_ PyObject* pStdIn)
{
  if (pStdErr != nullptr)
    ClosePythonFile(pStdErr);
  if (pStdIn != nullptr)
    ClosePythonFile(pStdIn);
  if (pStdOut != nullptr)
    ClosePythonFile(pStdOut);
}

BOOL CPyDebugFrameWnd::SetupRedirectIO(_In_opt_ PyObject* pSysModule, _In_ CPyObject& pStdOut, _In_ CPyObject& pStdIn, _In_ CPyObject& pStdErr)
{
  //Redirect IO if we have a console
  if (!m_bConsoleAllocated)
    return TRUE;

  if (pSysModule == nullptr)
  {
    AddOutputText(IDS_PYDEBUG_FAIL_IMPORT_SYS, GetPythonErrorString());
    return FALSE;
  }
  CPyObject pIOModule{PyImport_ImportModule("io")};
  if (pIOModule == nullptr)
  {
    AddOutputText(IDS_PYDEBUG_FAIL_IMPORT_IO, GetPythonErrorString());
    return FALSE;
  }

  //Redirect sys.stdout to CONOUT$ using the "io" and "sys" modules
  m_pOriginalStdErr.Release();
#pragma warning(suppress: 26486)
  if (!RedirectIO(pIOModule, pSysModule, "stdout", "CONOUT$", "wt", pStdOut, m_pOriginalStdErr))
    return FALSE;

  //Redirect sys.stdin to CONIN$ using the "io" and "sys" modules
  m_pOriginalStdIn.Release();
#pragma warning(suppress: 26486)
  if (!RedirectIO(pIOModule, pSysModule, "stdin", "CONIN$", "rt", pStdIn, m_pOriginalStdIn))
  {
    ResetIO(pSysModule);
    CloseIOFiles(pStdErr, pStdIn, pStdOut);
    return FALSE;
  }

  //Redirect sys.stderr to CONOUT$ using the "io" and "sys" modules
  m_pOriginalStdErr.Release();
#pragma warning(suppress: 26486)
  if (!RedirectIO(pIOModule, pSysModule, "stderr", "CONOUT$", "wt", pStdErr, m_pOriginalStdErr))
  {
    ResetIO(pSysModule);
    CloseIOFiles(pStdErr, pStdIn, pStdOut);
    return FALSE;
  }

  return TRUE;
}

void CPyDebugFrameWnd::ResetIO(_In_opt_ PyObject* pSysModule)
{
  //Reset sys.stdout
  if (m_pOriginalStdErr)
    ResetIO(pSysModule, "stdout", m_pOriginalStdErr);

  //Reset sys.stdin
  if (m_pOriginalStdIn)
    ResetIO(pSysModule, "stdin", m_pOriginalStdIn);

  //Reset sys.stderr
  if (m_pOriginalStdErr)
    ResetIO(pSysModule, "stderr", m_pOriginalStdErr);
}

BOOL CPyDebugFrameWnd::ResetIO(_In_opt_ PyObject* pSysModule, _In_opt_z_ LPCSTR pszPythonName, _In_opt_ PyObject* pOriginalObject)
{
  //Validate our parameters
  ASSERT(pSysModule != nullptr);

  const int nReturn{PyObject_SetAttrString(pSysModule, pszPythonName, pOriginalObject)};
  if (nReturn == -1)
  {
    AddOutputText(IDS_PYDEBUG_FAIL_RESET_IO, CString(pszPythonName), GetPythonErrorString());
    return false;
  }
  return true;
}

BOOL CPyDebugFrameWnd::SetupCodeObject(_In_ CPyObject& pBytes)
{
  //Convert the code text to run to a Python Unicode string
  CPyObject pUnicode{PyUnicode_FromWideChar(m_sCode, -1)};
  if (pUnicode == nullptr)
  {
    AddOutputText(IDS_PYDEBUG_FAIL_CONVERT_CODE, GetPythonErrorString());
    return FALSE;
  }
  ASSERT(pBytes == nullptr);
  pBytes.Attach(PyUnicode_AsUTF8String(pUnicode));
  if (pBytes == nullptr)
  {
    AddOutputText(IDS_PYDEBUG_FAIL_CONVERT_TO_UTF8, GetPythonErrorString());
    return FALSE;
  }

  return TRUE;
}

UINT CPyDebugFrameWnd::DebugThread()
{
  const DWORD dwCurrentThreadId{GetCurrentThreadId()};
  m_dwDebuggedThreadID = dwCurrentThreadId;
  CString sCurrentThreadID;
  sCurrentThreadID.Format(_T("%x"), dwCurrentThreadId);
  AddOutputText(IDS_PYDEBUG_DEBUGGING_THREAD_START, sCurrentThreadID);

  //Acquire the GIL
  CPyGILAutoLock lock;

  //Create the capsule object (used to allow us to pass a C++ pointer to the PyEval_SetTrace API)
  CPyObject pThisObject{PyCapsule_New(this, g_CAPSULE_IDENTIFIER, nullptr)};
  if (pThisObject == nullptr)
  {
    AddOutputText(IDS_PYDEBUG_FAIL_CREATE_CAPSULE, GetPythonErrorString());
    m_DebugState = DEBUG_STATE::DEBUG_THREAD_EXITED;
    return 1;
  }

  //Initialize the Python debugger API "PyEval_SetTrace"
  AddOutputText(IDS_PYDEBUG_INITIALIZE_PYTHON_DEBUGGER);
  PyEval_SetTrace(_TraceCallback, pThisObject);

  //Redirect IO if we have a console
  CPyObject pSysModule{PyImport_ImportModule("sys")};
  CPyObject pStdOut;
  CPyObject pStdIn;
  CPyObject pStdErr;
#pragma warning(suppress: 26486)
  if (!SetupRedirectIO(pSysModule, pStdOut, pStdIn, pStdErr))
  {
    //Tidy up before we exit
    PyEval_SetTrace(nullptr, nullptr);
    m_DebugState = DEBUG_STATE::DEBUG_THREAD_EXITED;
    return 1;
  }

  //Convert the code text to run to a Python bytes object
  CPyObject pBytes;
#pragma warning(suppress: 26486)
  if (!SetupCodeObject(pBytes))
  {
    //Tidy up before we exit
    ResetIO(pSysModule);
    CloseIOFiles(pStdErr, pStdIn, pStdOut);
    PyEval_SetTrace(nullptr, nullptr);
    m_DebugState = DEBUG_STATE::DEBUG_THREAD_EXITED;
    return 1;
  }

  //Run the code using PyRun_SimpleStringFlags
  AddOutputText(IDS_PYDEBUG_DEBUGGING_CODE);
  PyCompilerFlags cf;
  cf.cf_flags = 0;
  const char* szCode{PyBytes_AsString(pBytes)};
  const int nReturn{PyRun_SimpleStringFlags(szCode, &cf)};
  if (nReturn != 0)
    AddOutputTextI(IDS_PYDEBUG_FAIL_RUN_CODE, nReturn);

  AddOutputTextU(IDS_PYDEBUG_DEBUGGING_THREAD_EXIT, dwCurrentThreadId);

  //Tidy up before we exit
  ResetIO(pSysModule);
  CloseIOFiles(pStdErr, pStdIn, pStdOut);
  PyEval_SetTrace(nullptr, nullptr);
  m_DebugState = DEBUG_STATE::DEBUG_THREAD_EXITED;

  return nReturn;
}

bool CPyDebugFrameWnd::DebugCode(_In_ const CStringW& sCode, _In_ const CString& sModuleName, _In_ START_DEBUGGING_TYPE eType)
{
  //Ensure the worker thread is fully closed down if required
  OnDebugStop();

  //Allocate a console for debugging
  if (m_wndProperties.GetSafeHwnd() && m_wndProperties.GetAllocateConsole())
  {
    if (!AllocConsole())
    {
      AddOutputTextU(IDS_PYDEBUG_FAIL_ALLOCATE_CONSOLE, GetLastError());
      return false;
    }
    m_bConsoleAllocated = true;
    SetConsoleTitle(sModuleName);
    //Prevent the close button in the console window from also closing this app
    HWND hConsoleWnd{::GetConsoleWindow()};
    if (hConsoleWnd != nullptr)
    {
      HMENU hMenu{::GetSystemMenu(hConsoleWnd, FALSE)};
      if (hMenu != nullptr)
        DeleteMenu(hMenu, SC_CLOSE, MF_BYCOMMAND);
    }
  }
  else
    AddOutputText(IDS_PYDEBUG_ERROR_NOT_ALLOCATING_CONSOLE);

  //Hive away the code in a member variable
  m_sCode = sCode;

  //Hive away the module name in a member variable
  m_sModuleName = sModuleName;

  //Hive away the main GUI thread id
  m_dwGUIThreadID = GetCurrentThreadId();

  //Create the thread which runs the Python code 
  ASSERT(m_pDebugThread == nullptr);
  m_pDebugThread = AfxBeginThread(_DebugThread, this, THREAD_PRIORITY_IDLE, 0, CREATE_SUSPENDED);
  if (m_pDebugThread == nullptr)
  {
    AddOutputTextU(IDS_PYDEBUG_FAIL_CREATE_DEBUGGING_THREAD, GetLastError());
    return false;
  }
  m_pDebugThread->m_bAutoDelete = FALSE;
  m_DebugState = DEBUG_STATE::RUNNING;
  m_StartDebuggingType = eType;
  m_pDebugThread->ResumeThread();

  return true;
}

CPyDebugDoc* CPyDebugFrameWnd::FindDebuggerDocument(_In_ const CString& sFilename, _In_ bool bAllowOpen)
{
  //Try the currently active document first
  CPyDebugDoc* pActiveDoc{FindActiveDebuggerDocument()};
  if (pActiveDoc != nullptr)
  {
    if ((sFilename == pActiveDoc->GetPathName()))
      return pActiveDoc;
  }

  //Check to see if the file is already open
#pragma warning(suppress: 26496)
  AFXASSUME(m_pDocTemplate != nullptr); //Did you forget to call CPyDebugFrameWnd::SetDocTemplate
  POSITION posDoc{m_pDocTemplate->GetFirstDocPosition()};
  while (posDoc != nullptr)
  {
    CDocument* pDoc{m_pDocTemplate->GetNextDoc(posDoc)};
#pragma warning(suppress: 26496)
    AFXASSUME(pDoc != nullptr);
    if (pDoc->IsKindOf(RUNTIME_CLASS(CPyDebugDoc)) && (pDoc->GetPathName() == sFilename))
#pragma warning(suppress: 26466)
      return static_cast<CPyDebugDoc*>(pDoc);
  }

  //Ok, the file is not already open, lets open it if allowed
  if (bAllowOpen)
#pragma warning(suppress: 26466)
    return static_cast<CPyDebugDoc*>(m_pDocTemplate->OpenDocumentFile(sFilename));

  return nullptr;
}

void CPyDebugFrameWnd::HandleInstructionPointerUpdate()
{
  CSingleLock sl{&m_csStackFrames, TRUE};

  //Remove the instruction pointer marker from all currently open documents
  RemoveAllInstructionPointers();

  //Add the instruction pointer to the correct document
#pragma warning(suppress: 26446)
  const CPyDebugStackFrame& stackFrame{m_StackFrames[0]};
  const CPyDebugDoc* pDoc{FindDebuggerDocument(stackFrame.m_sModule, true)};
  if (pDoc != nullptr)
  {
#pragma warning(suppress: 26429)
    auto* pView{pDoc->GetView()};
    ASSERT(pView != nullptr);
    auto& rCtrl{pView->GetCtrl()};

    //Readd the new instruction pointer marker
    rCtrl.MarkerAdd(stackFrame.m_LineNumber - 1, SC_PYTHON_MARK_INSTRUCTION_POINTER);
    rCtrl.GotoPos(rCtrl.PositionFromLine(stackFrame.m_LineNumber - 1));

    //Finally activate the parent frame
#pragma warning(suppress: 26429)
    CFrameWnd* pActiveFrame{pView->GetParentFrame()};
    ASSERT(pActiveFrame != nullptr);
    pActiveFrame->ActivateFrame();
  }
}

LRESULT CPyDebugFrameWnd::OnDebugNotification(_In_ WPARAM wParam, _In_ LPARAM lParam)
{
  switch (wParam)
  {
    case DEBUG_NOTIFCATION_UPDATE_WATCH_WINDOW2:
    {
#pragma warning(suppress: 26472)
      const int nVariablesWndRequestIndex{static_cast<int>(lParam)};
      m_wndWatch.m_wndCtrl.UpdateVariableWindow(nVariablesWndRequestIndex);
      break;
    }
    case DEBUG_NOTIFCATION_UPDATE_WATCH_WINDOW:
    {
#pragma warning(suppress: 26472)
      const int nVariablesWndRequestIndex{static_cast<int>(lParam)};
      m_wndWatch.m_wndCtrl.UpdateVariable(nVariablesWndRequestIndex);
      break;
    }
    case DEBUG_NOTIFCATION_UPDATE_INSTRUCTION_POINTER:
    {
      //Handle update the instruction pointer
      HandleInstructionPointerUpdate();

      //Also update the locals window
      m_wndLocals.m_wndCtrl.UpdateVariableWindow();

      //Also update the globals window
      m_wndGlobals.m_wndCtrl.UpdateVariableWindow();

      //Also update the Built-Ins window
      m_wndBuiltIns.m_wndCtrl.UpdateVariableWindow();

      //Also update the Call Stack window
      UpdateCallStackWindow();

      //Also update the Modules window
      {
        m_wndModules.m_wndCtrl.SetRedraw(FALSE);
        m_wndModules.m_wndCtrl.DeleteAllItems();
        m_wndModules.m_wndCtrl.EnableWindow(TRUE);
        CSingleLock sl{&m_csModules, TRUE};
        const auto nModules{m_Modules.size()};
        if (nModules == 0)
        {
          CString sText;
#pragma warning(suppress: 26496)
          ENSURE(sText.LoadString(IDS_PYDEBUG_NO_MODULES_DEFINED));
          m_wndModules.m_wndCtrl.InsertItem(0, sText, 1);
        }
        else
        {
          int i{0};
          for (const auto& module : m_Modules)
          {
            const int nIndex{m_wndModules.m_wndCtrl.InsertItem(i, module.m_sName, 0)};
            i++;
            m_wndModules.m_wndCtrl.SetItemText(nIndex, 1, module.m_sDetails);
          }
        }
        m_wndModules.m_wndCtrl.SetRedraw(TRUE);
      }

      //Also update the breakpoints window
      UpdateBreakpointsWindow();

      //Also update the Threads window
      {
        m_wndThreads.m_wndCtrl.SetRedraw(FALSE);
        m_wndThreads.m_wndCtrl.DeleteAllItems();
        m_wndThreads.m_wndCtrl.EnableWindow(TRUE);
        CSingleLock sl{&m_csThreads, TRUE};
        const auto nThreads{m_Threads.size()};
        if (nThreads == 0)
        {
          CString sText;
#pragma warning(suppress: 26496)
          ENSURE(sText.LoadString(IDS_PYDEBUG_NO_THREADS));
          const int nIndex{m_wndThreads.m_wndCtrl.InsertItem(0, sText, 1)};
          m_wndThreads.m_wndCtrl.SetItemData(nIndex, 0);
        }
        else
        {
          const DWORD dwDebuggedThreadId{m_dwDebuggedThreadID};
          int i{0};
          for (const auto& thread : m_Threads)
          {
            CString sThreadID;
            sThreadID.Format(_T("%u"), thread.m_dwThreadID);
            const int nIndex{m_wndThreads.m_wndCtrl.InsertItem(i, sThreadID, thread.m_bHit ? 0 : 1)};
            i++;
            m_wndThreads.m_wndCtrl.SetItemData(nIndex, thread.m_bHit);

            CString sText;
            if (thread.m_dwThreadID == m_dwGUIThreadID)
#pragma warning(suppress: 26496)
              ENSURE(sText.LoadString(IDS_PYDEBUG_MAIN_THREAD));
            else if (thread.m_dwThreadID == dwDebuggedThreadId)
#pragma warning(suppress: 26496)
              ENSURE(sText.LoadString(IDS_PYDEBUG_PYTHON_DEBUGGER_THREAD));
            else
#pragma warning(suppress: 26496)
              ENSURE(sText.LoadString(IDS_PYDEBUG_WORKER_THREAD));
            m_wndThreads.m_wndCtrl.SetItemText(nIndex, 1, sText);
            CString sPriority;
            switch (thread.m_nPriority)
            {
              case THREAD_PRIORITY_ABOVE_NORMAL:
              {
#pragma warning(suppress: 26496)
                ENSURE(sPriority.LoadString(IDS_PYDEBUG_ABOVE_NORMAL));
                break;
              }
              case THREAD_PRIORITY_BELOW_NORMAL:
              {
#pragma warning(suppress: 26496)
                ENSURE(sPriority.LoadString(IDS_PYDEBUG_BELOW_NORMAL));
                break;
              }
              case THREAD_PRIORITY_HIGHEST:
              {
#pragma warning(suppress: 26496)
                ENSURE(sPriority.LoadString(IDS_PYDEBUG_HIGHEST));
                break;
              }
              case THREAD_PRIORITY_IDLE:
              {
#pragma warning(suppress: 26496)
                ENSURE(sPriority.LoadString(IDS_PYDEBUG_IDLE));
                break;
              }
              case THREAD_PRIORITY_LOWEST:
              {
#pragma warning(suppress: 26496)
                ENSURE(sPriority.LoadString(IDS_PYDEBUG_LOWEST));
                break;
              }
              case THREAD_PRIORITY_NORMAL:
              {
#pragma warning(suppress: 26496)
                ENSURE(sPriority.LoadString(IDS_PYDEBUG_NORMAL));
                break;
              }
              case THREAD_PRIORITY_TIME_CRITICAL:
              {
#pragma warning(suppress: 26496)
                ENSURE(sPriority.LoadString(IDS_PYDEBUG_TIME_CRITICAL));
                break;
              }
              default:
              {
#pragma warning(suppress: 26496)
                ENSURE(sPriority.LoadString(IDS_PYDEBUG_UNKNOWN));
                break;
              }
            }
            m_wndThreads.m_wndCtrl.SetItemText(nIndex, 2, sPriority);
          }
        }
        m_wndThreads.m_wndCtrl.SetRedraw(TRUE);
      }

      //Also update the Watch window
      m_wndWatch.m_wndCtrl.UpdateVariableWindow();

      break;
    }
    default:
    {
      ASSERT(FALSE);
      break;
    }
  }

  return 0L;
}

int CPyDebugFrameWnd::_TraceCallback(_In_opt_ PyObject* obj, _In_opt_ PyFrameObject* frame, _In_ int what, _In_opt_ PyObject* arg)
{
  //Convert from the C to the C++ world
#pragma warning(suppress: 26429)
  auto pThis{static_cast<CPyDebugFrameWnd*>(PyCapsule_GetPointer(obj, g_CAPSULE_IDENTIFIER))};
  ASSERT(pThis != nullptr);
  return pThis->TraceCallback(obj, frame, what, arg);
}

void CPyDebugFrameWnd::CaptureModulesDetails()
{
  CSingleLock sl{&m_csModules, TRUE};
  m_Modules.clear();
  PyObject* pModules{PyImport_GetModuleDict()};
  if (pModules != nullptr)
  {
    Py_ssize_t pos{0};
    PyObject* pKey{nullptr};
    PyObject* pValue{nullptr};
#pragma warning(suppress: 26486)
    while (PyDict_Next(pModules, &pos, &pKey, &pValue))
    {
      if ((pKey != nullptr) && (pValue != nullptr))
      {
        CPyDebugModule module;
        module.m_sName = PyToCString(pKey);
        module.m_sDetails = PyToCString(pValue);
        m_Modules.push_back(module);
      }
    }
  }
  std::sort(m_Modules.begin(), m_Modules.end(), CPyDebugLessThanPyModule());
}

void CPyDebugFrameWnd::CaptureThreadsDetails(_In_ DWORD dwThreadIdBeingHit)
{
  CSingleLock sl{&m_csThreads, TRUE};
  m_Threads.clear();

  //Capture all the threads in the current process using ToolHelp32
  const DWORD dwCurrentProcessID{GetCurrentProcessId()};
  HANDLE h{CreateToolhelp32Snapshot(TH32CS_SNAPTHREAD, 0)};
  if (h != INVALID_HANDLE_VALUE) 
  {
    THREADENTRY32 te;
    te.dwSize = sizeof(te);
    if (Thread32First(h, &te)) 
    {
      do
      {
        if (dwCurrentProcessID == te.th32OwnerProcessID)
        {
          CPyDebugThread thread;
          thread.m_dwThreadID = te.th32ThreadID;
          thread.m_bHit = (thread.m_dwThreadID == dwThreadIdBeingHit);
          HANDLE hThread{OpenThread(THREAD_QUERY_INFORMATION, FALSE, thread.m_dwThreadID)};
          if (hThread != nullptr)
          {
            #pragma warning(suppress: 6001)
            thread.m_nPriority = GetThreadPriority(hThread);
            CloseHandle(hThread);
          }

          m_Threads.push_back(thread);
        }
      }
      while (Thread32Next(h, &te));
    }
    CloseHandle(h);
  }
}

void CPyDebugFrameWnd::HandleBreakpointsForTraceCallback(_In_ PyFrameObject* frame, _Out_ bool& bBreakAll)
{
  //Set the output parameter to a default value
  bBreakAll = false;

  //Implement support for breaking on breakpoints here
  const int nLineNumber = PyFrame_GetLineNumber(frame);
  CSingleLock sl{&m_csBreakpoints, TRUE};
  for (auto& breakpoint : m_Breakpoints)
  {
    if (breakpoint.m_bEnabled && (breakpoint.m_LineNumber == nLineNumber))
    {
      breakpoint.m_bHit = true;
      bBreakAll = true;
    }
    else
      breakpoint.m_bHit = false;
  }
}

void CPyDebugFrameWnd::HandleBreakpointsForTraceCallback2(_In_ PyFrameObject* frame)
{
  //Implement support for breaking on breakpoints here
  const int nLineNumber = PyFrame_GetLineNumber(frame);
  CSingleLock sl{&m_csBreakpoints, TRUE};
  for (auto& breakpoint : m_Breakpoints)
  {
    if (breakpoint.m_bEnabled && (breakpoint.m_LineNumber == nLineNumber))
      breakpoint.m_bHit = true;
    else
      breakpoint.m_bHit = false;
  }
}

void CPyDebugFrameWnd::CaptureStackFrame(_In_ PyFrameObject* frame, _Out_ CPyDebugStackFrame& CapturedFrame)
{
  //Setup the new stack frame details
  CapturedFrame.m_LineNumber = PyFrame_GetLineNumber(frame);
  PyCodeObject* code = PyFrame_GetCode(frame);
#pragma warning(suppress: 26489)
  CapturedFrame.m_sFunction = PyToCString(code->co_name);
#pragma warning(suppress: 26489)
  CapturedFrame.m_sModule = PyToCString(code->co_filename);
  Py_DECREF(code);
  if (CapturedFrame.m_sModule == _T("<string>")) //Patch up m_sCurrentIPFilename because we are using "PyRun_SimpleStringFlags"
    CapturedFrame.m_sModule = m_sModuleName;
}

void CPyDebugFrameWnd::HandleVariableNameChangeForTraceCallback()
{
  CSingleLock sl{&m_wndWatch.m_wndCtrl.m_csVariables, TRUE};
#pragma warning(suppress: 26446)
  CPyDebugVariable& variable{m_wndWatch.m_wndCtrl.m_Variables[m_nVariablesWndRequestIndex]};

  //Reset the changed state for all variables
  m_wndWatch.m_wndCtrl.ResetChangedStateForAllVariables();

  //Get access to the "eval" function
  CPyObject pBuiltins{PyImport_ImportModule("builtins")};
  CPyObject pEval;
  if (pBuiltins != nullptr)
    pEval = PyObject_GetAttrString(pBuiltins, "eval");

  //Capture the current value for the variable
#pragma warning(suppress: 26486)
  m_wndWatch.m_wndCtrl.CaptureVariableDetail(variable, pEval, false);
  variable.m_bChanged = false;

  //Update the UI
  PostMessage(WM_PYTHON_DEBUG_NOTIFICATION, DEBUG_NOTIFCATION_UPDATE_WATCH_WINDOW, m_nVariablesWndRequestIndex);
}

void CPyDebugFrameWnd::HandleVariableValueChangeForTraceCallback()
{
  CSingleLock sl{&m_wndWatch.m_wndCtrl.m_csVariables, TRUE};
#pragma warning(suppress: 26446)
  CPyDebugVariable& variable{m_wndWatch.m_wndCtrl.m_Variables[m_nVariablesWndRequestIndex]};

  //Reset the changed state for all variables
  m_wndWatch.m_wndCtrl.ResetChangedStateForAllVariables();

  //Get access to the "eval" function
  CPyObject pBuiltins{PyImport_ImportModule("builtins")};
  CPyObject pEval;
  if (pBuiltins != nullptr)
    pEval = PyObject_GetAttrString(pBuiltins, "eval");

  //Set the current value for the variable
  if (m_wndWatch.m_wndCtrl.SetVariableValue(variable, m_sVariablesWndRequestString))
#pragma warning(suppress: 26486)
    m_wndWatch.m_wndCtrl.CaptureVariableDetail(variable, pEval, true);

  //Next update the state of all ancestor and descendant variables from the current variable level
  int nAncestorIndex{m_wndWatch.m_wndCtrl.FindParentVariableIndex(m_nVariablesWndRequestIndex)};
  while (nAncestorIndex != -1)
  {
#pragma warning(suppress: 26446)
    CPyDebugVariable& ancestorVariable{m_wndWatch.m_wndCtrl.m_Variables[nAncestorIndex]};
#pragma warning(suppress: 26486)
    m_wndWatch.m_wndCtrl.CaptureVariableDetail(ancestorVariable, pEval, true);

    //Prepare for the next loop
    nAncestorIndex = m_wndWatch.m_wndCtrl.FindParentVariableIndex(nAncestorIndex);
  }
  bool bMoreChildren{true};
  const int nVariables{m_wndWatch.m_wndCtrl.GetItemCount()};
  for (int i{m_nVariablesWndRequestIndex+1}; (i<nVariables) && bMoreChildren; i++)
  {
#pragma warning(suppress: 26446)
    CPyDebugVariable& descendantVariable{m_wndWatch.m_wndCtrl.m_Variables[i]};
    bMoreChildren = descendantVariable.m_nLevel > variable.m_nLevel;
    if (bMoreChildren)
#pragma warning(suppress: 26486)
      m_wndWatch.m_wndCtrl.CaptureVariableDetail(descendantVariable, pEval, true);
  }

  //Update the UI
  PostMessage(WM_PYTHON_DEBUG_NOTIFICATION, DEBUG_NOTIFCATION_UPDATE_WATCH_WINDOW2, m_nVariablesWndRequestIndex);
}

void CPyDebugFrameWnd::HandleVariableExpansionForTraceCallback(_In_ CPyDebugVariablesListCtrl& variablesWnd)
{
  CSingleLock sl{&variablesWnd.m_csVariables, TRUE};
#pragma warning(suppress: 26446)
  CPyDebugVariable& variable{variablesWnd.m_Variables[m_nVariablesWndRequestIndex]};
  variable.m_bCollapsed = false;
  CString sVariableName{variable.m_sName};
  const int nVariableLevel = variable.m_nLevel;

  //Get access to the "eval" function
  CPyObject pBuiltins{PyImport_ImportModule("builtins")};
  CPyObject pEval;
  if (pBuiltins != nullptr)
    pEval = PyObject_GetAttrString(pBuiltins, "eval");

  //Get access to the variable as a callable object
  CPyObject pArgs{Py_BuildValue("(s)", CStringA(variable.m_sName).GetString())};
  PyErr_Clear();
  CPyObject pObject{PyObject_Call(pEval, pArgs, nullptr)};
  if (pObject != nullptr)
  {
    //Use the Sequence protocol to get the sub variables
    if (PySequence_Check(pObject))
    {
      const auto nSize{PySequence_Size(pObject)};
      for (Py_ssize_t i{0}; i<nSize; i++)
      {
        CPyObject subItem{PySequence_GetItem(pObject, i)};

        //Add the sub variable to the in-memory array
        CPyDebugVariable tempVariable;
#pragma warning(suppress: 26472)
        tempVariable.m_sName.Format(_T("%s[%d]"), sVariableName.GetString(), static_cast<int>(i));
        tempVariable.m_nLevel = nVariableLevel + 1;
        variablesWnd.m_Variables.insert(variablesWnd.m_Variables.cbegin() + m_nVariablesWndRequestIndex + i + 1, tempVariable);
      }
    }

    //Use the Mapping protocol to get the sub variables
    if (PyMapping_Check(pObject))
    {
      CPyObject pItems{PyMapping_Items(pObject)};
      const auto nSize{PyMapping_Size(pItems)};
      for (Py_ssize_t i{0}; i<nSize; i++)
      {
        CPyObject pPair{PySequence_GetItem(pItems, i)};
        CPyObject pKey{PySequence_GetItem(pPair, 0)};

        //Add the sub variable to the in-memory array
        CPyDebugVariable tempVariable;
        tempVariable.m_sName.Format(_T("%s['%s']"), sVariableName.GetString(), CPyDebugFrameWnd::PyToCString(pKey).GetString());
        tempVariable.m_nLevel = nVariableLevel + 1;
        variablesWnd.m_Variables.insert(variablesWnd.m_Variables.cbegin() + m_nVariablesWndRequestIndex + i + 1, tempVariable);
      }
    }

    //Also get any sub objects via the dictionary
    CPyObject pDict{PyObject_GenericGetDict(pObject, nullptr)};
    if (pDict != nullptr)
    {
      Py_ssize_t pos{0};
      PyObject* pKey{nullptr};
      PyObject* pValue{nullptr};
      int i{0};
#pragma warning(suppress: 26486)
      while (PyDict_Next(pDict, &pos, &pKey, &pValue))
      {
        if ((pKey != nullptr) && (pValue != nullptr))
        {
          //Add the sub variable to the in-memory array
          CPyDebugVariable tempVariable;
          tempVariable.m_sName.Format(_T("%s.%s"), sVariableName.GetString(), CPyDebugFrameWnd::PyToCString(pKey).GetString());
          tempVariable.m_nLevel = nVariableLevel + 1;
          variablesWnd.m_Variables.insert(variablesWnd.m_Variables.cbegin() + m_nVariablesWndRequestIndex + i + 1, tempVariable);

          //Prepare for the next loop
          ++i;
        }
      }
    }
  }
  PyErr_Clear();
  variablesWnd.CaptureVariablesDetails(false);

  //Update the UI
  PostMessage(WM_PYTHON_DEBUG_NOTIFICATION, DEBUG_NOTIFCATION_UPDATE_WATCH_WINDOW2, m_nVariablesWndRequestIndex);
}

int CPyDebugFrameWnd::TraceCallback(_In_opt_ PyObject* /*obj*/, _In_opt_ PyFrameObject* frame, _In_ int what, _In_opt_ PyObject* arg)
{
  DEBUG_STATE eExpected{DEBUG_STATE::REQUEST_CONTINUE};
  if (m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::RUNNING))
    m_wndWatch.m_wndCtrl.ResetChangedStateForAllVariables();

  eExpected = DEBUG_STATE::REQUEST_STOP;
  if (m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::ACKNOWLEDGE_STOP))
  {
    AddOutputText(IDS_PYDEBUG_CALLING_PYERR_SETSTRING);
    PyErr_SetString(PyExc_InterruptedError, "code being debugged has been stopped");
    return 1;
  }

  eExpected = DEBUG_STATE::REQUEST_BREAK_ALL;
  if (m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL))
  {
    //Remember the instruction pointer, locals details
    m_wndLocals.m_wndCtrl.CaptureVariablesDetails(PyEval_GetLocals());
    m_wndGlobals.m_wndCtrl.CaptureVariablesDetails(PyEval_GetGlobals());
    m_wndBuiltIns.m_wndCtrl.CaptureVariablesDetails(PyEval_GetBuiltins());
    CaptureModulesDetails();
    CaptureThreadsDetails(GetCurrentThreadId());
    {
      CSingleLock sl{&m_csStackFrames, TRUE};
#pragma warning(suppress: 26446 26486)
      CaptureStackFrame(frame, m_StackFrames[0]);
    }
    m_wndWatch.m_wndCtrl.CaptureVariablesDetails(true);

    //Ask the UI to update the Instruction Pointer UI
    PostMessage(WM_PYTHON_DEBUG_NOTIFICATION, DEBUG_NOTIFCATION_UPDATE_INSTRUCTION_POINTER, -1);

    //Block further execution of the Python code
    while (m_DebugState == DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL)
    {
      Sleep(0);

      //Update a watch window variable name if required
      VARIABLES_WND_REQUEST eVariablesWndExpected{VARIABLES_WND_REQUEST::CHANGE_VARIABLE_NAME};
      if (m_VariablesWndRequest.compare_exchange_strong(eVariablesWndExpected, VARIABLES_WND_REQUEST::NO_REQUEST))
        HandleVariableNameChangeForTraceCallback();

      //Update a watch window variable value if required
      eVariablesWndExpected = VARIABLES_WND_REQUEST::CHANGE_VARIABLE_VALUE;
      if (m_VariablesWndRequest.compare_exchange_strong(eVariablesWndExpected, VARIABLES_WND_REQUEST::NO_REQUEST))
        HandleVariableValueChangeForTraceCallback();

      //Expand a watch window evariable if required
      eVariablesWndExpected = VARIABLES_WND_REQUEST::EXPAND_VARIABLE;
      if (m_VariablesWndRequest.compare_exchange_strong(eVariablesWndExpected, VARIABLES_WND_REQUEST::NO_REQUEST))
      {
        switch (m_VariablesWndRequestType)
        {
          case CPyDebugVariablesListCtrl::TYPE::BUILTINS:
          {
            HandleVariableExpansionForTraceCallback(m_wndBuiltIns.m_wndCtrl);
            break;
          }
          case CPyDebugVariablesListCtrl::TYPE::GLOBALS:
          {
            HandleVariableExpansionForTraceCallback(m_wndGlobals.m_wndCtrl);
            break;
          }
          case CPyDebugVariablesListCtrl::TYPE::LOCALS:
          {
            HandleVariableExpansionForTraceCallback(m_wndLocals.m_wndCtrl);
            break;
          }
          case CPyDebugVariablesListCtrl::TYPE::WATCH:
          {
            HandleVariableExpansionForTraceCallback(m_wndWatch.m_wndCtrl);
            break;
          }
          default:
          {
            ASSERT(FALSE);
            break;
          }
        }
      }
    }
  }

  switch (what)
  {
    case PyTrace_CALL:
    {
      //Update the stack frame in memory
      CPyDebugStackFrame stackFrame;
      CaptureStackFrame(frame, stackFrame);
      {
        CSingleLock sl{&m_csStackFrames, TRUE};
        m_StackFrames.push_front(stackFrame);
      }
      break;
    }
    case PyTrace_RETURN:
    {
      //Update the stack frame in memory
      {
        CSingleLock sl{&m_csStackFrames, TRUE};
        m_StackFrames.pop_front();
      }
      break;
    }
    case PyTrace_LINE:
    {
      //Update the stack frame in memory
      {
        CSingleLock sl{&m_csStackFrames, TRUE};
#pragma warning(suppress: 26446)
        CPyDebugStackFrame& stackFrame{m_StackFrames[0]};
        stackFrame.m_LineNumber = PyFrame_GetLineNumber(frame);
      }

      bool bBreakAll{false};
      if (m_DebugState == DEBUG_STATE::REQUEST_STEP_INTO) //Implement Step Into functionality
      {
        bBreakAll = true;
      }
      else if ((m_DebugState == DEBUG_STATE::REQUEST_STEP_OVER) || (m_DebugState == DEBUG_STATE::RUNNING_REQUEST_STEP_OVER)) //Implement Step Over functionality
      {
        if (m_DebugState == DEBUG_STATE::REQUEST_STEP_OVER)
          m_wndWatch.m_wndCtrl.ResetChangedStateForAllVariables();

        //Find the index of the "m_StackFrameAtRequestStepOver" in the stack frame
        int nStartStackFrameIndex{-1};
        {
          CSingleLock sl{&m_csStackFrames, TRUE};
#pragma warning(suppress: 26486)
          auto iter{std::find_if(m_StackFrames.begin(), m_StackFrames.end(), [this](const CPyDebugStackFrame& stackframe) noexcept
           {
             return ((stackframe.m_sFunction == m_StackFrameAtRequestStepOver.m_sFunction) && (stackframe.m_sModule == m_StackFrameAtRequestStepOver.m_sModule));
           })};
#pragma warning(suppress: 26486)
          if (iter != m_StackFrames.end())
#pragma warning(suppress: 26472 26486)
            nStartStackFrameIndex = static_cast<int>(iter - m_StackFrames.begin());
        }
        bBreakAll = (nStartStackFrameIndex <= 0);

        //Also give breakpoints a chance to fire
        if (!bBreakAll)
          HandleBreakpointsForTraceCallback(frame, bBreakAll);
        else
          HandleBreakpointsForTraceCallback2(frame);
      }
      else if ((m_DebugState == DEBUG_STATE::REQUEST_STEP_OUT) || (m_DebugState == DEBUG_STATE::RUNNING_REQUEST_STEP_OUT)) //Implement Step Out functionality)
      {
        if (m_DebugState == DEBUG_STATE::REQUEST_STEP_OUT)
          m_wndWatch.m_wndCtrl.ResetChangedStateForAllVariables();

        //Break if the m_StackFrameAtRequestStepOut is not now on the stack frame
        bBreakAll = true;
        {
          CSingleLock sl{&m_csStackFrames, TRUE};
#pragma warning(suppress: 26486)
          auto iter{std::find_if(m_StackFrames.begin(), m_StackFrames.end(), [this](const CPyDebugStackFrame& stackframe) noexcept
            {
              return ((stackframe.m_sFunction == m_StackFrameAtRequestStepOut.m_sFunction) && (stackframe.m_sModule == m_StackFrameAtRequestStepOut.m_sModule));
            })};
#pragma warning(suppress: 26486)
          if (iter != m_StackFrames.end())
            bBreakAll = false;
        }

        //Also give breakpoints a chance to fire
        if (!bBreakAll)
          HandleBreakpointsForTraceCallback(frame, bBreakAll);
        else
          HandleBreakpointsForTraceCallback2(frame);
      }
      else
      {
        //Implement Step Into/Step Over from start
        START_DEBUGGING_TYPE eExpectedType{START_DEBUGGING_TYPE::BREAK_ON_FIRST_STATEMENT};
        if (m_StartDebuggingType.compare_exchange_strong(eExpectedType, START_DEBUGGING_TYPE::NORMAL_START))
        {
          bBreakAll = TRUE;
          HandleBreakpointsForTraceCallback2(frame);
        }
        else
          HandleBreakpointsForTraceCallback(frame, bBreakAll);
      }

      if (bBreakAll)
      {
        //Remember the instruction pointer, locals details
        m_wndLocals.m_wndCtrl.CaptureVariablesDetails(PyEval_GetLocals());
        m_wndGlobals.m_wndCtrl.CaptureVariablesDetails(PyEval_GetGlobals());
        m_wndBuiltIns.m_wndCtrl.CaptureVariablesDetails(PyEval_GetBuiltins());
        CaptureModulesDetails();
        CaptureThreadsDetails(GetCurrentThreadId());
        {
          CSingleLock sl{&m_csStackFrames, TRUE};
#pragma warning(suppress: 26446 26486)
          CaptureStackFrame(frame, m_StackFrames[0]);
        }
        m_wndWatch.m_wndCtrl.CaptureVariablesDetails(true);

        m_DebugState = DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL;

        //Ask the UI to update the Instruction Pointer UI
        if (GetSafeHwnd())
          PostMessage(WM_PYTHON_DEBUG_NOTIFICATION, DEBUG_NOTIFCATION_UPDATE_INSTRUCTION_POINTER);

        //Block further execution of the Python code
        while (m_DebugState == DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL)
        {
          Sleep(0);

          //Update a watch window variable name if required
          VARIABLES_WND_REQUEST eVariablesWndExpected{VARIABLES_WND_REQUEST::CHANGE_VARIABLE_NAME};
          if (m_VariablesWndRequest.compare_exchange_strong(eVariablesWndExpected, VARIABLES_WND_REQUEST::NO_REQUEST))
            HandleVariableNameChangeForTraceCallback();

          //Update a watch window variable value if required
          eVariablesWndExpected = VARIABLES_WND_REQUEST::CHANGE_VARIABLE_VALUE;
          if (m_VariablesWndRequest.compare_exchange_strong(eVariablesWndExpected, VARIABLES_WND_REQUEST::NO_REQUEST))
            HandleVariableValueChangeForTraceCallback();

          //Expand a watch window evariable if required
          eVariablesWndExpected = VARIABLES_WND_REQUEST::EXPAND_VARIABLE;
          if (m_VariablesWndRequest.compare_exchange_strong(eVariablesWndExpected, VARIABLES_WND_REQUEST::NO_REQUEST))
          {
            switch (m_VariablesWndRequestType)
            {
              case CPyDebugVariablesListCtrl::TYPE::BUILTINS:
              {
                HandleVariableExpansionForTraceCallback(m_wndBuiltIns.m_wndCtrl);
                break;
              }
              case CPyDebugVariablesListCtrl::TYPE::GLOBALS:
              {
                HandleVariableExpansionForTraceCallback(m_wndGlobals.m_wndCtrl);
                break;
              }
              case CPyDebugVariablesListCtrl::TYPE::LOCALS:
              {
                HandleVariableExpansionForTraceCallback(m_wndLocals.m_wndCtrl);
                break;
              }
              case CPyDebugVariablesListCtrl::TYPE::WATCH:
              {
                HandleVariableExpansionForTraceCallback(m_wndWatch.m_wndCtrl);
                break;
              }
              default:
              {
                ASSERT(FALSE);
                break;
              }
            }
          }
        }
      }
      else
      {
        eExpected = DEBUG_STATE::REQUEST_STEP_OVER;
        m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::RUNNING_REQUEST_STEP_OVER);
        eExpected = DEBUG_STATE::REQUEST_STEP_OUT;
        m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::RUNNING_REQUEST_STEP_OUT);
        eExpected = DEBUG_STATE::NOT_DEBUGGING;
        m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::RUNNING);
        eExpected = DEBUG_STATE::ACKNOWLEDGE_STOP;
        m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::RUNNING);
        eExpected = DEBUG_STATE::REQUEST_BREAK_ALL;
        m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::RUNNING);
        eExpected = DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL;
        m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::RUNNING);
        eExpected = DEBUG_STATE::REQUEST_CONTINUE;
        m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::RUNNING);
        eExpected = DEBUG_STATE::REQUEST_STEP_INTO;
        m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::RUNNING);
      }
      break;
    }
    case PyTrace_EXCEPTION:
    {
      const int nLineNumber = PyFrame_GetLineNumber(frame);
      PyCodeObject* code = PyFrame_GetCode(frame);
#pragma warning(suppress: 26489 26489)
      AddOutputText(IDS_PYDEBUG_UNHANDLED_EXECPTION_OCCURRED, PyToCString(arg), nLineNumber, PyToCString(code->co_name), PyToCString(code->co_filename));
      Py_DECREF(code);
      break;
    }
    default:
    {
      break;
    }
  }

  return 0;
}

void CPyDebugFrameWnd::UpdateCallStackWindow()
{
  m_wndCallStack.m_wndCtrl.SetRedraw(FALSE);
  m_wndCallStack.m_wndCtrl.DeleteAllItems();
  m_wndCallStack.m_wndCtrl.EnableWindow(TRUE);
  CSingleLock sl{&m_csStackFrames, TRUE};
  const auto nFrames{m_StackFrames.size()};
  if (nFrames == 0)
  {
    CString sText;
#pragma warning(suppress: 26496)
    ENSURE(sText.LoadString(IDS_PYDEBUG_NO_STACK_FRAME));
    m_wndCallStack.m_wndCtrl.InsertItem(0, sText);
  }
  else
  {
    int i{0};
#pragma warning(suppress: 26486 26489)
    for (const auto& frame : m_StackFrames)
    {
      const int nIndex{m_wndCallStack.m_wndCtrl.InsertItem(i, frame.m_sModule, i==0 ? 0 : 1)};
      i++;
      m_wndCallStack.m_wndCtrl.SetItemText(nIndex, 1, frame.m_sFunction);
      CString sLine;
#pragma warning(suppress: 26472)
      sLine.Format(_T("%d"), static_cast<int>(frame.m_LineNumber));
      m_wndCallStack.m_wndCtrl.SetItemText(nIndex, 2, sLine);
    }
  }
  m_wndCallStack.m_wndCtrl.SetRedraw(TRUE);
}

void CPyDebugFrameWnd::UpdateBreakpointsWindow()
{
  m_wndBreakpoints.m_wndCtrl.SetRedraw(FALSE);
  m_wndBreakpoints.m_wndCtrl.DeleteAllItems();
  m_wndBreakpoints.m_wndCtrl.EnableWindow(TRUE);
  CSingleLock sl{&m_csBreakpoints, TRUE};
  int i{0};
  for (const auto& breakpoint : m_Breakpoints)
  {
    CString sText;
    CString sLineNumber;
#pragma warning(suppress: 26472)
    sLineNumber.Format(_T("%d"), static_cast<int>(breakpoint.m_LineNumber));
    AfxFormatString2(sText, IDS_PYDEBUG_BREAKPOINT_DETAILS, breakpoint.m_sModule, sLineNumber);
    const int nIndex{m_wndBreakpoints.m_wndCtrl.InsertItem(i, sText, breakpoint.m_bEnabled ? 0 : 2)};
    i++;
    m_wndBreakpoints.m_wndCtrl.SetCheck(nIndex, breakpoint.m_bEnabled);
    m_wndBreakpoints.m_wndCtrl.SetItemData(nIndex, breakpoint.m_bHit);
  }
  m_wndBreakpoints.m_wndCtrl.SetRedraw(TRUE);
}

void CPyDebugFrameWnd::OnDebugDisableAllBreakpoints()
{
  //Disable all breakpoints in-memory
  {
    CSingleLock sl{&m_csBreakpoints, TRUE};
    for (auto& breakpoint : m_Breakpoints)
      breakpoint.m_bEnabled = false;
  }

  //Ask each open document to disable all its breakpoints
#pragma warning(suppress: 26496)
  AFXASSUME(m_pDocTemplate != nullptr); //Did you forget to call CPyDebugFrameWnd::SetDocTemplate
  POSITION posDoc{m_pDocTemplate->GetFirstDocPosition()};
  while (posDoc != nullptr)
  {
    CDocument* pDoc{m_pDocTemplate->GetNextDoc(posDoc)};
#pragma warning(suppress: 26496)
    AFXASSUME(pDoc != nullptr);
    if (pDoc->IsKindOf(RUNTIME_CLASS(CPyDebugDoc)))
    {
#pragma warning(suppress: 26429 26466)
      auto pDebugDoc{static_cast<CPyDebugDoc*>(pDoc)};
      ASSERT(pDebugDoc != nullptr);
      pDebugDoc->DisableAllBreakpoints();
    }
  }

  //Update the breakpoints window
  UpdateBreakpointsWindow();
}

void CPyDebugFrameWnd::OnDebugToggleBreakpoint()
{
  //Validate our parameters
#pragma warning(suppress: 26429)
  CPyDebugDoc* pDoc{FindActiveDebuggerDocument()};
  ASSERT(pDoc != nullptr);
#pragma warning(suppress: 26429)
  auto* pView{pDoc->GetView()}; //NOLINT(clang-analyzer-core.CallAndMessage)
  ASSERT(pView != nullptr);

  auto& rCtrl{pView->GetCtrl()};
  const Sci_Position nPos{rCtrl.GetCurrentPos()};
  CString sThisModule{pDoc->GetPathName()};
  const auto line{rCtrl.LineFromPosition(nPos)};
  if (rCtrl.MarkerGet(line) & 0x3) //We are only looking for enabled or disabled breakpoints
  {
    {
      CSingleLock sl{&m_csBreakpoints, TRUE};

      //Remove from in-memory
      m_Breakpoints.erase(std::remove_if(m_Breakpoints.begin(), m_Breakpoints.end(), [line, &sThisModule](const CPyDebugBreakpoint& breakpoint) noexcept
        {
          return ((breakpoint.m_LineNumber == (line + 1)) && (breakpoint.m_sModule == sThisModule));
        }), m_Breakpoints.end());
    }

    //Remove from the UI
    rCtrl.MarkerDelete(line, SC_PYTHON_MARK_ENABLED_BREAKPOINT);
    rCtrl.MarkerDelete(line, SC_PYTHON_MARK_DISABLED_BREAKPOINT);

    //Update the breakpoints window
    UpdateBreakpointsWindow();
  }
  else
  {
    //Add to the in-memory array
    CPyDebugBreakpoint breakpoint;
    const DEBUG_STATE eDebugState{m_DebugState};
    if ((eDebugState == CPyDebugFrameWnd::DEBUG_STATE::NOT_DEBUGGING) || (eDebugState == CPyDebugFrameWnd::DEBUG_STATE::DEBUG_THREAD_EXITED))
      breakpoint.m_bEnabled = true;
    else
      breakpoint.m_bEnabled = !pDoc->IsModified();
    breakpoint.m_LineNumber = line + 1;
    breakpoint.m_sModule = sThisModule;

    {
      CSingleLock sl{&m_csBreakpoints, TRUE};
      m_Breakpoints.push_back(breakpoint);
    }

    //Add to the UI
    rCtrl.MarkerAdd(line, breakpoint.m_bEnabled ? SC_PYTHON_MARK_ENABLED_BREAKPOINT : SC_PYTHON_MARK_DISABLED_BREAKPOINT);

    //Update the breakpoints window
    UpdateBreakpointsWindow();
  }
}

#pragma warning(suppress: 26429)
void CPyDebugFrameWnd::OnUpdateDebugToggleBreakpoint(_In_ CCmdUI* pCmdUI)
{
  //Get the active document
  const CPyDebugDoc* pDoc{FindActiveDebuggerDocument()};
  if (pDoc == nullptr)
  {
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
    return;
  }

  //Validate our parameters
#pragma warning(suppress: 26429)
  auto* pView{pDoc->GetView()};
  ASSERT(pView != nullptr);

  const auto& rCtrl{pView->GetCtrl()};
#pragma warning(suppress: 26486)
  pCmdUI->Enable(rCtrl.GetWindowTextLength() != 0);
}

void CPyDebugFrameWnd::OnDebugDeleteAllBreakpoints()
{
  //Remove from in-memory
  {
    CSingleLock sl{&m_csBreakpoints, TRUE};
    m_Breakpoints.clear();
  }

  //Ask each open document to disable all its breakpoints
#pragma warning(suppress: 26496)
  AFXASSUME(m_pDocTemplate != nullptr); //Did you forget to call CPyDebugFrameWnd::SetDocTemplate
  POSITION posDoc{m_pDocTemplate->GetFirstDocPosition()};
  while (posDoc != nullptr)
  {
    CDocument* pDoc{m_pDocTemplate->GetNextDoc(posDoc)};
#pragma warning(suppress: 26496)
    AFXASSUME(pDoc != nullptr);
    if (pDoc->IsKindOf(RUNTIME_CLASS(CPyDebugDoc)))
    {
#pragma warning(suppress: 26429 26466)
      auto pDebugDoc{static_cast<CPyDebugDoc*>(pDoc)};
      ASSERT(pDebugDoc != nullptr);
      pDebugDoc->DeleteAllBreakpoints();
    }
  }

  //Update the breakpoints window
  UpdateBreakpointsWindow();
} 

void CPyDebugFrameWnd::OnDebugNewBreakpoint()
{
  //Delegate to the toggle breakpoint method
  OnDebugToggleBreakpoint();
}

#pragma warning(suppress: 26429)
void CPyDebugFrameWnd::OnUpdateDebugNewBreakpoint(_In_ CCmdUI* pCmdUI)
{
  //Get the active document
  const CPyDebugDoc* pDoc{FindActiveDebuggerDocument()};
  if (pDoc == nullptr)
  {
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
    return;
  }

  //Validate our parameters
#pragma warning(suppress: 26429)
  auto* pView{pDoc->GetView()};
  ASSERT(pView != nullptr);

  //Only allow creating a new breakpoint if one does not already exist on the line
  auto& rCtrl{pView->GetCtrl()};
  const Sci_Position nPos{rCtrl.GetCurrentPos()};
  CString sThisModule{pDoc->GetPathName()};
  const auto line{rCtrl.LineFromPosition(nPos)};
#pragma warning(suppress: 26486)
  pCmdUI->Enable((rCtrl.GetWindowTextLength() != 0) && rCtrl.MarkerGet(line) == 0);
}

#pragma warning(suppress: 26429)
void CPyDebugFrameWnd::OnUpdateDebugBreakAll(_In_ CCmdUI* pCmdUI)
{
  const DEBUG_STATE eDebugState{m_DebugState};
#pragma warning(suppress: 26486)
  pCmdUI->Enable((eDebugState == DEBUG_STATE::RUNNING) || (eDebugState == DEBUG_STATE::RUNNING_REQUEST_STEP_OVER) || (eDebugState == DEBUG_STATE::RUNNING_REQUEST_STEP_OUT));
}

#pragma warning(suppress: 26440)
void CPyDebugFrameWnd::OnDebugBreakAll()
{
  m_DebugState = DEBUG_STATE::REQUEST_BREAK_ALL;
}

void CPyDebugFrameWnd::OnDebugRestart()
{
  OnDebugStop();
  OnDebugStart();
}

#pragma warning(suppress: 26429)
void CPyDebugFrameWnd::OnUpdateDebugRestart(_In_ CCmdUI* pCmdUI)
{
  const DEBUG_STATE eDebugState{m_DebugState};
#pragma warning(suppress: 26486)
  pCmdUI->Enable((eDebugState != DEBUG_STATE::NOT_DEBUGGING) && (eDebugState != DEBUG_STATE::DEBUG_THREAD_EXITED));
}

void CPyDebugFrameWnd::OnDebugStepInto()
{
  DEBUG_STATE eExpected{DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL};
  if (!m_DebugState.compare_exchange_strong(eExpected, DEBUG_STATE::REQUEST_STEP_INTO))
      StartDebug(START_DEBUGGING_TYPE::BREAK_ON_FIRST_STATEMENT);
}

#pragma warning(suppress: 26429)
void CPyDebugFrameWnd::OnUpdateDebugStepInto(_In_ CCmdUI* pCmdUI)
{
  const DEBUG_STATE eDebugState{m_DebugState};
  if ((eDebugState == DEBUG_STATE::DEBUG_THREAD_EXITED) || (eDebugState == DEBUG_STATE::NOT_DEBUGGING))
  {
    //Get the active document
    const CPyDebugDoc* pDoc{FindActiveDebuggerDocument()};
    if (pDoc != nullptr)
    {
      //Allow debugging so long as their is some code and we are not debugging already, the debugging thread has exit or we are currently break all in the debugger
#pragma warning(suppress: 26429)
      auto* pView{pDoc->GetView()};
      ASSERT(pView != nullptr);
#pragma warning(suppress: 26486)
      pCmdUI->Enable(pView->GetCtrl().GetWindowTextLength() != 0);
    }
    else
#pragma warning(suppress: 26486)
      pCmdUI->Enable(FALSE);
  }
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(eDebugState == DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL);
}

void CPyDebugFrameWnd::OnDebugStepOver()
{
  if (m_DebugState == DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL) //NOLINT(clang-analyzer-cplusplus.NewDelete)
  {
    //Hive away the current stack frame before we request the step over
    CSingleLock sl{&m_csStackFrames, TRUE};
    if (m_StackFrames.size())
#pragma warning(suppress: 26446 26486)
        m_StackFrameAtRequestStepOver = m_StackFrames[0];
    else
        m_StackFrameAtRequestStepOver = CPyDebugStackFrame();

    m_DebugState = DEBUG_STATE::REQUEST_STEP_OVER;
  }
  else
    StartDebug(START_DEBUGGING_TYPE::BREAK_ON_FIRST_STATEMENT);
}

#pragma warning(suppress: 26429)
void CPyDebugFrameWnd::OnUpdateDebugStepOver(_In_ CCmdUI* pCmdUI)
{
  const DEBUG_STATE eDebugState{m_DebugState};
  if ((eDebugState == DEBUG_STATE::DEBUG_THREAD_EXITED) || (eDebugState == DEBUG_STATE::NOT_DEBUGGING))
  {
    //Get the active document
    const CPyDebugDoc* pDoc{FindActiveDebuggerDocument()};
    if (pDoc != nullptr)
    {
      //Allow debugging so long as their is some code and we are not debugging already, the debugging thread has exit or we are currently break all in the debugger
#pragma warning(suppress: 26429)
      auto* pView{pDoc->GetView()};
      ASSERT(pView != nullptr);
#pragma warning(suppress: 26486)
      pCmdUI->Enable(pView->GetCtrl().GetWindowTextLength() != 0);
    }
    else
#pragma warning(suppress: 26486)
      pCmdUI->Enable(FALSE);
  }
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(eDebugState == DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL);
}

void CPyDebugFrameWnd::DeleteBreakpoint(_In_ int nBreakpointIndex)
{
  //Pull out the requested breakpoint to delete
  CSingleLock sl{&m_csBreakpoints, TRUE};
#pragma warning(suppress: 26446)
  const CPyDebugBreakpoint& breakpoint{m_Breakpoints[nBreakpointIndex]};

  //Find the correct document and ask it to delete the breakpoint
  CPyDebugDoc* pDoc{FindDebuggerDocument(breakpoint.m_sModule, false)};
  if (pDoc != nullptr)
    pDoc->DeleteBreakpoint(breakpoint);

  //Remove from the in-memory array
  m_Breakpoints.erase(m_Breakpoints.cbegin() + nBreakpointIndex);

  //Update the Breakpoints window
  m_wndBreakpoints.m_wndCtrl.DeleteItem(nBreakpointIndex);
}

void CPyDebugFrameWnd::UpdateBreakpoint(_In_ int nBreakpointIndex, _In_ bool bEnabled)
{
  //Pull out the requested breakpoint to delete
  CSingleLock sl{&m_csBreakpoints, TRUE};
#pragma warning(suppress: 26446)
  CPyDebugBreakpoint& breakpoint{m_Breakpoints[nBreakpointIndex]};

  if (bEnabled != breakpoint.m_bEnabled)
  {
    //Update the in-memory copy
    breakpoint.m_bEnabled = bEnabled;

    //Find the correct document and ask it to update the breakpoint
    CPyDebugDoc* pDoc{FindDebuggerDocument(breakpoint.m_sModule, false)};
    if (pDoc != nullptr)
    {
      //Force the breakpoint to disabled if we are currently debugging and the document is modified
      const DEBUG_STATE eDebugState{m_DebugState};
      if (breakpoint.m_bEnabled && (eDebugState != CPyDebugFrameWnd::DEBUG_STATE::NOT_DEBUGGING) && (eDebugState != CPyDebugFrameWnd::DEBUG_STATE::DEBUG_THREAD_EXITED) && breakpoint.m_bEnabled && pDoc->IsModified())
        breakpoint.m_bEnabled = false;

      pDoc->UpdateBreakpoint(breakpoint);
    }

    //Update the breakpoints window
    LVITEM lvi{};
    lvi.mask = LVIF_IMAGE;
    lvi.iItem = nBreakpointIndex;
    lvi.iImage = breakpoint.m_bEnabled ? 0 : 2;
    m_wndBreakpoints.m_wndCtrl.SetItem(&lvi);
  }
}

void CPyDebugFrameWnd::OnDebugStepOut()
{
  //Hive away the current stack frame before we request the step over
  {
    CSingleLock sl{&m_csStackFrames, TRUE};
    if (m_StackFrames.size())
#pragma warning(suppress: 26446 26486)
      m_StackFrameAtRequestStepOut = m_StackFrames[0];
    else
      m_StackFrameAtRequestStepOut = CPyDebugStackFrame{};
  }

  m_DebugState = DEBUG_STATE::REQUEST_STEP_OUT;
}

#pragma warning(suppress: 26429)
void CPyDebugFrameWnd::OnUpdateDebugStepOut(_In_ CCmdUI* pCmdUI)
{
#pragma warning(suppress: 26486)
  pCmdUI->Enable(m_DebugState == DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL);
}

void CPyDebugFrameWnd::UpdateUIForCodeContinue()
{
  const CPyDebugDoc* pDoc{FindActiveDebuggerDocument()};

  //Remove any IP marker
  if (pDoc != nullptr)
  {
#pragma warning(suppress: 26429)
    auto* pView{pDoc->GetView()};
    ASSERT(pView != nullptr);
    auto& rCtrl{pView->GetCtrl()};
    rCtrl.MarkerDeleteAll(SC_PYTHON_MARK_INSTRUCTION_POINTER);
  }

  //Reset the locals window
  m_wndLocals.m_wndCtrl.EnableWindow(FALSE);

  //Reset the globals window
  m_wndGlobals.m_wndCtrl.EnableWindow(FALSE);

  //Reset the BuiltIns window
  m_wndBuiltIns.m_wndCtrl.EnableWindow(FALSE);

  //Reset the Call Stack window
  LVITEM lvi{};
  lvi.mask = LVIF_IMAGE;
  lvi.iItem = 0;
  lvi.iImage = 1;
  m_wndCallStack.m_wndCtrl.SetItem(&lvi);
  m_wndCallStack.m_wndCtrl.EnableWindow(FALSE);

  //Reset the Modules window
  m_wndModules.m_wndCtrl.EnableWindow(FALSE);

  //Reset the breakpoints window
  const int nBreakpoints = m_wndBreakpoints.m_wndCtrl.GetItemCount();
  for (int i{0}; i<nBreakpoints; i++)
    m_wndBreakpoints.m_wndCtrl.SetItemData(i, 0);
  m_wndBreakpoints.m_wndCtrl.UpdateWindow();

  //Reset the Threads window
  m_wndThreads.m_wndCtrl.EnableWindow(FALSE);

  //Reset the Watch window
  m_wndWatch.m_wndCtrl.ResetChangedStateForAllVariables();
}

BOOL CPyDebugFrameWnd::OnCommand(_In_ WPARAM wParam, _In_ LPARAM lParam)
{
  //Validate our parameters
  ASSERT(m_wndOutput.m_pWndView != nullptr);

  //Let the output window handle some specific messages if it is the window in focus
  const CWnd* pWnd{GetFocus()};
  if (pWnd != nullptr)
  {
    const bool bHandleInOutputWindow{((m_wndOutput.m_pWndView->GetCtrl().GetSafeHwnd() == pWnd->GetSafeHwnd()) && 
                                      (LOWORD(wParam) == ID_EDIT_FIND || LOWORD(wParam) == ID_EDIT_SELECT_ALL ||
                                       LOWORD(wParam) == ID_EDIT_REPEAT || LOWORD(wParam) == ID_EDIT_COPY || LOWORD(wParam) == ID_EDIT_CLEAR))};
    if (bHandleInOutputWindow)
    {
      AfxCallWndProc(m_wndOutput.m_pWndView, m_wndOutput.m_pWndView->m_hWnd, WM_COMMAND, wParam, lParam);
      return TRUE;
    }
  }

  //Let the base class do its thing
  return __super::OnCommand(wParam, lParam);
}

void CPyDebugFrameWnd::OnWindowManager()
{
  ShowWindowsDialog();
}

void CPyDebugFrameWnd::OnFileOpen()
{
  //Validate our parameters
#pragma warning(suppress: 26429)
  CWinApp* pApp{AfxGetApp()};
  ASSERT(pApp != nullptr);
  ASSERT(pApp->m_pDocManager != nullptr);
  ASSERT(m_pDocTemplate != nullptr);

  //prompt the user (with the specific doc template)
  CString newName;
  if (!pApp->m_pDocManager->DoPromptFileName(newName, AFX_IDS_OPENFILE, OFN_HIDEREADONLY | OFN_FILEMUSTEXIST, TRUE, m_pDocTemplate))
    return;

  m_pDocTemplate->OpenDocumentFile(newName, TRUE);
}

void CPyDebugFrameWnd::OnFileNew()
{
  //Validate our parameters
  ASSERT(m_pDocTemplate != nullptr);

  m_pDocTemplate->OpenDocumentFile(nullptr);
}
