/*
Module : PyDebugMainFrm.h
Purpose: Defines the interface for the classes which encapsulate the main frame window functionality of PyDebug++
Created: PJN / 20-08-2015

Copyright (c) 2015 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef __PYDEBUGMAINFRM_H__
#define __PYDEBUGMAINFRM_H__

#ifndef CPYDEBUG_EXT_CLASS
#define CPYDEBUG_EXT_CLASS
#endif //#ifndef CPYDEBUG_EXT_CLASS

constexpr static int SC_PYTHON_MARK_ENABLED_BREAKPOINT = 0;
constexpr static int SC_PYTHON_MARK_DISABLED_BREAKPOINT = 1;
constexpr static int SC_PYTHON_MARK_INSTRUCTION_POINTER = 2;


//////////////////// Includes /////////////////////////////////////////////////

#include "PyDebugOutputWnd.h"
#include "PyDebugPropertiesWnd.h"
#include "PyDebugCallStackWnd.h"
#include "PyDebugModulesWnd.h"
#include "PyDebugBreakpointsWnd.h"
#include "PyDebugThreadsWnd.h"
#include "PyDebugVariablesWnd.h"
#include "PythonPtr.h"
#include "PyDebugModule.h"
#include "PyDebugBreakpoint.h"
#include "PyDebugStackFrame.h"
#include "PyDebugThread.h"
#include "PyDebugDocTemplate.h"


//////////////////// Classes //////////////////////////////////////////////////

class CPyDebugDoc;

class CPYDEBUG_EXT_CLASS CPyDebugLessThanPyVariable
{
public:
  inline bool operator() (_In_ const CPyDebugVariable& var1, _In_ const CPyDebugVariable& var2) noexcept
  {
    return (var1.m_sName < var2.m_sName);
  }
};


class CPYDEBUG_EXT_CLASS CPyDebugLessThanPyModule
{
public:
  inline bool operator() (_In_ const CPyDebugModule& module1, _In_ const CPyDebugModule& module2) noexcept
  {
    return (module1.m_sName < module2.m_sName);
  }
};


class CPYDEBUG_EXT_CLASS CPyDebugFrameWnd : public CMDIFrameWndEx
{
public:
//Enums
  enum class DEBUG_STATE
  {
    NOT_DEBUGGING,
    RUNNING,
    REQUEST_STOP,
    ACKNOWLEDGE_STOP,
    REQUEST_BREAK_ALL,
    ACKNOWLEDGE_BREAK_ALL,
    REQUEST_CONTINUE,
    DEBUG_THREAD_EXITED,
    REQUEST_STEP_INTO,
    REQUEST_STEP_OVER,
    REQUEST_STEP_OUT,
    RUNNING_REQUEST_STEP_OVER,
    RUNNING_REQUEST_STEP_OUT,
  };

  enum class VARIABLES_WND_REQUEST
  {
    NO_REQUEST,
    CHANGE_VARIABLE_NAME,
    CHANGE_VARIABLE_VALUE,
    EXPAND_VARIABLE
  };

//Constructors / Destructors
#pragma warning(suppress: 26455)
  CPyDebugFrameWnd() = default;
  CPyDebugFrameWnd(const CPyDebugFrameWnd&) = delete;
  CPyDebugFrameWnd(CPyDebugFrameWnd&&) = delete;
  ~CPyDebugFrameWnd(); //NOLINT(modernize-use-override)

//Methods
  CPyDebugFrameWnd& operator=(const CPyDebugFrameWnd&) = delete;
  CPyDebugFrameWnd& operator=(CPyDebugFrameWnd&&) = delete;
  void SetDocTemplate(_In_opt_ CPyDebugMultiDocTemplate* pTemplate) noexcept { m_pDocTemplate = pTemplate; };
  bool GetHiColorIcons() noexcept { return m_bHiColorIcons; }
  void DeleteBreakpoint(_In_ int nBreakpointIndex);
  void UpdateBreakpoint(_In_ int nBreakpointIndex, _In_ bool bEnabled);
  void NavigateToDocumentLine(_In_opt_z_ LPCTSTR pszFilename, _In_ Scintilla::Line lineNumber, _In_ bool bAllowOpen);
  _NODISCARD DEBUG_STATE GetDebugState() const noexcept { return m_DebugState; };

protected:
//Enums
  enum class START_DEBUGGING_TYPE
  {
    NOT_DEFINED,
    NORMAL_START,
    BREAK_ON_FIRST_STATEMENT
  };

//Member variables
  CMFCMenuBar m_wndMenuBar;
  CMFCToolBar m_wndStandardToolBar;
  CMFCToolBar m_wndDebugToolBar;
  CMFCStatusBar m_wndStatusBar;
  CPyDebugOutputWnd m_wndOutput;
  std::vector<CString> m_pendingOutputText;
  CCriticalSection m_csOutputText;
  CPyDebugVariablesWnd m_wndGlobals;
  CPyDebugVariablesWnd m_wndLocals;
  CPyDebugVariablesWnd m_wndBuiltIns;
  CPyDebugCallStackWnd m_wndCallStack;
  CPyDebugModulesWnd m_wndModules;
  CPyDebugBreakpointsWnd m_wndBreakpoints;
  CPyDebugThreadsWnd m_wndThreads;
  CPyDebugVariablesWnd m_wndWatch;
  CPyDebugPropertiesWnd m_wndProperties;
  bool m_bHiColorIcons{true};
  CPyDebugMultiDocTemplate* m_pDocTemplate{nullptr};
  CWinThread* m_pDebugThread{nullptr};
  CStringW m_sCode;
  CString m_sModuleName;
  std::atomic<DEBUG_STATE> m_DebugState{DEBUG_STATE::NOT_DEBUGGING};
  std::atomic<START_DEBUGGING_TYPE> m_StartDebuggingType{START_DEBUGGING_TYPE::NOT_DEFINED};
  CCriticalSection m_csBreakpoints;
  std::vector<CPyDebugBreakpoint> m_Breakpoints;
  CCriticalSection m_csStackFrames;
  std::deque<CPyDebugStackFrame> m_StackFrames;
  CPyDebugStackFrame m_StackFrameAtRequestStepOver;
  CPyDebugStackFrame m_StackFrameAtRequestStepOut;
  CCriticalSection m_csModules;
  std::vector<CPyDebugModule> m_Modules;
  CCriticalSection m_csThreads;
  std::vector<CPyDebugThread> m_Threads;
  bool m_bConsoleAllocated{false};
  DWORD m_dwGUIThreadID{0};
  std::atomic<DWORD> m_dwDebuggedThreadID{0};
  std::atomic<VARIABLES_WND_REQUEST> m_VariablesWndRequest{VARIABLES_WND_REQUEST::NO_REQUEST};
  int m_nVariablesWndRequestIndex{-1};
  CPyDebugVariablesListCtrl::TYPE m_VariablesWndRequestType{CPyDebugVariablesListCtrl::TYPE::NONE};
  CString m_sVariablesWndRequestString;
  CPyObject m_pOriginalStdOut;
  CPyObject m_pOriginalStdIn;
  CPyObject m_pOriginalStdErr;

//Methods
  static UINT _DebugThread(_In_opt_ LPVOID pParam);
  static int _TraceCallback(_In_opt_ PyObject* obj, _In_opt_ PyFrameObject* frame, _In_ int what, _In_opt_ PyObject* arg);
  static CString PyToCString(_In_opt_ PyObject* pObject);
  virtual CString GetPythonErrorString();
  virtual void AddOutputText(_In_z_ LPCTSTR pszText);
  virtual void AddOutputText(_In_ UINT nResourceID);
  virtual void AddOutputTextU(_In_ UINT nResourceID, _In_ UINT nValue);
  virtual void AddOutputTextI(_In_ UINT nResourceID, _In_ int nValue);
  virtual void AddOutputText(_In_ UINT nResourceID, _In_z_ LPCTSTR pszValue);
  virtual void AddOutputText(_In_ UINT nResourceID, _In_z_ LPCTSTR pszValue1, _In_z_ LPCTSTR pszValue2);
  virtual void AddOutputText(_In_ UINT nResourceID, _In_ int nValue1, _In_z_ LPCTSTR pszValue2, _In_z_ LPCTSTR pszValue3);
  virtual void AddOutputText(_In_ UINT nResourceID, _In_z_ LPCTSTR pszValue1, _In_ int nValue2, _In_z_ LPCTSTR pszValue3, _In_z_ LPCTSTR pszValue4);
  virtual BOOL CreateDockingWindows();
  virtual BOOL CreateOutputDockingWindow();
  virtual BOOL CreatePropertiesDockingWindow();
  virtual BOOL CreateGlobalsDockingWindow();
  virtual BOOL CreateLocalsDockingWindow();
  virtual BOOL CreateBuiltInsDockingWindow();
  virtual BOOL CreateCallStackDockingWindow();
  virtual BOOL CreateModulesDockingWindow();
  virtual BOOL CreateBreakpointsDockingWindow();
  virtual BOOL CreateThreadsDockingWindow();
  virtual BOOL CreateWatchDockingWindow();
  BOOL PreTranslateMessage(MSG* pMsg) override;
  virtual CPyDebugDoc* FindActiveDebuggerDocument();
  virtual CPyDebugDoc* FindDebuggerDocument(_In_ const CString& sFilename, _In_ bool bAllowOpen);
  virtual bool RedirectIO(_In_opt_ PyObject* pIOModule, _In_opt_ PyObject* pSysModule, _In_opt_z_ LPCSTR pszPythonName, _In_opt_z_ LPCSTR pszName, _In_opt_z_ LPCSTR pszMode, _In_ CPyObject& fileObject, _Inout_ CPyObject& originalIO);
  virtual void ResetIO(_In_opt_ PyObject* pSysModule);
  virtual BOOL ResetIO(_In_opt_ PyObject* pSysModule, _In_opt_z_ LPCSTR pszPythonName, _In_opt_ PyObject* pOriginalObject);
  virtual void ClosePythonFile(_In_opt_ PyObject* pFile);
  virtual int TraceCallback(_In_opt_ PyObject* obj, _In_opt_ PyFrameObject* frame, _In_ int what, _In_opt_ PyObject* arg);
  virtual void HandleBreakpointsForTraceCallback(_In_ PyFrameObject* frame, _Out_ bool& bBreakAll);
  virtual void HandleBreakpointsForTraceCallback2(_In_ PyFrameObject* frame);
  virtual void HandleVariableNameChangeForTraceCallback();
  virtual void HandleVariableValueChangeForTraceCallback();
  virtual void HandleVariableExpansionForTraceCallback(_In_ CPyDebugVariablesListCtrl& variablesWnd);
  virtual void HandleInstructionPointerUpdate();
  virtual void StartDebug(_In_ START_DEBUGGING_TYPE eType);
  virtual bool DebugCode(_In_ const CStringW& sCode, _In_ const CString& sModuleName, _In_ START_DEBUGGING_TYPE eType);
  virtual UINT DebugThread();
  virtual void CaptureModulesDetails();
  virtual void CaptureThreadsDetails(_In_ DWORD dwThreadIdBeingHit);
  virtual void UpdateBreakpointsWindow();
  virtual void UpdateCallStackWindow();
  virtual void CaptureStackFrame(_In_ PyFrameObject* frame, _Out_ CPyDebugStackFrame& CapturedFrame);
  virtual void UpdateUIForCodeContinue();
  virtual void CloseIOFiles(_In_ PyObject* pStdErr, _In_ PyObject* pStdOut, _In_ PyObject* pStdIn);
  virtual void RemoveAllInstructionPointers();
  BOOL OnCommand(_In_ WPARAM wParam, _In_ LPARAM lParam) override;
  BOOL GetToolbarButtonToolTipText(CMFCToolBarButton* pButton, CString& strTTText) override;
  void GetMessageString(UINT nID, CString& rMessage) const override;
  virtual BOOL SetupRedirectIO(_In_opt_ PyObject* pSysModule, _In_ CPyObject& pStdOut, _In_ CPyObject& pStdIn, _In_ CPyObject& pStdErr);
  virtual BOOL SetupCodeObject(_In_ CPyObject& pBytes);

//Message Handlers
  afx_msg int OnCreate(_In_ LPCREATESTRUCT lpCreateStruct);
  afx_msg void OnViewCustomize();
  afx_msg LRESULT OnToolbarCreateNew(_In_ WPARAM wParam, _In_ LPARAM lParam);
  afx_msg LRESULT HandlePendingOutputText(_In_ WPARAM wParam, _In_ LPARAM lParam);
  afx_msg void OnSysKeyDown(_In_ UINT nChar, _In_ UINT nRepCnt, _In_ UINT nFlags);
  afx_msg BOOL OnPaneCheck(_In_ UINT nID);
  afx_msg void OnDebugStart();
  afx_msg void OnUpdateDebugStart(_In_ CCmdUI* pCmdUI);
  afx_msg void OnDebugStop();
  afx_msg void OnUpdateDebugStop(_In_ CCmdUI* pCmdUI);
  afx_msg void OnDebugToggleBreakpoint();
  afx_msg void OnUpdateDebugToggleBreakpoint(_In_ CCmdUI* pCmdUI);
  afx_msg void OnDebugDeleteAllBreakpoints();
  afx_msg void OnDebugNewBreakpoint();
  afx_msg void OnUpdateDebugNewBreakpoint(_In_ CCmdUI* pCmdUI);
  afx_msg void OnDebugDisableAllBreakpoints();
  afx_msg void OnUpdateDebugBreakAll(_In_ CCmdUI* pCmdUI);
  afx_msg void OnDebugBreakAll();
  afx_msg void OnDebugRestart();
  afx_msg void OnUpdateDebugRestart(_In_ CCmdUI* pCmdUI);
  afx_msg void OnDebugStepInto();
  afx_msg void OnUpdateDebugStepInto(_In_ CCmdUI* pCmdUI);
  afx_msg void OnDebugStepOver();
  afx_msg void OnUpdateDebugStepOver(_In_ CCmdUI* pCmdUI);
  afx_msg void OnDebugStepOut();
  afx_msg void OnUpdateDebugStepOut(_In_ CCmdUI* pCmdUI);
  afx_msg LRESULT OnDebugNotification(_In_ WPARAM wParam, _In_ LPARAM lParam);
  afx_msg void OnWindowManager();
  afx_msg void OnFileOpen();
  afx_msg void OnFileNew();

  DECLARE_MESSAGE_MAP()

  DECLARE_DYNCREATE(CPyDebugFrameWnd)

  friend class CPyDebugDoc;
  friend class CPyDebugCallStackListCtrl;
  friend class CPyDebugBreakpointsListCtrl;
  friend class CPyDebugVariablesListCtrl;
};


#endif //#ifndef __PYDEBUGMAINFRM_H__
