/*
Module : PyDebugPropertiesWnd.cpp
Purpose: Defines the implementation for the classes which encapsulate the Properties window functionality of PyDebug++
Created: PJN / 20-08-2015
History: PJN / 01-08-2022 1. Updated all the code in to use C++ uniform initialization for all variable
                          declarations.

Copyright (c) 2015 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "PyDebugPropertiesWnd.h"
#include "resource.h"
#include "PyDebugMainFrm.h"


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CPyDebugPropertiesWnd, CDockablePane) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_CREATE()
  ON_WM_SIZE()
  ON_COMMAND(ID_PYDEBUG_EXPAND_ALL, &CPyDebugPropertiesWnd::OnExpandAllProperties)
  ON_COMMAND(ID_PYDEBUG_SORTPROPERTIES, &CPyDebugPropertiesWnd::OnSortProperties)
  ON_UPDATE_COMMAND_UI(ID_PYDEBUG_SORTPROPERTIES, &CPyDebugPropertiesWnd::OnUpdateSortProperties)
  ON_WM_SETFOCUS()
  ON_WM_SETTINGCHANGE()
END_MESSAGE_MAP()

void CPyDebugPropertiesWnd::AdjustLayout()
{
  if (GetSafeHwnd() == nullptr)
    return;

  CRect rectClient;
  GetClientRect(rectClient);
  const int cyTlb{m_wndToolBar.CalcFixedLayout(FALSE, TRUE).cy};
  m_wndToolBar.SetWindowPos(nullptr, rectClient.left, rectClient.top, rectClient.Width(), cyTlb, SWP_NOACTIVATE | SWP_NOZORDER);
  m_wndPropList.SetWindowPos(nullptr, rectClient.left, rectClient.top + cyTlb, rectClient.Width(), rectClient.Height() - cyTlb, SWP_NOACTIVATE | SWP_NOZORDER);
}

#pragma warning(suppress: 26434)
int CPyDebugPropertiesWnd::OnCreate(_In_ LPCREATESTRUCT lpCreateStruct)
{
  //Let the base class do its thing
  if (__super::OnCreate(lpCreateStruct) == -1)
    return -1;

  //Create the properties grid
  CRect rectDummy;
  rectDummy.SetRectEmpty();
  if (!m_wndPropList.Create(WS_VISIBLE | WS_CHILD, rectDummy, this, 2))
  {
    TRACE(_T("Failed to create properties grid\n"));
    return -1;
  }
  InitPropList();

  //Create the toolbar
  m_wndToolBar.Create(this, AFX_DEFAULT_TOOLBAR_STYLE, IDR_PYDEBUG_PROPERTIES);
  m_wndToolBar.LoadToolBar(IDR_PYDEBUG_PROPERTIES, 0, 0, TRUE /* Is locked */);
  m_wndToolBar.CleanUpLockedImages();
  ASSERT(m_pMainFrame != nullptr); //Did you forget to call SetMainFrame
  const bool bHiColorIcons{m_pMainFrame->GetHiColorIcons()};
  m_wndToolBar.LoadBitmap(bHiColorIcons ? IDB_PYDEBUG_PROPERTIES_HC : IDR_PYDEBUG_PROPERTIES, 0, 0, TRUE);
  m_wndToolBar.SetPaneStyle(m_wndToolBar.GetPaneStyle() | CBRS_TOOLTIPS | CBRS_FLYBY);
  m_wndToolBar.SetPaneStyle(m_wndToolBar.GetPaneStyle() & ~(CBRS_GRIPPER | CBRS_SIZE_DYNAMIC | CBRS_BORDER_TOP | CBRS_BORDER_BOTTOM | CBRS_BORDER_LEFT | CBRS_BORDER_RIGHT));
  m_wndToolBar.SetOwner(this);
  m_wndToolBar.SetRouteCommandsViaFrame(FALSE); //All commands will be routed via this control, not via the parent frame

  AdjustLayout();
  return 0;
}

#pragma warning(suppress: 26434)
void CPyDebugPropertiesWnd::OnSize(_In_ UINT nType, _In_ int cx, _In_ int cy)
{
  //Let the base class do its thing
  __super::OnSize(nType, cx, cy);

  AdjustLayout();
}

void CPyDebugPropertiesWnd::OnExpandAllProperties()
{
  m_wndPropList.ExpandAll();
}

void CPyDebugPropertiesWnd::OnSortProperties()
{
  m_wndPropList.SetAlphabeticMode(!m_wndPropList.IsAlphabeticMode());
}

#pragma warning(suppress: 26429)
void CPyDebugPropertiesWnd::OnUpdateSortProperties(_In_ CCmdUI* pCmdUI)
{
#pragma warning(suppress: 26486)
  pCmdUI->SetCheck(m_wndPropList.IsAlphabeticMode());
}

void CPyDebugPropertiesWnd::InitPropList()
{
  SetPropListFont();
  m_wndPropList.EnableHeaderCtrl(FALSE);
  m_wndPropList.EnableDescriptionArea();
  m_wndPropList.SetVSDotNetLook();
  m_wndPropList.MarkModifiedProperties();
  CString sGroup;
#pragma warning(suppress: 26496)
  ENSURE(sGroup.LoadString(IDS_PYDEBUG_GENERAL));
#pragma warning(suppress: 26400 26403 26409)
  auto pGroup1{new CMFCPropertyGridProperty{sGroup}};
  ASSERT(m_pAllocateConsole == nullptr);
  CString sProperty;
#pragma warning(suppress: 26496)
  ENSURE(sProperty.LoadString(IDS_PYDEBUG_ALLOCATE_CONSOLE));
  CString sDescription;
#pragma warning(suppress: 26496)
  ENSURE(sDescription.LoadString(IDS_PYDEBUG_ALLOCATE_CONSOLE_DESCRIPTION));
#pragma warning(suppress: 26409)
  m_pAllocateConsole = new CMFCPropertyGridProperty{sProperty, _variant_t{false}, sDescription};
  pGroup1->AddSubItem(m_pAllocateConsole);
  m_wndPropList.AddProperty(pGroup1);
}

#pragma warning(suppress: 26434)
void CPyDebugPropertiesWnd::OnSetFocus(_In_opt_ CWnd* pOldWnd)
{
  //Let the base class do its thing
  __super::OnSetFocus(pOldWnd);

  //Give the focus to the control
  m_wndPropList.SetFocus();
}

#pragma warning(suppress: 26434)
void CPyDebugPropertiesWnd::OnSettingChange(_In_ UINT uFlags, _In_opt_z_ LPCTSTR lpszSection)
{
  //Let the base class do its thing
  __super::OnSettingChange(uFlags, lpszSection);

  SetPropListFont();
}

void CPyDebugPropertiesWnd::SetPropListFont()
{
  //Delete the existing font
  ::DeleteObject(m_fntPropList.Detach());

  //Recreate the font
  LOGFONT lf;
  afxGlobalData.fontRegular.GetLogFont(&lf);
  NONCLIENTMETRICS info;
  info.cbSize = sizeof(info);
  afxGlobalData.GetNonClientMetrics(info);
  lf.lfHeight = info.lfMenuFont.lfHeight;
  lf.lfWeight = info.lfMenuFont.lfWeight;
  lf.lfItalic = info.lfMenuFont.lfItalic;
  m_fntPropList.CreateFontIndirect(&lf);

  //Update the property grid
  m_wndPropList.SetFont(&m_fntPropList);
}

bool CPyDebugPropertiesWnd::GetAllocateConsole()
{
  ASSERT(m_pAllocateConsole != nullptr);
  _variant_t varValue{m_pAllocateConsole->GetValue()};
  return varValue.operator BOOL() != FALSE;
}

void CPyDebugPropertiesWnd::SetAllocateConsole(_In_ bool bAllocateConsole)
{
  ASSERT(m_pAllocateConsole != nullptr);
  m_pAllocateConsole->SetValue(_variant_t{bAllocateConsole});
}

void CPyDebugPropertiesWnd::SetPyDebugMainFrame(_In_opt_ CPyDebugFrameWnd* pMainFrame) noexcept
{
  m_pMainFrame = pMainFrame;
}
