/*
Module : PyDebugThreadsWnd.cpp
Purpose: Defines the implementation for the classes which encapsulate the Threads window functionality of PyDebug++
Created: PJN / 20-08-2015
History: PJN / 01-08-2022 1. Updated all the code in to use C++ uniform initialization for all variable
                          declarations.

Copyright (c) 2015 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "PyDebugThreadsWnd.h"
#include "PyDebugMainFrm.h"
#include "resource.h"


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG

constexpr int THREADS_LIST_CTRL_ID = 1;


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CPyDebugThreadsListCtrl, CListCtrl) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_CREATE()
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT_EX(NM_CUSTOMDRAW, &CPyDebugThreadsListCtrl::OnCustomDraw)
END_MESSAGE_MAP()

#pragma warning(suppress: 26434)
int CPyDebugThreadsListCtrl::OnCreate(_In_ LPCREATESTRUCT lpCreateStruct)
{
  //Let the base class do its thing
  if (__super::OnCreate(lpCreateStruct) == -1)
    return -1;

  //Create the bold font we need
  CFont* pFont{GetFont()};
  if (pFont == nullptr)
    return -1;
  LOGFONT lf;
  pFont->GetLogFont(&lf);
  lf.lfWeight = FW_BOLD;
  if (!m_BoldFont.CreateFontIndirect(&lf))
  {
    TRACE(_T("Failed to create Breakpoints bold font\n"));
    return -1;
  }

  return 0;
}

#pragma warning(suppress: 26429)
BOOL CPyDebugThreadsListCtrl::OnCustomDraw(_In_ NMHDR* pNMHDR, _Inout_ LRESULT* pResult) 
{
#pragma warning(suppress: 26490)
  auto pCustomDraw{reinterpret_cast<NMLVCUSTOMDRAW*>(pNMHDR)};
  ASSERT(pCustomDraw != nullptr);

  switch (pCustomDraw->nmcd.dwDrawStage) 
  {
    case CDDS_PREPAINT:
    {
      *pResult = CDRF_NOTIFYITEMDRAW; //Tell the control that we are interested in item notifications
      break;
    }
    case CDDS_ITEMPREPAINT:
    {
      *pResult = CDRF_NOTIFYSUBITEMDRAW;
      break;
    }
    case CDDS_SUBITEM | CDDS_ITEMPREPAINT:
    {
#pragma warning(suppress: 26472)
      const int nItem{static_cast<int>(pCustomDraw->nmcd.dwItemSpec)};
      const DWORD_PTR dwItemData{GetItemData(nItem)};
      if (dwItemData)
      {
        //Select the right font
        SelectObject(pCustomDraw->nmcd.hdc, m_BoldFont);
        *pResult = CDRF_NEWFONT;
      }
      else
        *pResult = CDRF_DODEFAULT;
      break;
    }
    case CDDS_ITEMPOSTPAINT:
    {
      *pResult = CDRF_DODEFAULT;
      break;
    }
    default:
    {
      break;
    }
  }

  return TRUE; //Allow the message to be reflected again
}



#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CPyDebugThreadsWnd, CDockablePane) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_CREATE()
  ON_WM_SIZE()
  ON_WM_SETFOCUS()
END_MESSAGE_MAP()

#pragma warning(suppress: 26434)
int CPyDebugThreadsWnd::OnCreate(_In_ LPCREATESTRUCT lpCreateStruct)
{
  //Let the base class do its thing
  if (__super::OnCreate(lpCreateStruct) == -1)
    return -1;

  //Create the list control
  CRect rectDummy;
  rectDummy.SetRectEmpty();
  if (!m_wndCtrl.Create(WS_CHILD | WS_VISIBLE | LVS_REPORT | LVS_SHOWSELALWAYS, rectDummy, this, THREADS_LIST_CTRL_ID))
  {
    TRACE(_T("Failed to create Threads window\n"));
    return -1;
  }
  m_wndCtrl.SetExtendedStyle(LVS_EX_FULLROWSELECT | LVS_EX_LABELTIP | LVS_EX_GRIDLINES);
  m_wndCtrl.EnableWindow(FALSE);

  //Create the image list
  if (!m_ImageList.Create(IDB_PYDEBUG_THREADS, 16, 1, RGB(255, 0, 255)))
  {
    TRACE(_T("Failed to create Threads window image list\n"));
    return -1;
  }
  m_wndCtrl.SetImageList(&m_ImageList, LVSIL_SMALL);

  //Add the columns to the list control
  CRect rClient;
  GetClientRect(&rClient);
  CString sColumn;
  if (!sColumn.LoadString(IDS_PYDEBUG_THREAD_ID))
  {
    TRACE(_T("Failed to load Threads window column caption\n"));
    return -1;
  }
  m_wndCtrl.InsertColumn(0, sColumn, LVCFMT_LEFT, 100);
  if (!sColumn.LoadString(IDS_PYDEBUG_CATEGORY))
  {
    TRACE(_T("Failed to load Threads window column caption\n"));
    return -1;
  }
  m_wndCtrl.InsertColumn(1, sColumn, LVCFMT_LEFT, 200);
  if (!sColumn.LoadString(IDS_PYDEBUG_PRIORITY))
  {
    TRACE(_T("Failed to load Threads window column caption\n"));
    return -1;
  }
  m_wndCtrl.InsertColumn(2, sColumn, LVCFMT_LEFT, 200);

  return 0;
}

#pragma warning(suppress: 26434)
void CPyDebugThreadsWnd::OnSize(_In_ UINT nType, _In_ int cx, _In_ int cy)
{
  //Let the base class do its thing
  __super::OnSize(nType, cx, cy);

  //Make the control window covers the whole client area
  m_wndCtrl.SetWindowPos(nullptr, -1, -1, cx, cy, SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOZORDER);
}

#pragma warning(suppress: 26434)
void CPyDebugThreadsWnd::OnSetFocus(_In_opt_ CWnd* pOldWnd)
{
  //Let the base class do its thing
  __super::OnSetFocus(pOldWnd);

  //Give the focus to the control
  m_wndCtrl.SetFocus();
}
