/*
Module : PyDebugVariablesWnd.cpp
Purpose: Defines the implementation for the classes which encapsulate the Variables window functionality of PyDebug++
Created: PJN / 20-08-2015
History: PJN / 28-08-2016 1. The watch window now uses the Python Sequence and Mapping protocols instead of the 
                          Iterator protocol (in addition to the existing code which uses a dictionary) to determine if a
                          variable contains sub-variables. This now allows dict objects to be shown correctly in the 
                          various variable windows.
         PJN / 11-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 01-08-2022 1. Updated all the code in to use C++ uniform initialization for all variable declarations.

Copyright (c) 2015 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#ifndef _ALGORITHM_
#pragma message("To avoid this message, please put algorithm in your pre compiled header (normally stdafx.h)")
#include <algorithm>
#endif //#ifndef _ALGORITHM_
#include "PyDebugVariablesWnd.h"
#include "PyDebugMainFrm.h"
#include "resource.h"


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG

constexpr int VARIABLES_LIST_CTRL_EDIT_ID = 2;


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CPyDebugVariablesEdit, CEdit) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_GETDLGCODE()
END_MESSAGE_MAP()

#pragma warning(suppress: 26434)
UINT CPyDebugVariablesEdit::OnGetDlgCode()
{
#pragma warning(suppress: 26472)
  const UINT nChar{static_cast<UINT>(GetCurrentMessage()->wParam)};
  if (nChar == VK_TAB)
  {
    //Tab into the next edit box
#pragma warning(suppress: 26429 26466)
    auto pParent{static_cast<CPyDebugVariablesListCtrl*>(GetParent())};
    ASSERT(pParent != nullptr);
    pParent->CreateValueEditForNextItem();
  }
  else if (nChar == VK_ESCAPE)
  {
    //Close down this edit box without saving
#pragma warning(suppress: 26429 26466)
    auto pParent{static_cast<CPyDebugVariablesListCtrl*>(GetParent())};
    ASSERT(pParent != nullptr);
    pParent->DestroyValueEdit();
  }
  else if (nChar == VK_RETURN)
  {
    //Close down this edit box with saving
#pragma warning(suppress: 26429 26466)
    auto pParent{static_cast<CPyDebugVariablesListCtrl*>(GetParent())};
    ASSERT(pParent != nullptr);
    pParent->UpdateValueFromEdit();
    pParent->DestroyValueEdit();
  }

  return __super::OnGetDlgCode() | DLGC_WANTTAB;
}


#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CPyDebugVariablesListCtrl, CListCtrl) //NOLINT(modernize-avoid-c-arrays)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT(NM_DBLCLK, &CPyDebugVariablesListCtrl::OnNMDblClick)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT(NM_CLICK, &CPyDebugVariablesListCtrl::OnNMClick)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT(LVN_ENDLABELEDIT, &CPyDebugVariablesListCtrl::OnLvnEndlabeledit)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT(LVN_BEGINLABELEDIT, &CPyDebugVariablesListCtrl::OnLvnBeginlabeledit)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT_EX(NM_CUSTOMDRAW, &CPyDebugVariablesListCtrl::OnCustomDraw)
  ON_WM_CHAR()
  ON_WM_PAINT()
  ON_WM_SIZE()
  ON_WM_MOUSEWHEEL()
  ON_WM_DESTROY()
  ON_WM_VSCROLL()
  ON_WM_HSCROLL()
  ON_WM_KILLFOCUS()
  ON_WM_CANCELMODE()
END_MESSAGE_MAP()

void CPyDebugVariablesListCtrl::CreateValueEditForNextItem()
{
  //Nothing to do if we have only one variable
  if ((m_nValueEditItem == 0) && (GetItemCount() == 2))
    return;

  int nNextItem{m_nValueEditItem};
  UpdateValueFromEdit();
  DestroyValueEdit();

  SetItemState(m_nValueEditItem, 0, LVIS_SELECTED | LVIS_FOCUSED);
  if (nNextItem < (GetItemCount() - 2))
    ++nNextItem;
  else
    nNextItem = 0;
  SetItemState(nNextItem, LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED);
  CreateValueEdit(nNextItem);
}

#pragma warning(suppress: 26429)
void CPyDebugVariablesListCtrl::OnLvnBeginlabeledit(_In_ NMHDR* /*pNMHDR*/, _Inout_ LRESULT *pResult)
{
  //Default the output parameter to 1
  *pResult = 1;

  //Do hit-testing to see where we should edit
  LVHITTESTINFO lvhit;
  lvhit.pt = GetCurrentMessage()->pt;
  ScreenToClient(&lvhit.pt);
  const int nItem{SubItemHitTest(&lvhit)};
  if (m_bEditable)
  {
    if ((nItem != -1))
    {
      if ((lvhit.iSubItem == 0) && (lvhit.flags & LVHT_ONITEM) && IsTopLevelVariable(nItem))
        *pResult = 0;
    }
    else
    {
      //If the hit testing fails on the current message position then fall back to editing the first selected item
      const int nCurSel{GetCurSel()};
      if (nCurSel != -1)
        *pResult = 0;
    }
  }
}

#pragma warning(suppress: 26429)
void CPyDebugVariablesListCtrl::OnNMClick(_In_ NMHDR* /*pNMHDR*/, _Inout_ LRESULT* pResult)
{
  //Validate our parameters
  ASSERT(pResult != nullptr);
  ASSERT(m_pMainFrame != nullptr);

  //Do hit-testing to see where we should edit
  LVHITTESTINFO lvhit;
  lvhit.pt = GetCurrentMessage()->pt;
  ScreenToClient(&lvhit.pt);
  const int nItem{SubItemHitTest(&lvhit)};

  //If the in place edit is already active then close it
  if (m_wndValueEdit.GetSafeHwnd())
    DestroyValueEdit();

  //Default the output parameter to 0
  *pResult = 0;

  //Start editing the sub-item if it is the second column and we are editable and the variable is defined and we are in break mode
  if (m_bEditable && (nItem != -1) && (lvhit.iSubItem == 1) && (lvhit.flags & LVHT_ONITEM) && GetItemText(nItem, 0).GetLength() && 
      (m_pMainFrame->GetDebugState() == CPyDebugFrameWnd::DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL))
  {
    CSingleLock sl{&m_csVariables, TRUE};

    //Pull out the variable to operate on
#pragma warning(suppress: 26446)
    const CPyDebugVariable& variable{m_Variables[nItem]};

    if (!variable.m_bError)
    {
      CreateValueEdit(nItem);
      *pResult = 1;
    }
  }
}

#pragma warning(suppress: 26429)
void CPyDebugVariablesListCtrl::OnNMDblClick(_In_ NMHDR* /*pNMHDR*/, _Inout_ LRESULT* pResult)
{
  //Validate our parameters
  ASSERT(pResult != nullptr);

  //Do hit-testing to see where we should expand / contract the current object
  LVHITTESTINFO lvhit;
  lvhit.pt = GetCurrentMessage()->pt;
  ScreenToClient(&lvhit.pt);
  const int nItem{SubItemHitTest(&lvhit)};
  if ((nItem != -1) && ((lvhit.flags & LVHT_ONITEMICON) || (lvhit.flags & LVHT_ONITEMLABEL)))
  {
    HandleVariableExpansionContraction(nItem);
    *pResult = 1;
  }
  else
  {
    //Let the base class do its thing
    *pResult = 0;
  }
}

BOOL CPyDebugVariablesListCtrl::CreateValueEdit(_In_ int nItem)
{
  //Destroy the edit box if need be
  if (m_wndValueEdit.GetSafeHwnd())
  {
    UpdateValueFromEdit();
    DestroyValueEdit();
  }

  //Get the rect of the sub-item
  CRect rectSubItem;
  GetSubItemRect(nItem, 1, LVIR_BOUNDS, rectSubItem);

  //Create the in-place rect
  CString sText{GetItemText(nItem, 1)};
  if (!m_wndValueEdit.Create(WS_BORDER | WS_VISIBLE | WS_CHILD | ES_LEFT | ES_AUTOHSCROLL | ES_WANTRETURN, rectSubItem, this, VARIABLES_LIST_CTRL_EDIT_ID))
    return FALSE;
  m_wndValueEdit.SetWindowText(sText);
  m_wndValueEdit.SetFont(GetFont());
  m_wndValueEdit.SetFocus();

  m_nValueEditItem = nItem;

  return TRUE;
}

void CPyDebugVariablesListCtrl::DestroyValueEdit()
{
  //Reset our variables for in-place editing values
  if (m_wndValueEdit.GetSafeHwnd())
    m_wndValueEdit.SendMessage(WM_CLOSE);
  m_nValueEditItem = -1;
}

void CPyDebugVariablesListCtrl::HandleDeleteVariable()
{
  //Validate our parameters
  if (!m_bEditable)
    return;

  const int nCurSel{GetCurSel()};

  //Delete all the selected top level items
  for (int i{GetItemCount()-2}; i>=0; i--) //The -2 ensures we do not delete the last item in the control which is the dummy placeholder
  {
    if (IsTopLevelVariable(i) && GetItemState(i, LVIS_SELECTED) & LVIS_SELECTED)
    {
      //Delete from our UI
      DeleteVariable(i);
    }
  }

  //Reselect an appropriate item if required
  int nNewCurSel{nCurSel};
  if ((nNewCurSel > 0) && (nNewCurSel == (GetItemCount() - 1)))
    --nNewCurSel;
  if (nNewCurSel != -1)
    SetItemState(nNewCurSel, LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED);
}

void CPyDebugVariablesListCtrl::UpdateValueFromEdit()
{
  //Update the text in the UI
  CString sNewValue;
  m_wndValueEdit.GetWindowText(sNewValue);
  SetItemText(m_nValueEditItem, 1, sNewValue);

  //Update the variable
  UpdateVariableValue(m_nValueEditItem, sNewValue);
}

#pragma warning(suppress: 26434)
void CPyDebugVariablesListCtrl::OnSize(_In_ UINT nType, _In_ int cx, _In_ int cy)
{
  //close down the edit control
  if (m_wndValueEdit.GetSafeHwnd())
  {
    UpdateValueFromEdit();
    DestroyValueEdit();
  }

  //let the parent class do its thing
  __super::OnSize(nType, cx, cy);
}

#pragma warning(suppress: 26434)
BOOL CPyDebugVariablesListCtrl::OnMouseWheel(_In_ UINT nFlags, _In_ short zDelta, _In_ CPoint pt)
{
  //close down the edit control
  if (m_wndValueEdit.GetSafeHwnd())
  {
    UpdateValueFromEdit();
    DestroyValueEdit();
  }

  //Let the parent class do its thing
  return __super::OnMouseWheel(nFlags, zDelta, pt);
}

#pragma warning(suppress: 26434)
void CPyDebugVariablesListCtrl::OnDestroy()
{
  //Destroy the old combo or edit box if need be
  DestroyValueEdit();

  //Let the parent class do its thing
  __super::OnDestroy();
}

#pragma warning(suppress: 26434)
void CPyDebugVariablesListCtrl::OnVScroll(_In_ UINT nSBCode, _In_ UINT nPos, _In_opt_ CScrollBar* pScrollBar)
{
  //close down the edit control
  if (m_wndValueEdit.GetSafeHwnd())
  {
    UpdateValueFromEdit();
    DestroyValueEdit();
  }

  //Let the parent class do its thing
  __super::OnVScroll(nSBCode, nPos, pScrollBar);
}

#pragma warning(suppress: 26434)
void CPyDebugVariablesListCtrl::OnHScroll(_In_ UINT nSBCode, _In_ UINT nPos, _In_opt_ CScrollBar* pScrollBar)
{
  //close down the edit control
  if (m_wndValueEdit.GetSafeHwnd())
  {
    UpdateValueFromEdit();
    DestroyValueEdit();
  }

  //Let the parent class do its thing
  __super::OnHScroll(nSBCode, nPos, pScrollBar);
}

#pragma warning(suppress: 26434)
void CPyDebugVariablesListCtrl::OnCancelMode()
{
  DestroyValueEdit();

  //Let the base class do its thing
  __super::OnCancelMode();
}

void CPyDebugVariablesListCtrl::HandleLoseFocusLogic(_In_opt_ CWnd* pNewWnd)
{
  //Destroy the edit box if we are losing focus to something else
  bool bForeignWnd{true};
#pragma warning(suppress: 26462)
  HWND hEdit{m_wndValueEdit.GetSafeHwnd()};
  if (hEdit && (pNewWnd != nullptr))
    bForeignWnd = (pNewWnd->GetSafeHwnd() != hEdit);

  if (bForeignWnd && hEdit)
  {
    UpdateValueFromEdit();
    DestroyValueEdit();
  }
}

#pragma warning(suppress: 26434)
void CPyDebugVariablesListCtrl::OnKillFocus(_In_opt_ CWnd* pNewWnd)
{
  //Call the helper function which does the main work
  HandleLoseFocusLogic(pNewWnd);

  //Let the parent class do its thing
  __super::OnKillFocus(pNewWnd);
}

int CPyDebugVariablesListCtrl::FindParentVariableIndex(_In_ int nVariableIndex)
{
  //The first item by definition cannot be a child item
  if (nVariableIndex == 0)
    return -1;

  CSingleLock sl{&m_csVariables, TRUE};

  //Pull out the variable to operate on
#pragma warning(suppress: 26446)
  const CPyDebugVariable& variable{m_Variables[nVariableIndex]};

  //Find the parent item of the current item
  for (int i{nVariableIndex-1}; i >= 0; i--)
  {
#pragma warning(suppress: 26446)
    if (m_Variables[i].m_nLevel < variable.m_nLevel)
      return i;
  }
  return -1;
}

int CPyDebugVariablesListCtrl::FindLastChildVariableIndex(_In_ int nVariableParentIndex)
{
  //What will be the return value from this method
  int nIndexLastChildItem{-1};

  CSingleLock sl{&m_csVariables, TRUE};

  //Pull out the variable to operate on
#pragma warning(suppress: 26446)
  const CPyDebugVariable& variable{m_Variables[nVariableParentIndex]};

  bool bFoundChild{true};
  int i{nVariableParentIndex + 1};
#pragma warning(suppress: 26472)
  const int nSize{static_cast<int>(m_Variables.size())};
  while (bFoundChild && (i < nSize))
  {
#pragma warning(suppress: 26446)
    if (m_Variables[i].m_nLevel > variable.m_nLevel)
      nIndexLastChildItem = i;
    else
      bFoundChild = false;

    //Prepare for the next loop
    ++i;
  }

  return nIndexLastChildItem;
}

BOOL CPyDebugVariablesListCtrl::PreTranslateMessage(MSG* pMsg)
{
  bool bHandled{false};
  CEdit* pEdit{GetEditControl()};
  if (m_bEditable && (pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_F2) && (pEdit == nullptr) && (m_wndValueEdit.GetSafeHwnd() == nullptr))
  {
    const int nCurSel{GetCurSel()};
    if (nCurSel != -1)
      EditLabel(nCurSel);
    bHandled = true;
  }
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_RIGHT) && (pEdit == nullptr) && (m_wndValueEdit.GetSafeHwnd() == nullptr))
  {
    const int nCurSel{GetCurSel()};
    if (nCurSel != -1)
      HandleVariableExpansion(nCurSel);
    bHandled = true;
  }
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_LEFT) && (pEdit == nullptr) && (m_wndValueEdit.GetSafeHwnd() == nullptr))
  {
    const int nCurSel{GetCurSel()};
    if (nCurSel != -1)
      HandleVariableLeftArrow(nCurSel);
    bHandled = true;
  }
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_DELETE))
  {
    if ((pEdit == nullptr) && (m_wndValueEdit.GetSafeHwnd() == nullptr))
    {
      HandleDeleteVariable();
      bHandled = true;
    }
    else
    {
      TranslateMessage(pMsg);
      DispatchMessage(pMsg);
      bHandled = true;
    }
  }
  else if (pEdit != nullptr)
  {
    if ((pMsg->message == WM_KEYDOWN) && (::GetKeyState(VK_CONTROL) < 0))
    {
      bHandled = true;
      switch (pMsg->wParam)
      {
        case _T('Z'):
        {
          pEdit->Undo();
          break;
        }
        case _T('X'):
        {
          pEdit->Cut();
          break;
        }
        case _T('C'):
        {
          pEdit->Copy();
          break;
        }
        case _T('V'):
        {
          pEdit->Paste();
          break;
        }
        default:
        {
          bHandled = false;
          break;
        }
      }
    }
  }

  if (bHandled)
    return TRUE;
  else
    return __super::PreTranslateMessage(pMsg);
}

int CPyDebugVariablesListCtrl::GetCurSel() const
{
  const int nCount{GetItemCount()};
  for (int i{0}; i<nCount; i++)
  {
    if (GetItemState(i, LVIS_SELECTED) & LVIS_SELECTED)
      return i;
  }
  return -1;
}

#pragma warning(suppress: 26434)
void CPyDebugVariablesListCtrl::OnChar(_In_ UINT nChar, _In_ UINT nRepCnt, _In_ UINT nFlags)
{
  //Any character will start in place editing of the item
  if (GetEditControl())
    __super::OnChar(nChar, nRepCnt, nFlags);
  else if (m_bEditable)
  {
    const int nCurSel{GetCurSel()};
    if (nCurSel != -1)
    {
      CEdit* pEdit{EditLabel(nCurSel)};
      if (pEdit != nullptr)
      {
        const MSG* pMsg{GetCurrentMessage()};
        pEdit->PostMessage(pMsg->message, pMsg->wParam, pMsg->lParam);
      }
    }
  }
  else
    __super::OnChar(nChar, nRepCnt, nFlags);
}

#pragma warning(suppress: 26429 26461)
void CPyDebugVariablesListCtrl::OnLvnEndlabeledit(_In_ NMHDR* pNMHDR, _Inout_ LRESULT* pResult)
{
#pragma warning(suppress: 26490)
  const NMLVDISPINFO* pDispInfo{reinterpret_cast<NMLVDISPINFO*>(pNMHDR)};
  if (pDispInfo->item.pszText)
  {
#pragma warning(suppress: 26486)
    if (_tcslen(pDispInfo->item.pszText) == 0)
    {
      if (pDispInfo->item.iItem != (GetItemCount() - 1)) //Don't allow the last dummy item to be deleted
      {
        DeleteVariable(pDispInfo->item.iItem);

        //Reselect an appropriate item if required
        int nNewCurSel{pDispInfo->item.iItem};
        if ((nNewCurSel > 0) && (nNewCurSel == (GetItemCount() - 1)))
          --nNewCurSel;
        if (nNewCurSel != -1)
          SetItemState(nNewCurSel, LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED);

        *pResult = TRUE;
      }
      else
        *pResult = FALSE;
    }
    else
    {
#pragma warning(suppress: 26486)
      UpdateVariableName(pDispInfo->item.iItem, pDispInfo->item.pszText);

      //Auto select the dummy item if we have just finished editing the last real variable
      if (pDispInfo->item.iItem == (GetItemCount() - 2))
      {
        SetItemState(pDispInfo->item.iItem, 0, LVIS_SELECTED | LVIS_FOCUSED);
        SetItemState(pDispInfo->item.iItem + 1, LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED);
      }

      *pResult = TRUE;
    }
  }
  else
    *pResult = FALSE;
}

#pragma warning(suppress: 26429)
BOOL CPyDebugVariablesListCtrl::OnCustomDraw(_In_ NMHDR* pNMHDR, _Inout_ LRESULT* pResult)
{
#pragma warning(suppress: 26490)
  auto pCustomDraw{reinterpret_cast<NMLVCUSTOMDRAW*>(pNMHDR)};
  ASSERT(pCustomDraw != nullptr);

  switch (pCustomDraw->nmcd.dwDrawStage)
  {
    case CDDS_PREPAINT:
    {
      *pResult = CDRF_NOTIFYITEMDRAW; //Tell the control that we are interested in item notifications
      break;
    }
    case CDDS_ITEMPREPAINT:
    {
      *pResult = CDRF_NOTIFYSUBITEMDRAW;
      break;
    }
    case CDDS_SUBITEM | CDDS_ITEMPREPAINT:
    {
#pragma warning(suppress: 26472)
      const int nItem{static_cast<int>(pCustomDraw->nmcd.dwItemSpec)};
      const DWORD_PTR dwItemData{GetItemData(nItem)};
      if ((pCustomDraw->iSubItem == 1) && dwItemData)
        pCustomDraw->clrText = RGB(255, 0, 0);
      else
        pCustomDraw->clrText = RGB(0, 0, 0);
      *pResult = CDRF_DODEFAULT;
      break;
    }
    case CDDS_ITEMPOSTPAINT:
    {
      *pResult = CDRF_DODEFAULT;
      break;
    }
    default:
    {
      break;
    }
  }

  return TRUE; //Allow the message to be reflected again
}

void CPyDebugVariablesListCtrl::ResetChangedStateForAllVariables()
{
  CSingleLock sl{&m_csVariables, TRUE};
  for (auto& variable : m_Variables)
    variable.m_bChanged = false;
}

void CPyDebugVariablesListCtrl::ResetValuesForAllVariables()
{
  CSingleLock sl{&m_csVariables, TRUE};
  int i{0};
  for (auto& variable : m_Variables)
  {
    variable.m_bChanged = false;
    variable.m_bError = false;
    variable.m_sType.Empty();
    variable.m_sValue.Empty();
    variable.m_bExpandable = false;
    variable.m_bCollapsed = true;

    //Delete any possible child variable
    DeleteChildVariables(i);
    i++;
  }
}

void CPyDebugVariablesListCtrl::DeleteChildVariables(_In_ int nVariableParentIndex)
{
  CSingleLock sl{&m_csVariables, TRUE};

  //Find the index of the last child of this item
  const int nIndexLastChildItem{FindLastChildVariableIndex(nVariableParentIndex)};
  if (nIndexLastChildItem == -1)
    return;

  //Remove from the in-memory array
  m_Variables.erase(m_Variables.cbegin() + nVariableParentIndex + 1, m_Variables.cbegin() + nIndexLastChildItem + 1);

  //Delete in the UI
  if (GetSafeHwnd())
  {
    for (int j{nIndexLastChildItem}; j>nVariableParentIndex; j--)
      DeleteItem(j);
  }
}

void CPyDebugVariablesListCtrl::DeleteVariable(_In_ int nVariableIndex)
{
  CSingleLock sl{&m_csVariables, TRUE};

#ifdef _DEBUG
#pragma warning(suppress: 26446)
  const CPyDebugVariable& variable{m_Variables[nVariableIndex]};
  ASSERT(variable.m_nLevel == 0); //Should only be called for top level variables
#endif //#ifdef _DEBUG

  DeleteChildVariables(nVariableIndex); //Delete any possible child variables

  //Remove from the in-memory array
  m_Variables.erase(m_Variables.cbegin() + nVariableIndex);

  //Delete in the UI
  DeleteItem(nVariableIndex);
}

void CPyDebugVariablesListCtrl::AddVariable(_In_ int nVariableIndex)
{
  CSingleLock sl{&m_csVariables, TRUE};
#pragma warning(suppress: 26446)
  CPyDebugVariable& variable{m_Variables[nVariableIndex]};
  LVITEM lvi{};
  lvi.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_INDENT;
  lvi.iItem = nVariableIndex;
  lvi.iImage = GetVariableImage(variable);
  lvi.iIndent = variable.m_nLevel;
  lvi.pszText = variable.m_sName.GetBuffer();
  const int nIndex{InsertItem(&lvi)};
  variable.m_sName.ReleaseBuffer();
  SetItemText(nIndex, 1, variable.m_sValue);
  SetItemText(nIndex, 2, variable.m_sType);
  SetItemData(nIndex, variable.m_bChanged);
}

void CPyDebugVariablesListCtrl::UpdateVariable(_In_ int nVariableIndex)
{
  CSingleLock sl{&m_csVariables, TRUE};
#pragma warning(suppress: 26446)
  CPyDebugVariable& variable{m_Variables[nVariableIndex]};
  LVITEM lvi{};
  lvi.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_INDENT;
  lvi.iItem = nVariableIndex;
  lvi.iImage = GetVariableImage(variable);
  lvi.iIndent = variable.m_nLevel;
  lvi.pszText = variable.m_sName.GetBuffer();
  const BOOL bResult{SetItem(&lvi)};
  variable.m_sName.ReleaseBuffer();
  if (!bResult)
    return;
  SetItemText(nVariableIndex, 1, variable.m_sValue);
  SetItemText(nVariableIndex, 2, variable.m_sType);
  SetItemData(nVariableIndex, variable.m_bChanged);
}

void CPyDebugVariablesListCtrl::AddDummyItemIfNecessary()
{
  bool bAddDummyItem{false};
  CSingleLock sl{&m_csVariables, TRUE};
  const auto nSize{m_Variables.size()};
  if (m_bEditable)
  {
    if (nSize)
#pragma warning(suppress: 26446)
      bAddDummyItem = (m_Variables[nSize - 1].m_sName.GetLength() != 0);
    else
      bAddDummyItem = true;
  }
  if (bAddDummyItem)
  {
    CPyDebugVariable dummyVariable;
    m_Variables.push_back(dummyVariable);
#pragma warning(suppress: 26472)
    AddVariable(static_cast<int>(nSize));
  }
}

void CPyDebugVariablesListCtrl::UpdateVariableWindow(_In_ int nSelectionSel)
{
  if (GetSafeHwnd() == nullptr)
    return;

  SetRedraw(FALSE);
  DeleteAllItems();
  EnableWindow(TRUE);
  CSingleLock sl{&m_csVariables, TRUE};
#pragma warning(suppress: 26472)
  const int nVariables{static_cast<int>(m_Variables.size())};

  //Readd a dummy empty item if required
  AddDummyItemIfNecessary();

  //Add all the variables
  for (int i{0}; i<nVariables; i++)
    AddVariable(i);
  SetRedraw(TRUE);

  if (nSelectionSel != -1)
    SetItemState(nSelectionSel, LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED);
}

#pragma warning(suppress: 26440)
int CPyDebugVariablesListCtrl::GetVariableImage(_In_ const CPyDebugVariable& variable)
{
  int nImage{2};
  if (variable.m_sName.IsEmpty())
    nImage = 3;
  else if (variable.m_bError)
    nImage = 1;
  else if (variable.m_bExpandable)
  {
    if (variable.m_bCollapsed)
      nImage = 0;
    else
      nImage = 4;
  }
  return nImage;
}

bool CPyDebugVariablesListCtrl::IsTopLevelVariable(_In_ int nVariableIndex)
{
  CSingleLock sl{&m_csVariables, TRUE};
#pragma warning(suppress: 26446)
  return m_Variables[nVariableIndex].m_nLevel == 0;
}

void CPyDebugVariablesListCtrl::HandleVariableLeftArrow(_In_ int nVariableIndex)
{
  //Pull out the watch variable to update
  CSingleLock sl{&m_csVariables, TRUE};

  //Pull out the variable to operate on
#pragma warning(suppress: 26446)
  CPyDebugVariable& variable{m_Variables[nVariableIndex]};

  //Navigate to the parent item if we can
  if (!variable.m_bExpandable || variable.m_bCollapsed)
  {
    const int nParentIndex{FindParentVariableIndex(nVariableIndex)};
    if (nParentIndex != -1)
    {
      SetItemState(nVariableIndex, 0, LVIS_SELECTED | LVIS_FOCUSED);
      SetItemState(nParentIndex, LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED);
      EnsureVisible(nParentIndex, FALSE);
    }
    return;
  }

  //Mark this item as collapsed in memory and in the UI
  variable.m_bCollapsed = true;
  UpdateVariable(nVariableIndex);

  //Find the index of the last child of this item
  const int nIndexLastChildItem{FindLastChildVariableIndex(nVariableIndex)};

  //Now delete all variables up to the parent in memory
  if (nIndexLastChildItem != -1)
    m_Variables.erase(m_Variables.cbegin() + nVariableIndex + 1, m_Variables.cbegin() + nIndexLastChildItem + 1);

  //Also remove all the items from the UI
  if (nIndexLastChildItem != -1)
  {
    SetRedraw(FALSE);
    for (int i{nIndexLastChildItem}; i>nVariableIndex; i--)
      DeleteItem(i);
    SetRedraw(TRUE);
  }
}

void CPyDebugVariablesListCtrl::HandleVariableExpansionContraction(_In_ int nVariableIndex)
{
  //Pull out the watch variable to update
  CSingleLock sl{&m_csVariables, TRUE};

  //Pull out the variable to operate on
#pragma warning(suppress: 26446)
  const CPyDebugVariable& variable{m_Variables[nVariableIndex]};

  //Nothing to do if the variable is not expandable
  if (!variable.m_bExpandable)
    return;

  if (!variable.m_bCollapsed)
    HandleVariableLeftArrow(nVariableIndex);
  else
    HandleVariableExpansion(nVariableIndex);
}

void CPyDebugVariablesListCtrl::UpdateVariableName(_In_ int nVariableIndex, _In_z_ LPCTSTR pszNewName)
{
  //Validate our parameters
  ASSERT(m_pMainFrame != nullptr);

  CSingleLock sl{&m_csVariables, TRUE};

  //Pull out the watch variable to update
#pragma warning(suppress: 26446)
  CPyDebugVariable& variable{m_Variables[nVariableIndex]};

  //Nothing to do if the name has not changed
  if (variable.m_sName == pszNewName)
    return;

  //Update the in-memory array
  variable.m_sName = pszNewName;

  if (m_pMainFrame->GetDebugState() == CPyDebugFrameWnd::DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL)
  {
    //Request the worker thread to update the variable name
    m_pMainFrame->m_nVariablesWndRequestIndex = nVariableIndex;
    m_pMainFrame->m_VariablesWndRequest = CPyDebugFrameWnd::VARIABLES_WND_REQUEST::CHANGE_VARIABLE_NAME;
  }

  //Force contraction
  HandleVariableLeftArrow(nVariableIndex);

  //Update the variable in the UI
  UpdateVariable(nVariableIndex);

  //Readd the dummy item if necessary
  AddDummyItemIfNecessary();
}

void CPyDebugVariablesListCtrl::UpdateVariableValue(_In_ int nVariableIndex, _In_z_ LPCTSTR pszValue)
{
  //Validate our parameters
  ASSERT(m_pMainFrame != nullptr);

  CSingleLock sl{&m_csVariables, TRUE};

  //Pull out the watch variable to update
#pragma warning(suppress: 26446)
  const CPyDebugVariable& variable{m_Variables[nVariableIndex]};

  //Nothing to do if the value has not changed
  if (variable.m_sValue == pszValue)
    return;

  if (m_pMainFrame->GetDebugState() == CPyDebugFrameWnd::DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL)
  {
    //Request the worker thread to update the variable value
    m_pMainFrame->m_nVariablesWndRequestIndex = nVariableIndex;
    m_pMainFrame->m_sVariablesWndRequestString = pszValue;
    m_pMainFrame->m_VariablesWndRequest = CPyDebugFrameWnd::VARIABLES_WND_REQUEST::CHANGE_VARIABLE_VALUE;
  }

  //Force contraction
  HandleVariableLeftArrow(nVariableIndex);

  //Update the variable in the UI
  UpdateVariable(nVariableIndex);
}

void CPyDebugVariablesListCtrl::HandleVariableExpansion(_In_ int nVariableIndex)
{
  //Validate our parameters
  ASSERT(m_pMainFrame != nullptr);

  CSingleLock sl{&m_csVariables, TRUE};
#pragma warning(suppress: 26446)
  CPyDebugVariable variable{m_Variables[nVariableIndex]};

  //Nothing to do if the variable is not expandable
  if (!variable.m_bExpandable)
    return;

  //Nothing to do if the object is already expanded
  if (!variable.m_bCollapsed)
    return;

  //Nothing to do if current state of the code is not breaked
  if (m_pMainFrame->GetDebugState() != CPyDebugFrameWnd::DEBUG_STATE::ACKNOWLEDGE_BREAK_ALL)
    return;

  //Request the worker thread to update the expansion state
  m_pMainFrame->m_VariablesWndRequestType = m_Type;
  m_pMainFrame->m_nVariablesWndRequestIndex = nVariableIndex;
  m_pMainFrame->m_VariablesWndRequest = CPyDebugFrameWnd::VARIABLES_WND_REQUEST::EXPAND_VARIABLE;
}

void CPyDebugVariablesListCtrl::CaptureVariablesDetails(_In_opt_ PyObject* pObject)
{
  //Get access to the "eval" function
  CPyObject pBuiltins{PyImport_ImportModule("builtins")};
  CPyObject pEval;
  if (pBuiltins != nullptr)
    pEval = PyObject_GetAttrString(pBuiltins, "eval");

  CSingleLock sl{&m_csVariables, TRUE};
  m_Variables.clear();
  if (pObject != nullptr)
  {
    Py_ssize_t pos{0};
    PyObject* pKey{nullptr};
    PyObject* pValue{nullptr};
#pragma warning(suppress: 26486)
    while (PyDict_Next(pObject, &pos, &pKey, &pValue))
    {
      if ((pKey != nullptr) && (pValue != nullptr))
      {
        CPyDebugVariable variable;
        variable.m_sName = CPyDebugFrameWnd::PyToCString(pKey);
        variable.m_sValue = CPyDebugFrameWnd::PyToCString(pValue);
        variable.m_sType = pValue->ob_type->tp_name;
#pragma warning(suppress: 26486)
        variable.m_bExpandable = IsExpandableVariable(variable.m_sName, pEval);
        m_Variables.push_back(variable);
      }
    }
  }
  std::sort(m_Variables.begin(), m_Variables.end(), CPyDebugLessThanPyVariable());
}

void CPyDebugVariablesListCtrl::CaptureVariablesDetails(_In_ bool bUpdateChangeState)
{
  //Get access to the "eval" function
  CPyObject pBuiltins{PyImport_ImportModule("builtins")};
  CPyObject pEval;
  if (pBuiltins != nullptr)
    pEval = PyObject_GetAttrString(pBuiltins, "eval");

  CSingleLock sl{&m_csVariables, TRUE};
  for (auto& variable : m_Variables)
#pragma warning(suppress: 26486)
    CaptureVariableDetail(variable, pEval, bUpdateChangeState);
}

bool CPyDebugVariablesListCtrl::SetVariableValue(_In_ CPyDebugVariable& variable, _In_z_ LPCTSTR pszValue)
{
  //Validate our parameters
  ASSERT(m_pMainFrame != nullptr);

  //What will be the return value from this method
  bool bSuccess{true};

  CStringA sCode;
  sCode.Format("%s = %s", CStringA(variable.m_sName).GetString(), CStringA(pszValue).GetString());
  PyErr_Clear();
#pragma warning(suppress: 26477)
  const int nRun{PyRun_SimpleString(sCode)};
  if (nRun != 0)
  {
    m_pMainFrame->AddOutputText(IDS_PYDEBUG_FAIL_SET_VARIABLE_VALUE, nRun, variable.m_sName, pszValue);
    bSuccess = false;
  }

  return bSuccess;
}

bool CPyDebugVariablesListCtrl::IsExpandableVariable(_In_ LPCTSTR pszVariableName, _In_ PyObject* pEval)
{
  //What will be the return value from this method
  bool bExpandable{false};

  //Get access to the variable as a callable object
  CPyObject pArgs{Py_BuildValue("(s)", CStringA(pszVariableName).GetString())};
  PyErr_Clear();
  CPyObject pObject{PyObject_Call(pEval, pArgs, nullptr)};
  PyErr_Clear();
  if (pObject != nullptr)
  {
    //A variable is expandable if it supports the Python sequence protocol, the Python map protocol or has a dictionary which returns at least one item
    bExpandable = (PySequence_Check(pObject) != 0);
    if (!bExpandable)
      bExpandable = (PyMapping_Check(pObject) != 0);
    if (!bExpandable)
    {
      //Use the dictionary to find if there are any sub variables
      CPyObject pDict{PyObject_GenericGetDict(pObject, nullptr)};
      if (pDict != nullptr)
      {
        Py_ssize_t pos{0};
        PyObject* pKey{nullptr};
        PyObject* pValue{nullptr};
#pragma warning(suppress: 26486)
        if (PyDict_Next(pDict, &pos, &pKey, &pValue))
        {
          if ((pKey != nullptr) && (pValue != nullptr))
            bExpandable = true;
        }
      }
    }
  }
  PyErr_Clear();
  return bExpandable;
}

void CPyDebugVariablesListCtrl::CaptureVariableDetail(_In_ CPyDebugVariable& variable, _In_ PyObject* pEval, _In_ bool bUpdateChangeState)
{
  if (variable.m_sName.GetLength()) //Skip over the special empty placeholder variable
  {
    if (pEval != nullptr)
    {
      CPyObject pArgs{Py_BuildValue("(s)", CStringA(variable.m_sName).GetString())};
      PyErr_Clear();
      CPyObject pObject{PyObject_Call(pEval, pArgs, nullptr)};
      PyErr_Clear();
      if (pObject == nullptr)
      {
        CString sNewValue;
        AfxFormatString1(sNewValue, IDS_PYDEBUG_IDENTIFIER_IS_UNDEFINED, variable.m_sName);
        if (sNewValue != variable.m_sValue)
        {
          variable.m_sValue = sNewValue;
          if (bUpdateChangeState)
            variable.m_bChanged = true;
        }
        variable.m_bError = true;
        variable.m_sType.Empty();
      }
      else
      {
        CString sNewValue{CPyDebugFrameWnd::PyToCString(pObject)};
        if (sNewValue != variable.m_sValue)
        {
          variable.m_sValue = sNewValue;
          if (bUpdateChangeState)
            variable.m_bChanged = true;
        }
        variable.m_bError = false;
        variable.m_sType = pObject->ob_type->tp_name;
        variable.m_bExpandable = IsExpandableVariable(variable.m_sName, pEval);
      }
    }
    else
    {
      CString sNewValue;
#pragma warning(suppress: 26496)
      ENSURE(sNewValue.LoadString(IDS_PYDEBUG_IDENTIFIER_EVALUATION_FAILURE));
      if (sNewValue != variable.m_sValue)
      {
        variable.m_sValue = sNewValue;
        if (bUpdateChangeState)
          variable.m_bChanged = true;
      }
      variable.m_bError = true;
      variable.m_sType.Empty();
    }
  }
}


#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CPyDebugVariablesWnd, CDockablePane) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_CREATE()
  ON_WM_SIZE()
  ON_WM_SETFOCUS()
END_MESSAGE_MAP()

void CPyDebugVariablesWnd::SetEditable(bool bEditable) noexcept
{
  m_bEditable = bEditable;
  m_wndCtrl.SetEditable(bEditable);
}

#pragma warning(suppress: 26434)
int CPyDebugVariablesWnd::OnCreate(_In_ LPCREATESTRUCT lpCreateStruct)
{
  //Validate our parameters
  ASSERT(m_pMainFrame != nullptr); //Did you forget to call SetMainFrame

  //Let the base class do its thing
  if (__super::OnCreate(lpCreateStruct) == -1)
    return -1;

  //Create the list control
  CRect rectDummy;
  rectDummy.SetRectEmpty();
  DWORD dwStyle{WS_CHILD | WS_VISIBLE | WS_TABSTOP | LVS_REPORT | LVS_SHOWSELALWAYS};
  if (m_bEditable)
    dwStyle |= LVS_EDITLABELS;
  if (!m_wndCtrl.Create(dwStyle, rectDummy, this, 2))
  {
    TRACE(_T("Failed to create variables window\n"));
    return -1;
  }
  m_wndCtrl.SetExtendedStyle(LVS_EX_FULLROWSELECT | LVS_EX_LABELTIP | LVS_EX_GRIDLINES);
  if (!m_bEditable)
    m_wndCtrl.EnableWindow(FALSE);

  //Create the image list
  const bool bHiColorIcons{m_pMainFrame->GetHiColorIcons()};
  BOOL bImageListCreated{m_ImageList.Create(24, 16, bHiColorIcons ? ILC_MASK | ILC_COLOR24 : ILC_MASK, 0, 0)};
  if (bImageListCreated)
  {
    CBitmap bmp;
    bImageListCreated = bmp.LoadBitmap(IDB_PYDEBUG_WATCH);
    if (bImageListCreated)
      m_ImageList.Add(&bmp, RGB(255, 0, 255));
  }
  if (!bImageListCreated)
  {
    TRACE(_T("Failed to create watch window image list\n"));
    return -1;
  }
  m_wndCtrl.SetImageList(&m_ImageList, LVSIL_SMALL);

  //Add the columns to the list control
  CString sColumn;
  if (!sColumn.LoadString(IDS_PYDEBUG_NAME))
  {
    TRACE(_T("Failed to load Variables window column caption\n"));
    return -1;
  }
  m_wndCtrl.InsertColumn(0, sColumn, LVCFMT_LEFT, 200);
  if (!sColumn.LoadString(IDS_PYDEBUG_VALUE))
  {
    TRACE(_T("Failed to load Variables window column caption\n"));
    return -1;
  }
  m_wndCtrl.InsertColumn(1, sColumn, LVCFMT_LEFT, 350);
  if (!sColumn.LoadString(IDS_PYDEBUG_TYPE))
  {
    TRACE(_T("Failed to load Variables window column caption\n"));
    return -1;
  }
  m_wndCtrl.InsertColumn(2, sColumn, LVCFMT_LEFT, 200);

  return 0;
}

#pragma warning(suppress: 26434)
void CPyDebugVariablesWnd::OnSize(_In_ UINT nType, _In_ int cx, _In_ int cy)
{
  //Let the base class do its thing
  __super::OnSize(nType, cx, cy);

  //Make the control window covers the whole client area
  m_wndCtrl.SetWindowPos(nullptr, -1, -1, cx, cy, SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOZORDER);
}

#pragma warning(suppress: 26434)
void CPyDebugVariablesWnd::OnSetFocus(_In_opt_ CWnd* pOldWnd)
{
  //Let the base class do its thing
  __super::OnSetFocus(pOldWnd);

  //Give the focus to the control
  m_wndCtrl.SetFocus();
}

void CPyDebugVariablesWnd::SetPyDebugMainFrame(_In_opt_ CPyDebugFrameWnd* pMainFrame) noexcept
{
  m_pMainFrame = pMainFrame;
  m_wndCtrl.m_pMainFrame = pMainFrame;
}
