/*
Module : PyDebugVariablesWnd.h
Purpose: Defines the interface for the classes which encapsulate the Variables windows functionality of PyDebug++
Created: PJN / 20-08-2015

Copyright (c) 2015 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef __PYDEBUGVARIABLESWND_H__
#define __PYDEBUGVARIABLESWND_H__

#ifndef CPYDEBUG_EXT_CLASS
#define CPYDEBUG_EXT_CLASS
#endif //#ifndef CPYDEBUG_EXT_CLASS


//////////////////// Includes /////////////////////////////////////////////////

#include "PythonPtr.h"


//////////////////// Classes //////////////////////////////////////////////////

//A variable as shown in the PyDebug++ windows
class CPYDEBUG_EXT_CLASS CPyDebugVariable
{
public:
//Member variables
  CString m_sName;
  CString m_sValue;
  CString m_sType;
  bool m_bError{false};
  bool m_bChanged{false};
  bool m_bExpandable{false};
  bool m_bCollapsed{false};
  int m_nLevel{0};
};


//Class which represents an inplace edit control for editing variable values
class CPYDEBUG_EXT_CLASS CPyDebugVariablesEdit : public CEdit
{
protected:
  //Message handlers
  afx_msg UINT OnGetDlgCode();

  DECLARE_MESSAGE_MAP()
};


class CPyDebugFrameWnd;


//Class which implements the list control for the watch window
class CPYDEBUG_EXT_CLASS CPyDebugVariablesListCtrl : public CListCtrl
{
public:
//Enums
  enum class TYPE
  {
    NONE,
    GLOBALS,
    LOCALS,
    BUILTINS,
    WATCH
  };

//Methods
  void SetEditable(_In_ bool bEditable) noexcept { m_bEditable = bEditable; }
  void SetType(_In_ TYPE type) noexcept { m_Type = type; }

protected:
//Member variables
  CPyDebugVariablesEdit m_wndValueEdit;
  int m_nValueEditItem{-1};
  CPyDebugFrameWnd* m_pMainFrame{nullptr};
  bool m_bEditable{false};
  CCriticalSection m_csVariables;
  std::vector<CPyDebugVariable> m_Variables;
  TYPE m_Type{TYPE::NONE};

//Methods
  virtual void HandleDeleteVariable();
  virtual void UpdateValueFromEdit();
  virtual BOOL CreateValueEdit(_In_ int nItem);
  virtual void CreateValueEditForNextItem();
  virtual void DestroyValueEdit();
  _NODISCARD virtual int GetCurSel() const;
  BOOL PreTranslateMessage(MSG* pMsg) override;
  virtual void HandleLoseFocusLogic(_In_opt_ CWnd* pNewWnd);
  virtual int FindLastChildVariableIndex(_In_ int nVariableParentIndex);
  virtual int FindParentVariableIndex(_In_ int nVariableIndex);
  virtual void ResetChangedStateForAllVariables();
  virtual void ResetValuesForAllVariables();
  virtual void DeleteChildVariables(_In_ int nVariableParentIndex);
  virtual void DeleteVariable(_In_ int nVariableIndex);
  virtual void AddVariable(_In_ int nVariableIndex);
  virtual void UpdateVariable(_In_ int nVariableIndex);
  virtual void UpdateVariableWindow(_In_ int nSelectionSel = -1);
  virtual int  GetVariableImage(_In_ const CPyDebugVariable& variable);
  virtual bool IsTopLevelVariable(_In_ int nVariableIndex);
  virtual void HandleVariableLeftArrow(_In_ int nVariableIndex);
  virtual void HandleVariableExpansionContraction(_In_ int nVariableIndex);
  virtual void AddDummyItemIfNecessary();
  virtual void UpdateVariableName(_In_ int nVariableIndex, _In_z_ LPCTSTR pszNewName);
  virtual void UpdateVariableValue(_In_ int nVariableIndex, _In_z_ LPCTSTR pszValue);
  virtual void HandleVariableExpansion(_In_ int nVariableIndex);
  virtual void CaptureVariablesDetails(_In_opt_ PyObject* pObject);
  virtual void CaptureVariablesDetails(_In_ bool bUpdateChangeState);
  virtual void CaptureVariableDetail(_In_ CPyDebugVariable& variable, _In_ PyObject* pEval, _In_ bool bUpdateChangeState);
  virtual bool SetVariableValue(_In_ CPyDebugVariable& variable, _In_z_ LPCTSTR pszValue);
  virtual bool IsExpandableVariable(_In_ LPCTSTR pszVariableName, _In_ PyObject* pEval);

//Message handlers
  afx_msg void OnNMClick(_In_ NMHDR* pNMHDR, _Inout_ LRESULT* pResult);
  afx_msg void OnLvnBeginlabeledit(_In_ NMHDR* pNMHDR, _Inout_ LRESULT* pResult);
  afx_msg void OnNMDblClick(_In_ NMHDR* pNMHDR, _Inout_ LRESULT* pResult);
  afx_msg void OnLvnEndlabeledit(_In_ NMHDR* pNMHDR, _Inout_ LRESULT* pResult);
  afx_msg void OnChar(_In_ UINT nChar, _In_ UINT nRepCnt, _In_ UINT nFlags);
  afx_msg BOOL OnCustomDraw(_In_ NMHDR* pNMHDR, _Inout_ LRESULT* pResult);
  afx_msg void OnSize(_In_ UINT nType, _In_ int cx, _In_ int cy);
  afx_msg BOOL OnMouseWheel(_In_ UINT nFlags, _In_ short zDelta, _In_ CPoint pt);
  afx_msg void OnDestroy();
  afx_msg void OnVScroll(_In_ UINT nSBCode, _In_ UINT nPos, _In_opt_ CScrollBar* pScrollBar);
  afx_msg void OnHScroll(_In_ UINT nSBCode, _In_ UINT nPos, _In_opt_ CScrollBar* pScrollBar);
  afx_msg void OnKillFocus(_In_opt_ CWnd* pNewWnd);
  afx_msg void OnCancelMode();

  DECLARE_MESSAGE_MAP()

  friend class CPyDebugVariablesWnd;
  friend class CPyDebugVariablesEdit;
  friend class CPyDebugFrameWnd;
};


class CPYDEBUG_EXT_CLASS CPyDebugVariablesWnd : public CDockablePane
{
public:
//Member variables
  CPyDebugVariablesListCtrl m_wndCtrl;

//Methods
  void SetEditable(bool bEditable) noexcept;
  void SetPyDebugMainFrame(_In_opt_ CPyDebugFrameWnd* pMainFrame) noexcept;

protected:
//Member variables
  CImageList m_ImageList;
  bool m_bEditable{false};
  CPyDebugFrameWnd* m_pMainFrame{nullptr};

//Message handlers
  afx_msg int OnCreate(_In_ LPCREATESTRUCT lpCreateStruct);
  afx_msg void OnSize(_In_ UINT nType, _In_ int cx, _In_ int cy);
  afx_msg void OnSetFocus(_In_opt_ CWnd* pOldWnd);

  DECLARE_MESSAGE_MAP()
};


#endif //#ifndef __PYDEBUGVARIABLESWND_H__
