/*
Module : PyDebugView.cpp
Purpose: Defines the implementation for the classes which encapsulate the MFC View functionality of PyDebug++
Created: PJN / 20-08-2015
History: PJN / 01-08-2022 1. Updated all the code in to use C++ uniform initialization for all variable
                          declarations.

Copyright (c) 2015 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "PyDebugDoc.h"
#include "PyDebugView.h"
#include "PyDebugMainFrm.h"
#include "PyDebugPP.h"


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG

//Taken from "lib\keyword.py" in the Python distribution
constexpr const TCHAR* g_pyKeyWords = _T("False None True __peg_parser__ and as assert async await break class continue def del elif else except finally for from global if import in is lambda nonlocal not or pass raise return try while with yield");


//////////////////// Implementation ///////////////////////////////////////////


#pragma warning(suppress: 26409 26433 26440 26477)
IMPLEMENT_DYNCREATE(CPyDebugView, CScintillaView)

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CPyDebugView, CScintillaView) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
  ON_WM_CREATE()
END_MESSAGE_MAP()

#pragma warning(suppress: 26434 26440)
CPyDebugDoc* CPyDebugView::GetDocument() const //non-debug version is inline
{
  ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CPyDebugDoc)));
#pragma warning(suppress: 26466)
  return static_cast<CPyDebugDoc*>(m_pDocument);
}

void CPyDebugView::SetAStyle(_In_ int style, _In_ COLORREF fore, _In_ COLORREF back, _In_ int size, _In_opt_z_ const char* face) 
{
  auto& rCtrl{GetCtrl()};

  rCtrl.StyleSetFore(style, fore);
  rCtrl.StyleSetBack(style, back);
  if (size >= 1)
    rCtrl.StyleSetSize(style, size);
  if (face != nullptr) 
    rCtrl.StyleSetFont(style, face);
}

void CPyDebugView::OnInitialUpdate()
{
  //Let the base class do its thing
  __super::OnInitialUpdate();

  auto& rCtrl{GetCtrl()};

  //Setup the Lexer
  rCtrl.SetILexer(m_pPythonLexer);
  rCtrl.SetKeyWords(0, g_pyKeyWords);
  SetAStyle(static_cast<int>(Scintilla::StylesCommon::Default), RGB(0, 0, 0), RGB(0xFF, 0xFF, 0xFF));
  rCtrl.StyleClearAll();
  SetAStyle(SCE_P_DEFAULT, RGB(0, 0, 0));
  SetAStyle(SCE_P_COMMENTLINE, RGB(0, 0x80, 0));
  SetAStyle(SCE_P_COMMENTBLOCK, RGB(0, 0x80, 0));
  SetAStyle(SCE_P_NUMBER, RGB(0, 0x80, 0x80));
  SetAStyle(SCE_P_WORD, RGB(0, 0, 0x80));
  rCtrl.StyleSetBold(SCE_P_WORD, 1);
  SetAStyle(SCE_P_WORD2, RGB(0, 0, 0x80));
  rCtrl.StyleSetBold(SCE_P_WORD2, 1);
  SetAStyle(SCE_P_STRING, RGB(0x80, 0, 0x80));
  SetAStyle(SCE_P_CHARACTER, RGB(0x80, 0, 0x80));
  SetAStyle(SCE_P_TRIPLE, RGB(0x80, 0, 0x80));
  SetAStyle(SCE_P_TRIPLEDOUBLE, RGB(0x80, 0, 0x80));
  SetAStyle(SCE_P_IDENTIFIER, RGB(0, 0, 0));
  SetAStyle(SCE_P_CLASSNAME, RGB(0, 0, 0));
  SetAStyle(SCE_P_DEFNAME, RGB(0, 0, 0));
  SetAStyle(SCE_P_OPERATOR, RGB(0x80, 0, 0));

  //Setup our markers
  rCtrl.SetMarginWidthN(1, 16);
  rCtrl.SetMarginTypeN(1, Scintilla::MarginType::Symbol);
  rCtrl.MarkerDefine(SC_PYTHON_MARK_ENABLED_BREAKPOINT, Scintilla::MarkerSymbol::Circle);
  rCtrl.MarkerSetFore(SC_PYTHON_MARK_ENABLED_BREAKPOINT, RGB(255, 0, 0));
  rCtrl.MarkerSetBack(SC_PYTHON_MARK_ENABLED_BREAKPOINT, RGB(255, 0, 0));
  rCtrl.MarkerDefine(SC_PYTHON_MARK_DISABLED_BREAKPOINT, Scintilla::MarkerSymbol::Circle);
  rCtrl.MarkerSetFore(SC_PYTHON_MARK_DISABLED_BREAKPOINT, RGB(255, 0, 0));
  rCtrl.MarkerDefine(SC_PYTHON_MARK_INSTRUCTION_POINTER, Scintilla::MarkerSymbol::ShortArrow);
  rCtrl.MarkerSetFore(SC_PYTHON_MARK_INSTRUCTION_POINTER, RGB(0, 0, 0));
  rCtrl.MarkerSetBack(SC_PYTHON_MARK_INSTRUCTION_POINTER, RGB(255, 255, 0));

  //Setup the notification for modified text
  rCtrl.SetModEventMask(Scintilla::ModificationFlags::InsertText | Scintilla::ModificationFlags::DeleteText);
}

void CPyDebugView::OnModified(_Inout_ Scintilla::NotificationData* /*pSCNotification*/)
{
#pragma warning(suppress: 26429)
  CPyDebugDoc* pDoc{GetDocument()};
  ASSERT(pDoc != nullptr);

  pDoc->UpdateBreakpointsFromUI();
}

#pragma warning(suppress: 26434)
int CPyDebugView::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
  //Let the base class do its thing
  if (__super::OnCreate(lpCreateStruct) == -1)
    return -1;

  //Create the Python Lexer
  m_pPythonLexer = theApp.m_pCreateLexer("python");
  if (m_pPythonLexer == nullptr)
    return -1;

  return 0;
}
