/*
Module : PythonPtr.h
Purpose: Defines the interface for a C++ wrapper class for Python Py_Object*'s as used in the PyDebug++ framework
Created: PJN / 20-08-2015

Copyright (c) 2015 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef __PYTHONPTR_H__
#define __PYTHONPTR_H__

#ifndef CPYDEBUG_EXT_CLASS
#define CPYDEBUG_EXT_CLASS
#endif //#ifndef CPYDEBUG_EXT_CLASS


//////////////////// Classes //////////////////////////////////////////////////

//Simple RAII wrapper for Python's "PyObject*"
class CPYDEBUG_EXT_CLASS CPyObject
{
public:
//Constructors / Destructors
  CPyObject() noexcept : p{nullptr}
  {
  }

  CPyObject(_Inout_opt_ PyObject* lp, bool bIncRef = false) noexcept : p{lp}
  {
    if (bIncRef && (p != nullptr))
      Py_INCREF(p);
  }

#pragma warning(suppress: 26495)
  CPyObject(_Inout_ const CPyObject& lp) noexcept : p{lp.p}
  {
    if (p != nullptr)
      Py_INCREF(p);
  }

  CPyObject(_Inout_ CPyObject&& lp) noexcept : p{nullptr}
  {
    lp.Swap(*this);
  }

  ~CPyObject() noexcept
  {
    if (p != nullptr)
#pragma warning(suppress: 26429 26486)
      Py_DECREF(p);
  }

//Methods
  PyObject* operator->() const noexcept
  {
    return p;
  }

  operator PyObject*() const noexcept
  {
    return p;
  }

  bool operator!() const noexcept
  {
    return (p == nullptr);
  }

  bool operator<(_In_opt_ PyObject* pT) const noexcept
  {
    return p < pT;
  }

  bool operator!=(_In_opt_ PyObject* pT) const noexcept
  {
    return !operator==(pT);
  }

  bool operator==(_In_opt_ PyObject* pT) const noexcept
  {
    return p == pT;
  }

  //Release the interface and set to nullptr
  void Release() noexcept
  {
    PyObject* pTemp{p};
    if (pTemp)
    {
      p = nullptr;
#pragma warning(suppress: 26429 26486)
      Py_DECREF(pTemp);
    }
  }

  //Attach to an existing interface (does not Py_INCREF)
  void Attach(_In_opt_ PyObject* p2) noexcept
  {
    if (p)
    {
#pragma warning(suppress: 26429 26486)
      Py_DECREF(p);
      //Attaching to the same object only works if duplicate references are being coalesced. Otherwise
      //re-attaching will cause the pointer to be released and may cause a crash on a subsequent dereference.
#pragma warning(suppress: 26477)
      ATLASSERT(p2 != p);
    }
    p = p2;
  }

  //Detach the interface (does not Release)
  PyObject* Detach() noexcept
  {
    PyObject* pt{p};
    p = nullptr;
    return pt;
  }

  _Check_return_ HRESULT CopyTo(_Outptr_result_maybenull_ PyObject** ppT) noexcept
  {
#pragma warning(suppress: 26477)
    ATLASSERT(ppT != nullptr);
    if (ppT == nullptr)
      return E_POINTER;
    *ppT = p;
    if (p)
      Py_INCREF(p);
    return S_OK;
  }

  PyObject* operator=(_Inout_ const CPyObject& lp) noexcept
  {
    if (*this != lp)
    {
      CPyObject{lp, false}.Swap(*this);
      if (p != nullptr)
        Py_INCREF(p);
    }
    return *this;
  }

  PyObject* operator=(_Inout_ CPyObject&& lp) noexcept
  {
    if (*this != lp)
      Swap(lp);
    return *this;
  }

//Member variables
  PyObject* p;

protected:
//Methods
  void Swap(CPyObject& other) noexcept
  {
    PyObject* pTemp{p};
    p = other.p;
    other.p = pTemp;
  }
};


#endif //#ifndef __PYTHONPTR_H__
